/*
 * Copyright 2011 Freescale Semiconductor, Inc.
 *
 * Implements the clk api defined in include/linux/clk.h
 *
 * Author: Chunhe Lan <Chunhe.Lan@freescale.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/spinlock.h>
#include <linux/delay.h>
#include <linux/clk.h>
#include <linux/of_platform.h>
#include <asm/clk_interface.h>
#include <asm/clock.h>

static LIST_HEAD(clocks);
static DEFINE_SPINLOCK(clk_lock);

static struct clk ipg_clk = {
	.name		= "pwm",
	.rate_hz	= 0,
	.users		= 1,		/* always on */
	.id		= 0,
};

static struct clk *all_clocks[] __initdata = {
	&ipg_clk,
};

/* clocks cannot be de-registered no refcounting necessary */
static struct clk *fsl_clk_get(struct device *dev, const char *id)
{
	struct clk *clk;

	list_for_each_entry(clk, &clocks, node) {
		if (strcmp(id, clk->name) == 0)
			return clk;
	}

	return ERR_PTR(-ENOENT);
}

static void fsl_clk_put(struct clk *clk)
{
}

static void __clk_enable(struct clk *clk)
{
	if (clk->parent)
		__clk_enable(clk->parent);
	if (clk->users++ == 0 && clk->mode)
		clk->mode(clk, 1);
}

static int fsl_clk_enable(struct clk *clk)
{
	unsigned long flags;

	spin_lock_irqsave(&clk_lock, flags);
	__clk_enable(clk);
	spin_unlock_irqrestore(&clk_lock, flags);

	return 0;
}

static void __clk_disable(struct clk *clk)
{
	BUG_ON(clk->users == 0);

	if (--clk->users == 0 && clk->mode)
		clk->mode(clk, 0);
	if (clk->parent)
		__clk_disable(clk->parent);
}

static void fsl_clk_disable(struct clk *clk)
{
	unsigned long flags;

	spin_lock_irqsave(&clk_lock, flags);
	__clk_disable(clk);
	spin_unlock_irqrestore(&clk_lock, flags);
}

static unsigned long fsl_clk_get_rate(struct clk *clk)
{
	unsigned long flags;
	unsigned long rate;

	spin_lock_irqsave(&clk_lock, flags);
	for (;;) {
		rate = clk->rate_hz;
		if (rate || !clk->parent)
			break;
		clk = clk->parent;
	}
	spin_unlock_irqrestore(&clk_lock, flags);

	return rate;
}

static struct clk_interface fsl_clk_functions = {
	.clk_get		= fsl_clk_get,
	.clk_put		= fsl_clk_put,
	.clk_enable		= fsl_clk_enable,
	.clk_disable		= fsl_clk_disable,
	.clk_get_rate		= fsl_clk_get_rate,
};

/* Register a new clock */
int clk_register(struct clk *clk)
{
	unsigned long flags;

	spin_lock_irqsave(&clk_lock, flags);
	list_add(&clk->node, &clocks);
	spin_unlock_irqrestore(&clk_lock, flags);

	return 0;
}
EXPORT_SYMBOL(clk_register);

static int fsl_register_clocks(void)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(all_clocks); i++)
		clk_register(all_clocks[i]);

	return 0;
}

static int __init fsl_clk_init(void)
{
	fsl_register_clocks();
	clk_functions = fsl_clk_functions;

	return 0;
}

arch_initcall(fsl_clk_init);
