/*
 * Copyright 2009-2011 Freescale Semiconductor, Inc.
 *
 * QorIQ (P1/P2) L2 controller init for Cache-SRAM instantiation
 *
 * Author: Vivek Mahajan <vivek.mahajan@freescale.com>
 * Modifier: Harninder Rai <harninder.rai@freescale.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/kernel.h>
#include <linux/of_platform.h>
#include <asm/io.h>

#include "fsl_85xx_cache_ctlr.h"

static char *sram_size;
static char *sram_offset;
static char *stash_size;
static char *stash_offset;
struct mpc85xx_l2ctlr __iomem *l2ctlr;

static long get_cache_sram_size(void)
{
	unsigned long val;

	if (!sram_size || (strict_strtoul(sram_size, 0, &val) < 0))
		return -EINVAL;

	return val;
}

static long get_cache_sram_offset(void)
{
	unsigned long val;

	if (!sram_offset || (strict_strtoul(sram_offset, 0, &val) < 0))
		return -EINVAL;

	return val;
}

static long get_stash_size(void)
{
	unsigned long val;

	if (!stash_size || (strict_strtoul(stash_size, 0, &val) < 0))
		return -EINVAL;

	return val;
}

static long get_stash_offset(void)
{
	unsigned long val;

	if (!stash_offset || (strict_strtoul(stash_offset, 0, &val) < 0))
		return -EINVAL;

	return val;
}

static int __init get_sram_size_from_cmdline(char *str)
{
	if (!str)
		return 0;

	sram_size = str;
	return 1;
}

static int __init get_sram_offset_from_cmdline(char *str)
{
	if (!str)
		return 0;

	sram_offset = str;
	return 1;
}

static int __init get_stash_size_from_cmdline(char *str)
{
	if (!str)
		return 0;

	stash_size = str;
	return 1;
}

static int __init get_stash_offset_from_cmdline(char *str)
{
	if (!str)
		return 0;

	stash_offset = str;
	return 1;
}

__setup("cache-sram-size=", get_sram_size_from_cmdline);
__setup("cache-sram-offset=", get_sram_offset_from_cmdline);
__setup("stash-size=", get_stash_size_from_cmdline);
__setup("stash-offset=", get_stash_offset_from_cmdline);

static int initialize_stash_region(struct of_device *dev,
		unsigned int l2cache_size)
{
	unsigned int size;
	u64 offset;
	u64 rem;
	unsigned char stash_config;

	size  = get_stash_size();
	if (size <= 0) {
		dev_err(&dev->dev, "Provide a legal stash size\n");
		return -EINVAL;
	}

	offset = get_stash_offset();
	if (offset <= 0) {
		dev_err(&dev->dev, "Provide a legal stash offset\n");
		return -EINVAL;
	}

	rem = (l2cache_size / 2) % size;
	stash_config = STASH_SIZE_INVALID * size / l2cache_size;
	if (rem || (stash_config & (stash_config - 1))) {
		dev_err(&dev->dev, "Invalid stash size in command line\n");
		return -EINVAL;
	}

	rem = offset & (size - 1);
	if (rem) {
		dev_err(&dev->dev, "Stash offset must be naturally aligned"
				"with stash size passed in command line\n");
		return -EINVAL;
	}

	if (!request_mem_region(offset, size, "fsl_85xx_stash_only_region")) {
		dev_err(&dev->dev, "%s: Request memory failed\n",
				dev->node->full_name);
		return -ENXIO;
	}

	out_be32(&l2ctlr->ewar0,
			offset & L2_EXT_ADDR_MSK_LO24);

#ifdef CONFIG_PHYS_64BIT
	out_be32(&l2ctlr->ewarea0,
			(offset >> 32) & L2_EXT_ADDR_MSK_HI4);
#endif

	clrsetbits_be32(&l2ctlr->ctl,
			L2CR_STASH_DIS | L2CR_L2E, L2CR_L2FI);

	switch (stash_config) {
	case STASH_SIZE_EIGHTH:
		setbits32(&l2ctlr->ctl,
			L2CR_L2E | L2CR_L2FI | L2CR_STASH_EIGHTH);
		break;

	case STASH_SIZE_QUARTER:
		setbits32(&l2ctlr->ctl,
			L2CR_L2E | L2CR_L2FI | L2CR_STASH_QUARTER);
		break;

	case STASH_SIZE_HALF:
	default:
		setbits32(&l2ctlr->ctl,
			L2CR_L2E | L2CR_L2FI | L2CR_STASH_HALF);
	}

	out_be32(&l2ctlr->ewcr0,
			L2_EXT_WRITE_EN | L2_EXT_WRITE_LOCK | (size >> 4));
	eieio();

	return 0;
}

static int initialize_sram_region(struct of_device *dev,
		unsigned int l2cache_size, struct sram_parameters sram_params)
{
	unsigned int rem;
	unsigned char ways;

	rem = l2cache_size % sram_params.sram_size;
	ways = LOCK_WAYS_FULL * sram_params.sram_size / l2cache_size;
	if (rem || (ways & (ways - 1))) {
		dev_err(&dev->dev, "Invalid cache-sram-size in command line\n");
		return -EINVAL;
	}

	/*
	 * Write bits[0-17] to srbar0
	 */
	out_be32(&l2ctlr->srbar0,
		sram_params.sram_offset & L2SRAM_BAR_MSK_LO18);

	/*
	 * Write bits[18-21] to srbare0
	 */
#ifdef CONFIG_PHYS_64BIT
	out_be32(&l2ctlr->srbarea0,
		(sram_params.sram_offset >> 32) & L2SRAM_BARE_MSK_HI4);
#endif

	clrsetbits_be32(&l2ctlr->ctl, L2CR_L2E, L2CR_L2FI);

	switch (ways) {
	case LOCK_WAYS_EIGHTH:
		setbits32(&l2ctlr->ctl,
			L2CR_L2E | L2CR_L2FI | L2CR_SRAM_EIGHTH);
		break;

	case LOCK_WAYS_TWO_EIGHTH:
		setbits32(&l2ctlr->ctl,
			L2CR_L2E | L2CR_L2FI | L2CR_SRAM_QUART);
		break;

	case LOCK_WAYS_HALF:
		setbits32(&l2ctlr->ctl,
			L2CR_L2E | L2CR_L2FI | L2CR_SRAM_HALF);
		break;

	case LOCK_WAYS_FULL:
	default:
		setbits32(&l2ctlr->ctl,
			L2CR_L2E | L2CR_L2FI | L2CR_SRAM_FULL);
		break;
	}
	eieio();

	return 0;
}

static int __devinit mpc85xx_l2ctlr_of_probe(struct of_device *dev,
			const struct of_device_id *match)
{
	long rval;
	const unsigned int *prop;
	unsigned int l2cache_size;
	struct sram_parameters sram_params;
	int ret = 0;

	if (!dev->node) {
		dev_err(&dev->dev, "Device's OF-node is NULL\n");
		return -EINVAL;
	}

	prop = of_get_property(dev->node, "cache-size", NULL);
	if (!prop) {
		dev_err(&dev->dev, "Missing L2 cache-size\n");
		return -EINVAL;
	}
	l2cache_size = *prop;

	l2ctlr = of_iomap(dev->node, 0);
	if (!l2ctlr) {
		dev_err(&dev->dev, "Can't map L2 controller\n");
		return -EINVAL;
	}

	sram_params.sram_size  = get_cache_sram_size();
	if (sram_params.sram_size <= 0) {
		dev_err(&dev->dev, "Provide a legal sram size\n");
		ret = -EINVAL;
		goto out_unmap;
	}

	sram_params.sram_offset = get_cache_sram_offset();
	if (sram_params.sram_offset <= 0) {
		dev_err(&dev->dev, "Provide a legal sram offset\n");
		ret = -EINVAL;
		goto out_unmap;
	}

	rval = initialize_sram_region(dev, l2cache_size, sram_params);
	if (rval == 0) {
		rval = instantiate_cache_sram(dev, sram_params);
		if (rval < 0)
			dev_err(&dev->dev, "Can't instantiate Cache-SRAM\n");
	} else {
		dev_err(&dev->dev, "Can't initialize Cache-SRAM\n");
	}

	rval = initialize_stash_region(dev, l2cache_size);
	if (rval < 0) {
		dev_err(&dev->dev, "Can't initialize stash region\n");
		ret = -EINVAL;
		goto out_unmap;
	}

	iounmap(l2ctlr);

	return 0;

out_unmap:
	iounmap(l2ctlr);
	return ret;
}

static int __devexit mpc85xx_l2ctlr_of_remove(struct of_device *dev)
{
	BUG_ON(!l2ctlr);

	iounmap(l2ctlr);
	remove_cache_sram(dev);
	dev_info(&dev->dev, "MPC85xx L2 controller unloaded\n");

	return 0;
}

static struct of_device_id mpc85xx_l2ctlr_of_match[] = {
	{
		.compatible = "fsl,p2020-l2-cache-controller",
	},
	{
		.compatible = "fsl,p2010-l2-cache-controller",
	},
	{
		.compatible = "fsl,p1020-l2-cache-controller",
	},
	{
		.compatible = "fsl,p1011-l2-cache-controller",
	},
	{
		.compatible = "fsl,p1013-l2-cache-controller",
	},
	{
		.compatible = "fsl,p1022-l2-cache-controller",
	},
	{
		.compatible = "fsl,p1010-l2-cache-controller",
	},
	{
		.compatible = "fsl,bsc9131-l2-cache-controller",
	},
};

static struct of_platform_driver mpc85xx_l2ctlr_of_platform_driver = {
		.name		= "fsl-l2ctlr",
		.match_table	= mpc85xx_l2ctlr_of_match,
	.probe		= mpc85xx_l2ctlr_of_probe,
	.remove		= __devexit_p(mpc85xx_l2ctlr_of_remove),
};

static __init int mpc85xx_l2ctlr_of_init(void)
{
	return of_register_platform_driver(&mpc85xx_l2ctlr_of_platform_driver);
}

static void __exit mpc85xx_l2ctlr_of_exit(void)
{
	of_unregister_platform_driver(&mpc85xx_l2ctlr_of_platform_driver);
}

subsys_initcall(mpc85xx_l2ctlr_of_init);
module_exit(mpc85xx_l2ctlr_of_exit);

MODULE_DESCRIPTION("Freescale MPC85xx L2 controller init");
MODULE_LICENSE("GPL v2");
