/*
 * Copyright (c) 2011-2012 Freescale Semiconductor, Inc.
 *
 * Freescale Subscriber Identification Module(SIM) device driver
 *
 * Author: Harninder Rai <harninder.rai@freescale.com>
 * Author: Ashish Kumar  <ashish.kumar@freescale.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/sched.h>
#include <linux/io.h>
#include <linux/poll.h>
#include <linux/miscdevice.h>
#include <linux/clk.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/interrupt.h>

#ifdef CONFIG_OF
#include <linux/of_platform.h>
/* Following include is for ARM platform
#else
#include <linux/platform_device.h> */
#endif

#include <linux/fsl_sim_interface.h>
#include <linux/fsl_devices.h>

#include "sim.h"

#define SIM_INTERNAL_CLK  0
#define SIM_RFU          -1

/* Default communication parameters: FI=372, DI=1, PI1=5V, II=50mA, WWT=10 */
#define SIM_PARAM_DEFAULT { 0, 1, 1, 5, 1, 0, 0, 0, 10 }

/* PTS structure location*/
#define SIM_PTS_PPSS 0
#define SIM_PTS_PPS0 1
#define SIM_PTS_PPS1 2
#define SIM_PTS_PPS2 3
#define SIM_PTS_PPS3 4
#define SIM_PTS_PCK 5
#define SIM_PTS_PPS0_BIT5 0x10
#define SIM_PTS_PPS0_BIT6 0x20
#define SIM_PTS_PPS0_BIT7 0x40
#define SIM_PTS_PPS0_BIT_76X 0x60
#define SIM_PTS_PPS0_BIT_7X5 0x50
#define SIM_PTS_PPS0_BIT_X65 0x30
#define SIM_PTS_PPS0_BIT_765 0x70

/* Interface character references */
#define SIM_IFC_TXI(letter, number) (letter + number * 4)
#define SIM_IFC_TA1   SIM_IFC_TXI(0, 0)
#define SIM_IFC_TB1   SIM_IFC_TXI(1, 0)
#define SIM_IFC_TC1   SIM_IFC_TXI(2, 0)
#define SIM_IFC_TD1   SIM_IFC_TXI(3, 0)
#define SIM_IFC_TA2   SIM_IFC_TXI(0, 1)
#define SIM_IFC_TB2   SIM_IFC_TXI(1, 1)
#define SIM_IFC_TC2   SIM_IFC_TXI(2, 1)
#define SIM_IFC_TD2   SIM_IFC_TXI(3, 1)
#define SIM_IFC_TA3   SIM_IFC_TXI(0, 2)
#define SIM_IFC_TB3   SIM_IFC_TXI(1, 2)
#define SIM_IFC_TC3   SIM_IFC_TXI(2, 2)
#define SIM_IFC_TD3   SIM_IFC_TXI(3, 2)
#define SIM_IFC_TA4   SIM_IFC_TXI(0, 3)
#define SIM_IFC_TB4   SIM_IFC_TXI(1, 3)
#define SIM_IFC_TC4   SIM_IFC_TXI(2, 3)
#define SIM_IFC_TD4   SIM_IFC_TXI(3, 3)

/* ATR and OPS states */
#define SIM_STATE_REMOVED              0
#define SIM_STATE_OPERATIONAL_IDLE     1
#define SIM_STATE_OPERATIONAL_COMMAND  2
#define SIM_STATE_OPERATIONAL_RESPONSE 3
#define SIM_STATE_OPERATIONAL_STATUS1  4
#define SIM_STATE_OPERATIONAL_STATUS2  5
#define SIM_STATE_DETECTED_ATR_T0       7
#define SIM_STATE_DETECTED_ATR_TS       8
#define SIM_STATE_DETECTED_ATR_TXI      9
#define SIM_STATE_DETECTED_ATR_THB      10
#define SIM_STATE_DETECTED_ATR_TCK      11
#define SIM_STATE_OPERATIONAL_PPSS     6
#define SIM_STATE_OPERATIONAL_PPS0     12
#define SIM_STATE_OPERATIONAL_PPS1     13
#define SIM_STATE_OPERATIONAL_PPS2     14
#define SIM_STATE_OPERATIONAL_PPS3     15
#define SIM_STATE_OPERATIONAL_PCK      16

/* Definitions of the offset of the SIM hardware registers */
#define PORT1_CNTL	0x00	/* 00 */
#define SETUP		0x04	/* 04 */
#define PORT1_DETECT	0x08	/* 08 */
#define PORT1_XMT_BUF	0x0C	/* 0c */
#define PORT1_RCV_BUF	0x10	/* 10 */
#define PORT0_CNTL	0x14	/* 14 */
#define CNTL		0x18	/* 18 */
#define CLK_PRESCALER	0x1C	/* 1c */
#define RCV_THRESHOLD	0x20	/* 20 */
#define ENABLE		0x24	/* 24 */
#define XMT_STATUS	0x28	/* 28 */
#define RCV_STATUS	0x2C	/* 2c */
#define INT_MASK	0x30	/* 30 */
#define PORTO_XMT_BUF	0x34	/* 34 */
#define PORT0_RCV_BUF	0x38	/* 38 */
#define PORT0_DETECT	0x3C	/* 3c */
#define DATA_FORMAT	0x40	/* 40 */
#define XMT_THRESHOLD	0x44	/* 44 */
#define GUARD_CNTL	0x48	/* 48 */
#define OD_CONFIG	0x4C	/* 4c */
#define RESET_CNTL	0x50	/* 50 */
#define CHAR_WAIT	0x54	/* 54 */
#define GPCNT		0x58	/* 58 */
#define DIVISOR		0x5C	/* 5c */
#define BWT		0x60	/* 60 */
#define BGT		0x64	/* 64 */
#define BWT_H		0x68	/* 68 */
#define XMT_FIFO_STAT	0x6C	/* 6c */
#define RCV_FIFO_CNT	0x70	/* 70 */
#define RCV_FIFO_WPTR	0x74	/* 74 */
#define RCV_FIFO_RPTR	0x78	/* 78 */

/* SIM port[0|1]_cntl register bits */
#define SIM_PORT_CNTL_SFPD   (1<<7)
#define SIM_PORT_CNTL_3VOLT  (1<<6)
#define SIM_PORT_CNTL_SCSP   (1<<5)
#define SIM_PORT_CNTL_SCEN   (1<<4)
#define SIM_PORT_CNTL_SRST   (1<<3)
#define SIM_PORT_CNTL_STEN   (1<<2)
#define SIM_PORT_CNTL_SVEN   (1<<1)
#define SIM_PORT_CNTL_SAPD   (1<<0)

/* SIM od_config register bits */
#define SIM_OD_CONFIG_OD_P1  (1<<1)
#define SIM_OD_CONFIG_OD_P0  (1<<0)

/* SIM enable register bits */
#define SIM_ENABLE_XMTEN     (1<<1)
#define SIM_ENABLE_RCVEN     (1<<0)

/* SIM int_mask register bits */
#define SIM_INT_MASK_RFEM    (1<<13)
#define SIM_INT_MASK_BGTM    (1<<12)
#define SIM_INT_MASK_BWTM    (1<<11)
#define SIM_INT_MASK_RTM     (1<<10)
#define SIM_INT_MASK_CWTM    (1<<9)
#define SIM_INT_MASK_GPCM    (1<<8)
#define SIM_INT_MASK_TDTFM   (1<<7)
#define SIM_INT_MASK_TFOM    (1<<6)
#define SIM_INT_MASK_XTM     (1<<5)
#define SIM_INT_MASK_TFEIM   (1<<4)
#define SIM_INT_MASK_ETCIM   (1<<3)
#define SIM_INT_MASK_OIM     (1<<2)
#define SIM_INT_MASK_TCIM    (1<<1)
#define SIM_INT_MASK_RIM     (1<<0)

/* SIM xmt_status register bits */
#define SIM_XMT_STATUS_GPCNT (1<<8)
#define SIM_XMT_STATUS_TDTF  (1<<7)
#define SIM_XMT_STATUS_TFO   (1<<6)
#define SIM_XMT_STATUS_TC    (1<<5)
#define SIM_XMT_STATUS_ETC   (1<<4)
#define SIM_XMT_STATUS_TFE   (1<<3)
#define SIM_XMT_STATUS_XTE   (1<<0)

/* SIM rcv_status register bits */
#define SIM_RCV_STATUS_BGT   (1<<11)
#define SIM_RCV_STATUS_BWT   (1<<10)
#define SIM_RCV_STATUS_RTE   (1<<9)
#define SIM_RCV_STATUS_CWT   (1<<8)
#define SIM_RCV_STATUS_CRCOK (1<<7)
#define SIM_RCV_STATUS_LRCOK (1<<6)
#define SIM_RCV_STATUS_RDRF  (1<<5)
#define SIM_RCV_STATUS_RFD   (1<<4)
#define SIM_RCV_STATUS_RFE   (1<<1)
#define SIM_RCV_STATUS_OEF   (1<<0)

/* SIM cntl register bits */
#define SIM_CNTL_BWTEN       (1<<15)
#define SIM_CNTL_XMT_CRC_LRC (1<<14)
#define SIM_CNTL_CRCEN       (1<<13)
#define SIM_CNTL_LRCEN       (1<<12)
#define SIM_CNTL_CWTEN       (1<<11)
#define SIM_CNTL_SAMPLE12    (1<<4)
#define SIM_CNTL_ONACK       (1<<3)
#define SIM_CNTL_ANACK       (1<<2)
#define SIM_CNTL_ICM         (1<<1)
#define SIM_CNTL_GPCNT_CLK_SEL(x)   ((x&0x03)<<9)
#define SIM_CNTL_GPCNT_CLK_SEL_MASK     (0x03<<9)
#define SIM_CNTL_BAUD_SEL(x)        ((x&0x07)<<6)
#define SIM_CNTL_BAUD_SEL_MASK          (0x07<<6)

/* SIM rcv_threshold register bits */
#define SIM_RCV_THRESHOLD_RTH(x)    ((x&0x0f)<<9)
#define SIM_RCV_THRESHOLD_RTH_MASK      (0x0f<<9)
#define SIM_RCV_THRESHOLD_RDT(x)   ((x&0x1ff)<<0)
#define SIM_RCV_THRESHOLD_RDT_MASK     (0x1ff<<0)

/* SIM xmt_threshold register bits */
#define SIM_XMT_THRESHOLD_XTH(x)    ((x&0x0f)<<4)
#define SIM_XMT_THRESHOLD_XTH_MASK      (0x0f<<4)
#define SIM_XMT_THRESHOLD_TDT(x)    ((x&0x0f)<<0)
#define SIM_XMT_THRESHOLD_TDT_MASK      (0x0f<<0)

/* SIM guard_cntl register bits */
#define SIM_GUARD_CNTL_RCVR11              (1<<8)
#define SIM_GIARD_CNTL_GETU(x)           (x&0xff)
#define SIM_GIARD_CNTL_GETU_MASK           (0xff)

/* SIM port[0|]_detect register bits */
#define SIM_PORT_DETECT_SPDS  (1<<3)
#define SIM_PORT_DETECT_SPDP  (1<<2)
#define SIM_PORT_DETECT_SDI   (1<<1)
#define SIM_PORT_DETECT_SDIM  (1<<0)

/* END of REGS definitions */

static void sim_power_off(struct sim_t *);
static void sim_data_reset(struct sim_t *);
static void sim_stop(struct sim_t *);
static void sim_deactivate(struct sim_t *);
static void sim_counter_enable(struct sim_t *);
static void sim_counter_disable(struct sim_t *sim, uint32_t);
static void sim_char_wait_enable(struct sim_t *);
static void sim_char_wait_disable(struct sim_t *);
static void sim_block_wait_enable(struct sim_t *);
static void sim_block_wait_disable(struct sim_t *);

static int sim_param_F[] = {
	SIM_INTERNAL_CLK, 372, 558, 744, 1116, 1488, 1860, SIM_RFU,
	SIM_RFU, 512, 768, 1024, 1536, 2048, SIM_RFU, SIM_RFU
};

static int sim_param_D[] = {
	SIM_RFU, 64 * 1, 64 * 2, 64 * 4, 64 * 8, 64 * 16, 64 * 32, 64 * 64,
	64 * 12, 64 * 20, SIM_RFU, SIM_RFU, SIM_RFU, SIM_RFU, SIM_RFU, SIM_RFU
};

/* Function: sim_calc_param
 *
 * Description: determine register values depending on communication parameters
 *
 * Parameters:
 * uint32_t fi             ATR frequency multiplier index
 * uint32_t di             ATR frequency divider index
 * uint32_t* ptr_divisor   location to store divisor result
 * uint32_t* ptr_sample12  location to store sample12 result
 *
 * Return Values:
 *  SIM_OK                          calculation finished without errors
 * -SIM_E_PARAM_DIVISOR_RANGE       calculated divisor > 255
 * -SIM_E_PARAM_FBYD_NOTDIVBY8OR12  F/D not divisable by 12 (as required)
 * -SIM_E_PARAM_FBYD_WITHFRACTION   F/D has a remainder
 * -SIM_E_PARAM_DI_INVALID          frequency multiplyer index not supported
 * -SIM_E_PARAM_FI_INVALID          frequency divider index not supported
 */

static int sim_calc_param(uint32_t fi, uint32_t di, uint32_t *ptr_divisor,
			  uint32_t *ptr_sample12)
{
	int32_t errval = SIM_OK;
	int32_t f = sim_param_F[fi];
	int32_t d = sim_param_D[di];
	int32_t stage2_fra = (64 * f) % d;
	int32_t stage2_div = (64 * f) / d;
	int32_t sample12 = 1;
	uint32_t divisor = 31;

	pr_debug("%s entering.\n", __func__);
	if ((f > 0) || (d > 0)) {
		if (stage2_fra == 0) {
			if ((stage2_div % 12) == 0) {
				sample12 = 1;
				divisor = stage2_div / 12;
			} else if ((stage2_div % 8) == 0) {
				sample12 = 0;
				divisor = stage2_div / 8;
			} else
				sample12 = -1;
			if (sample12 >= 0) {
				if (divisor < 256) {
					pr_debug("fi=%i\n", fi);
					pr_debug("di=%i\n", di);
					pr_debug("f=%i\n", f);
					pr_debug("d=%i/64\n", d);
					pr_debug("div=%i\n", stage2_div);
					pr_debug("divisor=%i\n", divisor);
					pr_debug("sample12=%i\n", sample12);

					*ptr_divisor = divisor;
					*ptr_sample12 = sample12;
					errval = SIM_OK;
				} else
					errval = -SIM_E_PARAM_DIVISOR_RANGE;
			} else
				errval = -SIM_E_PARAM_FBYD_NOTDIVBY8OR12;
		} else
			errval = -SIM_E_PARAM_FBYD_WITHFRACTION;
	} else
		errval = -SIM_E_PARAM_FI_INVALID;

	return errval;
};

/* Function: sim_set_param
 *
 * Description: apply communication parameters (setup devisor and sample12)
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 * sim_param_t* param      pointer to communication parameters
 *
 * Return Values:
 * see function sim_calc_param
 */

static int sim_set_param(struct sim_t *sim, sim_param_t *param)
{
	uint32_t divisor, sample12, reg_data;
	int errval;

	pr_debug("%s entering.\n", __func__);
	errval = sim_calc_param(param->FI, param->DI, &divisor, &sample12);
	if (errval == SIM_OK) {
		sim_write(divisor, sim->ioaddr + DIVISOR);
		if (sample12) {
			reg_data = sim_read(sim->ioaddr + CNTL);
			reg_data |= SIM_CNTL_SAMPLE12;
			sim_write(reg_data, sim->ioaddr + CNTL);
		} else {
			reg_data = sim_read(sim->ioaddr + CNTL);
			reg_data &= ~SIM_CNTL_SAMPLE12;
			sim_write(reg_data, sim->ioaddr + CNTL);
		}
	}

	return errval;
};

/* Function: sim_atr_received
 *
 * Description: this function is called whenever a valid ATR has been received.
 * It determines the communication parameters from the ATR received and notifies
 * the user space application with SIGIO.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_atr_received(struct sim_t *sim)
{
	sim_param_t param_default = SIM_PARAM_DEFAULT;
	sim->param_atr = param_default;

	pr_debug("%s entering.\n", __func__);
	if (sim->atrparser.ifc_valid & (1 << (SIM_IFC_TA1))) {
		sim->param_atr.FI = sim->atrparser.TXI[SIM_IFC_TA1] >> 4;
		sim->param_atr.DI = sim->atrparser.TXI[SIM_IFC_TA1] & 0x0f;
		pr_debug("valid FI DI %d %d\n", sim->param_atr.FI,
			sim->param_atr.DI);
	}
	if (sim->atrparser.ifc_valid & (1 << (SIM_IFC_TB1))) {
		sim->param_atr.PI1 = (sim->atrparser.TXI[SIM_IFC_TB1] >> 4)
		    & 0x07;
		sim->param_atr.II = sim->atrparser.TXI[SIM_IFC_TB1] & 0x07f;
	}
	if (sim->atrparser.ifc_valid & (1 << (SIM_IFC_TC1)))
		sim->param_atr.N = sim->atrparser.TXI[SIM_IFC_TC1];

	if (sim->atrparser.ifc_valid & (1 << (SIM_IFC_TD1)))
		sim->param_atr.T = sim->atrparser.TXI[SIM_IFC_TD1] & 0x0f;

	if (sim->atrparser.ifc_valid & (1 << (SIM_IFC_TB2)))
		sim->param_atr.PI2 = sim->atrparser.TXI[SIM_IFC_TB2];

	if (sim->atrparser.ifc_valid & (1 << (SIM_IFC_TC2)))
		sim->param_atr.WWT = sim->atrparser.TXI[SIM_IFC_TC2];

	if (sim->fasync)
		kill_fasync(&sim->fasync, SIGIO, POLL_IN);

};
static void sim_pps_valid(struct sim_t *sim, uint32_t data)
{
	uint8_t temp = 0x00;
	int i, j, count = 0;

	pr_debug("%s in %x\n", __func__, data);
	pr_debug("%s in %x\n ", __func__, sim->xmt_buffer[SIM_PTS_PPS0]);

	if (sim->xmt_buffer[SIM_PTS_PPS0] == data) {
		for (i = 0; i < sim->xfer.xmt_length; i++) {
			pr_debug("%s in %x %x\n", __func__, i,
				sim->xmt_buffer[i]);
			temp = temp ^ sim->xmt_buffer[i];
		}

		if (temp == 00) {
			for (j = 0; j < sim->xfer.xmt_length; j++) {
				if ((sim->xmt_buffer[j] && 1) == 0)
					sim->sim_pts_valid_flag = 0;
				else
					++count;
			}
		}
	}

	if (count == sim->xfer.xmt_length)
		sim->sim_pts_valid_flag = 1;
	pr_debug("count = %x sim_pts_valid_flag =%x\n", count,
			sim->sim_pts_valid_flag);
	return;
}

/* Function: sim_pts_received
 **
 ** Description: this function is called whenever a PTS sequence is called.
 ** It determines the communication parameters from the ATR received
 ** and compares with PTS PPS1 value and uses the appropriate value
 **
 ** Parameters:
 ** struct sim_t* sim  pointer to SIM device handler
 **/

static void sim_pts_received(struct sim_t *sim, uint32_t data)
{
	uint8_t pts_FI = ((data & 0xf0) >> 4);
	uint8_t pts_DI = (data & 0x0f);

	sim->param_pts = sim->param_atr;
	pr_debug("%s pts_FI=%x pts_di=%x\n", __func__, pts_FI, pts_DI);
	pr_debug("before pts_fi= %d pts_di= %d\n",
		sim->param_pts.FI, sim->param_pts.DI);
	if ((pts_FI <= sim->param_atr.FI) &&
		(pts_DI <= sim->param_atr.DI) &&
		((pts_DI && pts_FI) != 0)) {
			sim->param_pts.FI = pts_FI;
			sim->param_pts.DI = pts_DI;
			pr_debug("%s pts_FI=%x pts_di=%x\n", __func__, pts_FI,
				pts_DI);
	} else
		sim->sim_pts_valid_flag = 0;

	pr_debug("after pts_fi= %d pts_di= %d\n",
		sim->param_pts.FI, sim->param_pts.DI);

	return;
};

/* Function: sim_xmt_fill
 *
 * Description: fill the transmit FIFO until the FIFO is full or
 * the end of the transmission has been reached.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_xmt_fill(struct sim_t *sim)
{
	uint32_t reg_data;
	int bytesleft;
	int temp = sim->xmt_remaining;
	int count = 0;

	reg_data = sim_read(sim->ioaddr + XMT_FIFO_STAT);
	bytesleft = 16 - ((reg_data >> 8) & 0x0f);

	pr_debug("txfill: remaining=%i bytesleft=%i\n",
		 sim->xmt_remaining, bytesleft);
	if (bytesleft > sim->xmt_remaining)
		bytesleft = sim->xmt_remaining;

	sim->xmt_remaining -= bytesleft;
	for (; bytesleft > 0; bytesleft--) {
		++count;
		sim_write(sim->xmt_buffer[sim->xmt_pos],
			     sim->ioaddr + PORT1_XMT_BUF);
		if ((temp - count) == 1) {
			sim_block_wait_enable(sim);
			sim->sim_xmt_last_byte = 1;
		}
		sim->xmt_pos++;
	};
/* FIXME: optimization - keep filling until fifo full */
};

/* Function: sim_xmt_start
 *
 * Description: initiate a transfer
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 * int pos                 position in the xfer transmit buffer
 * int count               number of bytes to be transmitted
 */

static void sim_xmt_start(struct sim_t *sim, int pos, int count)
{
	uint32_t reg_data;

	pr_debug("%s: tx\n", __func__);
	sim->xmt_remaining = count;
	sim->xmt_pos = pos;
	sim_xmt_fill(sim);

	if (sim->xmt_remaining) {
		reg_data = sim_read(sim->ioaddr + INT_MASK);
		reg_data &= ~SIM_INT_MASK_TDTFM;
		sim_write(reg_data, sim->ioaddr + INT_MASK);
	} else {
		reg_data = sim_read(sim->ioaddr + INT_MASK);
		reg_data &= ~SIM_INT_MASK_TCIM;
		sim_write(reg_data, sim->ioaddr + INT_MASK);
	}
		sim_write(SIM_XMT_STATUS_TC | SIM_XMT_STATUS_TDTF,
			     sim->ioaddr + XMT_STATUS);
		reg_data = sim_read(sim->ioaddr + ENABLE);
		reg_data |= SIM_ENABLE_XMTEN;
		sim_write(reg_data, sim->ioaddr + ENABLE);
};

/* Function: sim_atr_add
 *
 * Description: add a byte to the raw ATR string
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 * uint8_t data            byte to be added
 */

static void sim_atr_add(struct sim_t *sim, uint32_t data)
{
	pr_debug("%s entering.\n", __func__);
	if (sim->atr.size < SIM_ATR_LENGTH_MAX)
		sim->atr.t[sim->atr.size++] = (uint8_t)data;
	else
		pr_debug("sim.c: ATR received is too big!\n");
};

/* Function: sim_fsm
 *
 * Description: main finite state machine running in ISR context.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 * uint8_t data            byte received
 */

static void sim_fsm(struct sim_t *sim, uint32_t data)
{
	uint32_t temp = 0, i = 0, reg_data0 = 0;
	uint32_t TDx_valid[4] = {0, 0};
	uint8_t TDx_first4bits[4] = {0, 0};
	uint8_t cnt_type0 = 0, cnt_type15 = 0, tck_true = 0;
	switch (sim->state) {
		/* OPS FSM */

	case SIM_STATE_OPERATIONAL_IDLE:
		pr_debug("data received unexpectidly (%04x)\n", data);
		break;

	case SIM_STATE_OPERATIONAL_COMMAND:
		sim_char_wait_disable(sim);
		if (data == 0x60) {
			pr_debug("NULL byte in Procedure byte value =60\n");
			break;
		} else if (((data & 0xf0) == 0x60) | ((data & 0xf0) == 0x90)) {
			pr_debug("6x or 9x state\n");
			sim->xfer.sw1 = (uint8_t)data;
			sim->state = SIM_STATE_OPERATIONAL_STATUS2;
		} else if (data == sim->xmt_buffer[1]) {
			pr_debug("data=%x rcv_length=%x in INS\n", data,
				sim->xfer.rcv_length);
			if (sim->xfer.rcv_length) {
				sim->ins_complimentary_flag = 0;
				sim->state = SIM_STATE_OPERATIONAL_RESPONSE;
			} else {
				sim->state = SIM_STATE_OPERATIONAL_STATUS1;
				if (sim->xfer.xmt_length >
					sim->xfer.xmt_position) {
					sim_xmt_start(sim,
						sim->xfer.xmt_position,
						sim->xfer.xmt_length -
						sim->xfer.xmt_position);
					break;
				}
			}
		} else if ((data == (0xff ^ sim->xmt_buffer[1])) &&
			(((0xff ^ sim->xmt_buffer[1]) & 0xf0) != 0x60) &&
			(((0xff ^ sim->xmt_buffer[1]) & 0xf0) != 0x90)
			) {
			pr_debug("Entering in the Ins complimentary state \n");
			if (sim->xfer.rcv_length) {
				sim->ins_complimentary_flag = 1;
				sim->state = SIM_STATE_OPERATIONAL_RESPONSE;
			} else {
				if (sim->xfer.xmt_length >
						sim->xfer.xmt_position) {
					sim_xmt_start(sim,
						sim->xfer.xmt_position, 1);
					sim->xfer.xmt_position =
						sim->xfer.xmt_position + 1;
					pr_debug("xfer.xmt_position=%x\n",
						sim->xfer.xmt_position);
					break;
				} else {
					sim->state =
						SIM_STATE_OPERATIONAL_STATUS1;
				}
			}
		} else {
			sim->errval = -SIM_E_NACK;
			complete(&sim->xfer_done);
		}
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_RESPONSE:
		sim_char_wait_disable(sim);
		sim->rcv_buffer[sim->rcv_count] = (uint8_t)data;
		sim->rcv_count++;
		if (sim->rcv_count == sim->xfer.rcv_length)
			sim->state = SIM_STATE_OPERATIONAL_STATUS1;
		else if (sim->ins_complimentary_flag == 1)
			sim->state = SIM_STATE_OPERATIONAL_COMMAND;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_STATUS1:
		sim_char_wait_disable(sim);
		sim->xfer.sw1 = (uint8_t)data;
		sim->state = SIM_STATE_OPERATIONAL_STATUS2;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_STATUS2:
		sim_char_wait_disable(sim);
		sim->xfer.sw2 = (uint8_t)data;
		sim->state = SIM_STATE_OPERATIONAL_IDLE;
		complete(&sim->xfer_done);
		break;

	case SIM_STATE_OPERATIONAL_PPSS:
		sim->rcv_buffer[sim->rcv_count] = (uint8_t)data;
		sim->rcv_count++;
		pr_debug("data PPSS %x\n", data);
		sim->state = SIM_STATE_OPERATIONAL_PPS0;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_PPS0:
		sim_char_wait_disable(sim);
		sim->rcv_buffer[sim->rcv_count] = (uint8_t)data;
		sim->rcv_count++;
		sim_pps_valid(sim, data);
		pr_debug("data PPS0 %x\n", data);
		if ((data & SIM_PTS_PPS0_BIT_765) == SIM_PTS_PPS0_BIT_765) {
			pr_debug("SIM_PTS_PPS0_BIT_765\n");
			sim->state = SIM_STATE_OPERATIONAL_PPS1;
			sim->sim_pts_pps2_flag = 1;
			sim->sim_pts_pps3_flag = 1;
		} else if ((data & SIM_PTS_PPS0_BIT_76X) ==
				SIM_PTS_PPS0_BIT_76X) {
			sim->state = SIM_STATE_OPERATIONAL_PPS2;
			sim->sim_pts_pps3_flag = 1;
		} else if ((data & SIM_PTS_PPS0_BIT_X65) ==
				SIM_PTS_PPS0_BIT_X65) {
			sim->state = SIM_STATE_OPERATIONAL_PPS1;
			sim->sim_pts_pps2_flag = 1;
		} else if ((data & SIM_PTS_PPS0_BIT_7X5) ==
				SIM_PTS_PPS0_BIT_7X5) {
			sim->state = SIM_STATE_OPERATIONAL_PPS1;
			sim->sim_pts_pps3_flag = 1;
		} else if (data & SIM_PTS_PPS0_BIT5)
			sim->state = SIM_STATE_OPERATIONAL_PPS1;
		else if (data & SIM_PTS_PPS0_BIT6)
			sim->state = SIM_STATE_OPERATIONAL_PPS2;
		else if (data & SIM_PTS_PPS0_BIT7)
			sim->state = SIM_STATE_OPERATIONAL_PPS3;
		else {
			sim->state = SIM_STATE_OPERATIONAL_PCK;
		}
		sim_param_t param_default = SIM_PARAM_DEFAULT;
		sim->param_pts = param_default;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_PPS1:
		sim_char_wait_disable(sim);
		sim->rcv_buffer[sim->rcv_count] = (uint8_t)data;
		sim->rcv_count++;
		sim_pts_received(sim, data);
		if (sim->sim_pts_pps3_flag && sim->sim_pts_pps2_flag)
			sim->state = SIM_STATE_OPERATIONAL_PPS2;
		else if (sim->sim_pts_pps2_flag == 1)
			sim->state = SIM_STATE_OPERATIONAL_PPS2;
		else if (sim->sim_pts_pps3_flag == 1)
			sim->state = SIM_STATE_OPERATIONAL_PPS3;
		else
			sim->state = SIM_STATE_OPERATIONAL_PCK;
		sim->sim_pts_pps2_flag = 0;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_PPS2:
		sim_char_wait_disable(sim);
		sim->rcv_buffer[sim->rcv_count] = (uint8_t)data;
		sim->rcv_count++;
		if (sim->sim_pts_pps3_flag == 1)
			sim->state = SIM_STATE_OPERATIONAL_PPS3;
		else
			sim->state = SIM_STATE_OPERATIONAL_PCK;
		sim->sim_pts_pps3_flag = 0;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_PPS3:
		sim_char_wait_disable(sim);
		sim->rcv_buffer[sim->rcv_count] = (uint8_t)data;
		sim->rcv_count++;
		sim->state = SIM_STATE_OPERATIONAL_PCK;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_OPERATIONAL_PCK:
		sim_char_wait_disable(sim);
		pr_debug("SIM_STATE_OPERATIONAL_PCK \n");
		sim->rcv_buffer[sim->rcv_count] = (uint8_t)data;
		sim->rcv_count++;
		pr_debug("rcv_count=%x xfer.rcv_lenght=%x \n", sim->rcv_count,
				sim->xfer.rcv_length);

		complete(&sim->xfer_done);
		if (sim->sim_pts_valid_flag == 1) {
			sim_set_param(sim, &sim->param_pts);
			sim->sim_pts_valid_flag = 0;
			sim->state = SIM_STATE_OPERATIONAL_IDLE;
			pr_info("COMM Parameters changed,"
				" PPS  Successful \n");
		} else {
			pr_info("deactivate occured due"
				" to incorrect PTS\n");
			sim_deactivate(sim);	/*  deactivate */
		}
		break;

		/* ATR FSM */

	case SIM_STATE_DETECTED_ATR_T0:
		reg_data0 = sim_read(sim->ioaddr + XMT_STATUS);
		sim_counter_disable(sim, reg_data0);
		sim_atr_add(sim, data);
		pr_debug("T0 %02x\n", data);
		sim->atrparser.T0 = (uint8_t)data;
		sim->state = SIM_STATE_DETECTED_ATR_TS;
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_DETECTED_ATR_TS:
		sim_char_wait_disable(sim);
		sim_atr_add(sim, data);
		pr_debug("TS %02x\n", data);
		sim->atrparser.TS = (uint8_t)data;
		if (data & 0xf0) {
			sim->atrparser.ifc_current_valid = (data >> 4) & 0x0f;
			sim->atrparser.num_hb = data & 0x0f;
			sim->atrparser.ifc_valid = 0;
			sim->state = SIM_STATE_DETECTED_ATR_TXI;
			sim->atrparser.cnt = 0;
			sim_char_wait_enable(sim);
		} else {
			goto sim_fsm_atr_thb;
		};
		break;

	case SIM_STATE_DETECTED_ATR_TXI:
		sim_char_wait_disable(sim);
		sim_atr_add(sim, data);
		i = ffs(sim->atrparser.ifc_current_valid) - 1;
		pr_debug("T%c%i %02x\n", 'A' + i, sim->atrparser.cnt + 1, data);
		sim->atrparser.TXI[SIM_IFC_TXI(i, sim->atrparser.cnt)] =
				(uint8_t)data;
		sim->atrparser.ifc_valid |= 1 << SIM_IFC_TXI(i,
							     sim->atrparser.
							     cnt);
		sim->atrparser.ifc_current_valid &= ~(1 << i);
		pr_debug("atrparser.cnt=%x\n", sim->atrparser.cnt);
		pr_debug("ifc_valid=%x\n", sim->atrparser.ifc_valid);
		if (sim->atrparser.ifc_current_valid == 0) {
			if (i == 3) {
				sim->atrparser.ifc_current_valid = (data >> 4)
				    & 0x0f;

				if (sim->atrparser.cnt > 3) {
					/* error */
					pr_debug("ERROR in atrparser.cnt !\n");
					break;
				} else
					sim->atrparser.cnt++;

				if (sim->atrparser.ifc_current_valid == 0)
					goto sim_fsm_atr_thb;
			} else {
sim_fsm_atr_thb:
				if (sim->atrparser.num_hb) {
					sim->state = SIM_STATE_DETECTED_ATR_THB;
					sim->atrparser.cnt = 0;
				} else {
					goto sim_fsm_atr_tck;
				};
			};
		};
		sim_char_wait_enable(sim);
		break;

	case SIM_STATE_DETECTED_ATR_THB:
		sim_char_wait_disable(sim);
		sim_atr_add(sim, data);
		pr_debug("THB%i %02x\n", sim->atrparser.cnt, data);
		sim->atrparser.THB[sim->atrparser.cnt] = (uint8_t)data;
		sim->atrparser.cnt++;
		sim_char_wait_enable(sim);

		if (sim->atrparser.cnt == sim->atrparser.num_hb) {
sim_fsm_atr_tck:
			for (i = 0; i <= 3; i++) {
				TDx_valid[i] =
					sim->atrparser.ifc_valid &
					(1 << (SIM_IFC_TXI(3, i)));

				if (TDx_valid[i]) {
					TDx_first4bits[i] =
					sim->atrparser.TXI[SIM_IFC_TXI(3, i)] &
						0x0f;
				}
			}

			for (i = 0; i <= 3; i++) {
				if (TDx_valid[i]) {

					if (TDx_first4bits[i] == 0)
						++cnt_type0;
					if (TDx_first4bits[i] == 15)
						++cnt_type15;
				}

			}

			for (i = 0; i <= 3; i++) {
				if (TDx_valid[i]) {
					if ((TDx_first4bits[i] > 0) &&
						(TDx_first4bits[i] < 15))
						tck_true = 1;
				}

			}
			for (i = 0; i <= 3; i++) {
				if (TDx_valid[i])
					temp = temp || TDx_first4bits[i];
			}

			if (((cnt_type0 && cnt_type15) ==
					SIM_PROTOCOL_T1) || tck_true)
				sim->state = SIM_STATE_DETECTED_ATR_TCK;
			else if (temp == 0)
				goto sim_fsm_atr_received;

		};
		break;

	case SIM_STATE_DETECTED_ATR_TCK:
		sim_char_wait_disable(sim);
		sim_atr_add(sim, data);
		/* checksum not required for T=0 */
		sim->atrparser.TCK = (uint8_t)data;
sim_fsm_atr_received:
		sim_char_wait_disable(sim);
		sim->state = SIM_STATE_OPERATIONAL_IDLE;
		sim->present = SIM_PRESENT_OPERATIONAL;
		sim_atr_received(sim);
		break;
	};
};

/* Function: sim_irq_handler
 *
 * Description: interrupt service routine.
 *
 * Parameters:
 * int irq                 interrupt number
 * void *dev_id            pointer to SIM device handler
 *
 * Return values:
 * IRQ_HANDLED             OS specific
 */

static irqreturn_t sim_irq_handler(int irq, void *dev_id)
{
	uint32_t reg_data = 0 , reg_data0 = 0, reg_data1 = 0;

	struct sim_t *sim = (struct sim_t *) dev_id;

	pr_debug("%s entering xmt_status=%x\n", __func__, sim_read(sim->ioaddr
				+ XMT_STATUS));

	reg_data0 = sim_read(sim->ioaddr + XMT_STATUS);

	if (reg_data0 & 0x100) {
		pr_info("Deactive Seq. Run.\nSIM does not meet timing"
			" requiremnet for the ATR at the start\n");
		sim_counter_disable(sim, reg_data0);
		sim_deactivate(sim);
	}

	reg_data0 = sim_read(sim->ioaddr + RCV_STATUS);

	if (reg_data0 & 0x100) {
		pr_info("Deactive Seq. Run.\nSIM does not meet timing"
			" requiremnet for char wait\n");
		sim_char_wait_disable(sim);
		sim_deactivate(sim);
	}

	if (reg_data0 & 0x400) {
		pr_info("Deactive Seq. Run.\nSIM does not meet timing"
			" requiremnet for block wait\n");
		sim_block_wait_disable(sim);
		sim_deactivate(sim);
	}

	reg_data0 = sim_read(sim->ioaddr + XMT_STATUS);
	reg_data1 = sim_read(sim->ioaddr + INT_MASK);
	if ((reg_data0 & SIM_XMT_STATUS_TC)
	    && (!(reg_data1 & SIM_INT_MASK_TCIM))) {
		pr_debug("TC_IRQ\n");
		sim_write(SIM_XMT_STATUS_TC, sim->ioaddr + XMT_STATUS);
		reg_data = sim_read(sim->ioaddr + INT_MASK);
		reg_data |= SIM_INT_MASK_TCIM;
		sim_write(reg_data, sim->ioaddr + INT_MASK);
		reg_data = sim_read(sim->ioaddr + ENABLE);
		reg_data &= ~SIM_ENABLE_XMTEN;
		sim_write(reg_data, sim->ioaddr + ENABLE);
	};

	reg_data0 = sim_read(sim->ioaddr + XMT_STATUS);
	reg_data1 = sim_read(sim->ioaddr + INT_MASK);
	if ((reg_data0 & SIM_XMT_STATUS_TDTF)
	    && (!(reg_data1 & SIM_INT_MASK_TDTFM))) {
		pr_debug("TDTF_IRQ\n");
		sim_write(SIM_XMT_STATUS_TDTF, sim->ioaddr + XMT_STATUS);
		sim_xmt_fill(sim);

		if (sim->xmt_remaining == 0) {
			sim_write(SIM_XMT_STATUS_TC,
				     sim->ioaddr + XMT_STATUS);
			reg_data = sim_read(sim->ioaddr + INT_MASK);
			reg_data &= ~SIM_INT_MASK_TCIM;
			reg_data |= SIM_INT_MASK_TDTFM;
			sim_write(reg_data, sim->ioaddr + INT_MASK);
		};
	};

	reg_data0 = sim_read(sim->ioaddr + RCV_STATUS);
	reg_data1 = sim_read(sim->ioaddr + INT_MASK);
	if ((reg_data0 & SIM_RCV_STATUS_RDRF)
	    && (!(reg_data1 & SIM_INT_MASK_RIM))) {
		pr_debug("%s RDRF_IRQ\n", __func__);
		sim_write(SIM_RCV_STATUS_RDRF, sim->ioaddr + RCV_STATUS);

		while (sim_read(sim->ioaddr + RCV_FIFO_CNT)) {
			uint32_t data;
			data = sim_read(sim->ioaddr + PORT1_RCV_BUF);
			pr_debug("RX = %02x state = %i\n", data, sim->state);
			if (data & 0x700) {
				if (sim->xfer_ongoing) {
					/* error */
					sim_char_wait_disable(sim);
					pr_info("ERROR ! in CWT/FE/PE\n");
					pr_info("CWT/FE/PE 0x%x\n",
						(data & 0x0700));
					return IRQ_HANDLED;
				};
			} else {
				sim_fsm(sim, data);
				if (sim->sim_xmt_last_byte == 1)
					sim_block_wait_disable(sim);
			}
		};
	};

	reg_data0 = sim_read(sim->ioaddr + PORT0_DETECT);
	if (reg_data0 & SIM_PORT_DETECT_SDI) {
		pr_debug("%s PD_IRQ\n", __func__);
		reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
		reg_data |= SIM_PORT_DETECT_SDI;
		sim_write(reg_data, sim->ioaddr + PORT0_DETECT);

		reg_data0 = sim_read(sim->ioaddr + PORT0_DETECT);
		if (reg_data0 & SIM_PORT_DETECT_SPDP) {
			reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
			reg_data &= ~SIM_PORT_DETECT_SPDS;
			sim_write(reg_data, sim->ioaddr + PORT0_DETECT);

			if (sim->present != SIM_PRESENT_REMOVED) {
				pr_debug("Removed sim card\n");
				sim->present = SIM_PRESENT_REMOVED;
				sim->state = SIM_STATE_REMOVED;

				if (sim->fasync)
					kill_fasync(&sim->fasync,
						    SIGIO, POLL_IN);
			};
		} else {
			reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
			reg_data |= SIM_PORT_DETECT_SPDS;
			sim_write(reg_data, sim->ioaddr + PORT0_DETECT);

			if (sim->present == SIM_PRESENT_REMOVED) {
				pr_debug("Inserted sim card\n");
				sim->state = SIM_STATE_DETECTED_ATR_T0;
				sim->present = SIM_PRESENT_DETECTED;

				if (sim->fasync)
					kill_fasync(&sim->fasync,
						    SIGIO, POLL_IN);
				sim_counter_enable(sim);
			};
		};
	};

	return IRQ_HANDLED;
};

/* Function: sim_power_on
 *
 * Description: run the power on sequence
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_power_on(struct sim_t *sim)
{
	uint32_t reg_data;

	/* power on sequence */
	pr_debug("%s Powering on the sim port.\n", __func__);
	reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
	reg_data |= SIM_PORT_CNTL_SVEN;
	sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
	msleep(10);
	reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
	reg_data |= SIM_PORT_CNTL_SCEN;
	sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
	msleep(10);
	reg_data = SIM_RCV_THRESHOLD_RTH(0) | SIM_RCV_THRESHOLD_RDT(1);
	sim_write(reg_data, sim->ioaddr + RCV_THRESHOLD);
	sim_write(SIM_RCV_STATUS_RDRF, sim->ioaddr + RCV_STATUS);
	reg_data = sim_read(sim->ioaddr + INT_MASK);
	reg_data &= ~(SIM_INT_MASK_RIM | SIM_INT_MASK_GPCM);
	sim_write(reg_data, sim->ioaddr + INT_MASK);
	sim_write(31, sim->ioaddr + DIVISOR);
	reg_data = sim_read(sim->ioaddr + CNTL);
	reg_data |= SIM_CNTL_SAMPLE12;
	sim_write(reg_data, sim->ioaddr + CNTL);
	reg_data = sim_read(sim->ioaddr + ENABLE);
	reg_data |= SIM_ENABLE_RCVEN;
	sim_write(reg_data, sim->ioaddr + ENABLE);
	reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
	reg_data |= SIM_PORT_CNTL_SRST;
	sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
	pr_debug("%s port0_ctl is 0x%x.\n", __func__,
		 sim_read(sim->ioaddr + PORT0_CNTL));
	pr_debug("%s cntl is 0x%x.\n", __func__,
		 sim_read(sim->ioaddr + CNTL));
	sim->power = SIM_POWER_ON;
	sim->sim_pts_pps2_flag = 0;
	sim->sim_pts_pps3_flag = 0;
	sim->sim_pts_valid_flag = 0;
};

/* Function: sim_power_off
 *
 * Description: run the power off sequence
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_power_off(struct sim_t *sim)
{
	uint32_t reg_data;

	pr_debug("%s entering.\n", __func__);
	/* sim_power_off sequence */
	reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
	reg_data &= ~SIM_PORT_CNTL_SCEN;
	sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
	reg_data = sim_read(sim->ioaddr + ENABLE);
	reg_data &= ~SIM_ENABLE_RCVEN;
	sim_write(reg_data, sim->ioaddr + ENABLE);
	reg_data = sim_read(sim->ioaddr + INT_MASK);
	reg_data |= SIM_INT_MASK_RIM;
	sim_write(reg_data, sim->ioaddr + INT_MASK);
	sim_write(0, sim->ioaddr + RCV_THRESHOLD);
	reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
	reg_data &= ~SIM_PORT_CNTL_SRST;
	sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
	reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
	reg_data &= ~SIM_PORT_CNTL_SVEN;
	sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
	sim->power = SIM_POWER_OFF;
};

/* Function: sim_start
 *
 * Description: ramp up the SIM interface
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_start(struct sim_t *sim)
{
	uint32_t reg_data, clk_div = 0;
	memset(&sim->atrparser, 0, sizeof(sim->atrparser));
	memset(&sim->atr, 0, sizeof(sim->atr));

	pr_debug("%s entering.\n", __func__);
	/* Configuring SIM for Operation */
	reg_data = SIM_XMT_THRESHOLD_XTH(0) | SIM_XMT_THRESHOLD_TDT(5);
	sim_write(reg_data, sim->ioaddr + XMT_THRESHOLD);
	sim_write(0, sim->ioaddr + SETUP);
	/* ~ 4 MHz */
#ifdef CONFIG_OF
	clk_div = sim->clk.prescaler;
/* Following code is for ARM platform
#else
	clk_rate = clk_get_rate(sim->clk);
	clk_div = clk_rate / sim->plat_data->clk_rate;
	if (clk_rate % sim->plat_data->clk_rate)
		clk_div++;*/
#endif
	pr_debug("%s prescaler is 0x%x.\n", __func__, clk_div);
	sim_write(clk_div, sim->ioaddr + CLK_PRESCALER);

	reg_data = SIM_CNTL_GPCNT_CLK_SEL(0) | SIM_CNTL_BAUD_SEL(7)
	    | SIM_CNTL_SAMPLE12 | SIM_CNTL_ANACK | SIM_CNTL_ICM;
	sim_write(reg_data, sim->ioaddr + CNTL);
	sim_write(31, sim->ioaddr + DIVISOR);
	reg_data = sim_read(sim->ioaddr + OD_CONFIG);
	reg_data |= SIM_OD_CONFIG_OD_P0;
	sim_write(reg_data, sim->ioaddr + OD_CONFIG);
	reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
	reg_data |= SIM_PORT_CNTL_3VOLT | SIM_PORT_CNTL_STEN;
	sim_write(reg_data, sim->ioaddr + PORT0_CNTL);

	/* presense detect */
	pr_debug("%s p0_det is 0x%x\n", __func__,
		 sim_read(sim->ioaddr + PORT0_DETECT));
	if (sim_read(sim->ioaddr + PORT0_DETECT) & SIM_PORT_DETECT_SPDP) {
		pr_debug("%s card removed\n", __func__);
		reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
		reg_data &= ~SIM_PORT_DETECT_SPDS;
		sim_write(reg_data, sim->ioaddr + PORT0_DETECT);
		sim->present = SIM_PRESENT_REMOVED;
		sim->state = SIM_STATE_REMOVED;
	} else {
		pr_debug("%s card inserted\n", __func__);
		reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
		reg_data |= SIM_PORT_DETECT_SPDS;
		sim_write(reg_data, sim->ioaddr + PORT0_DETECT);
		sim->present = SIM_PRESENT_DETECTED;
		sim->state = SIM_STATE_DETECTED_ATR_T0;
	};
	reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
	reg_data |= SIM_PORT_DETECT_SDI;
	reg_data &= ~SIM_PORT_DETECT_SDIM;
	sim_write(reg_data, sim->ioaddr + PORT0_DETECT);

	pr_debug("%s After p0_det is 0x%x\n", __func__,
		 sim_read(sim->ioaddr + PORT0_DETECT));

	if (sim->present == SIM_PRESENT_DETECTED)
		sim_power_on(sim);

};

/* Function: sim_stop
 *
 * Description: shut down the SIM interface
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_stop(struct sim_t *sim)
{
	pr_debug("%s entering.\n", __func__);
	sim_write(0, sim->ioaddr + SETUP);
	sim_write(0, sim->ioaddr + ENABLE);
	sim_write(0, sim->ioaddr + PORT0_CNTL);
	sim_write(0x06, sim->ioaddr + CNTL);
	sim_write(0, sim->ioaddr + CLK_PRESCALER);
	sim_write(0, sim->ioaddr + SETUP);
	sim_write(0, sim->ioaddr + OD_CONFIG);
	sim_write(0, sim->ioaddr + XMT_THRESHOLD);
	sim_write(0xb8, sim->ioaddr + XMT_STATUS);
	sim_write(4, sim->ioaddr + RESET_CNTL);
	mdelay(1);
};

/* Function: sim_data_reset
 *
 * Description: reset a SIM structure to default values
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_data_reset(struct sim_t *sim)
{
	sim_param_t param_default = SIM_PARAM_DEFAULT;
	sim->present = SIM_PRESENT_REMOVED;
	sim->state = SIM_STATE_REMOVED;
	sim->power = SIM_POWER_OFF;
	sim->errval = SIM_OK;
	memset(&sim->atrparser, 0, sizeof(sim->atrparser));
	memset(&sim->atr, 0, sizeof(sim->atr));
	sim->param_atr = param_default;
	memset(&sim->param, 0, sizeof(sim->param));
	memset(&sim->param_pts, 0, sizeof(sim->param_pts));
	memset(&sim->xfer, 0, sizeof(sim->xfer));
	sim->xfer_ongoing = 0;
	sim->xmt_remaining = 0;
	sim->xmt_pos = 0;
	sim->rcv_count = 0;
	memset(sim->rcv_buffer, 0, SIM_RCV_BUFFER_SIZE);
	memset(sim->xmt_buffer, 0, SIM_XMT_BUFFER_SIZE);
	sim->sim_pts_pps2_flag = 0;
	sim->sim_pts_pps3_flag = 0;
	sim->sim_pts_valid_flag = 0;
	sim->ins_complimentary_flag = 0;
	sim->sim_xmt_last_byte = 0;
};

/* Function: sim_cold_reset
 *
 * Description: cold reset the SIM interface, including card
 * power down and interface hardware reset.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_cold_reset(struct sim_t *sim)
{
	pr_debug("%s entering.\n", __func__);
	if (sim->present != SIM_PRESENT_REMOVED) {
		sim_power_off(sim);
		sim_stop(sim);
		sim_data_reset(sim);
		sim->state = SIM_STATE_DETECTED_ATR_T0;
		sim->present = SIM_PRESENT_DETECTED;
		msleep(50);
		sim_start(sim);
		sim_power_on(sim);
	};
};

/* Function: sim_warm_reset
 *
 * Description: warm reset the SIM interface: just invoke the
 * reset signal and reset the SIM structure for the interface.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_warm_reset(struct sim_t *sim)
{
	uint32_t reg_data;

	pr_debug("%s entering.\n", __func__);
	if (sim->present != SIM_PRESENT_REMOVED) {
		reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
		reg_data |= SIM_PORT_CNTL_SRST;
		sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
		sim_data_reset(sim);
		msleep(50);
		reg_data = sim_read(sim->ioaddr + PORT0_CNTL);
		reg_data &= ~SIM_PORT_CNTL_SRST;
		sim_write(reg_data, sim->ioaddr + PORT0_CNTL);
	};
};


/* Function: sim_deactivate
 *
 * Description: cold reset the SIM interface, including card
 * power down and interface hardware reset.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_deactivate(struct sim_t *sim)
{
	int reg_data = 0;

	pr_debug("%s entering.\n", __func__);
	if (sim->present != SIM_PRESENT_REMOVED) {
		sim_power_off(sim);
		sim_stop(sim);
		sim_data_reset(sim);
		sim->state = SIM_STATE_OPERATIONAL_IDLE;
		sim->present = SIM_PRESENT_REMOVED;
		reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
		sim_write(reg_data | SIM_PORT_DETECT_SDIM,
			sim->ioaddr + PORT0_DETECT);
		if (sim->fasync)
			kill_fasync(&sim->fasync, SIGIO, POLL_IN);
		msleep(50);
	};
};

/* Function: sim_counter_enable
 *
 * Description: set GPCNT counter to value 0x9c40
 * Enable GPCNT clock select as card clock
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_counter_enable(struct sim_t *sim)
{
	uint32_t reg_data = 0;

	pr_info("Entering.. %s\n", __func__);

	sim_write(0x9c40, sim->ioaddr + GPCNT);
	reg_data = sim_read(sim->ioaddr + CNTL);
	reg_data |= SIM_CNTL_GPCNT_CLK_SEL(0x01);
	sim_write(reg_data, sim->ioaddr + CNTL);

	pr_debug("XMT_STATUS= %x \n", sim_read(
		sim->ioaddr + XMT_STATUS));
	pr_debug("INT_MASK= %x b4 insert\n", sim_read(
		sim->ioaddr + INT_MASK));
	pr_debug("%s GPCNT=%x CNTL=%x\n", __func__,
		sim_read(sim->ioaddr + GPCNT),
		sim_read(sim->ioaddr + CNTL));

	return;
}

/* Function: sim_counter_disable
 *
 * Description: Disable GPCNTM in INT_MASK, Disbale GPCNT_CLK_SEL
 * Clear GPCNT bit in XMT_MASK
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 * XMT_STATUS in reg_data0
 */

static void sim_counter_disable(struct sim_t *sim, uint32_t reg_data0)
{
	uint32_t reg_data = 0, reg_data1 = 0;

	reg_data1 = sim_read(sim->ioaddr + INT_MASK);
	reg_data1 |= SIM_INT_MASK_GPCM;
	sim_write(reg_data1, sim->ioaddr + INT_MASK);
	reg_data = sim_read(sim->ioaddr + CNTL);
	reg_data &= ~SIM_CNTL_GPCNT_CLK_SEL_MASK;
	sim_write(reg_data, sim->ioaddr + CNTL);
	reg_data0 = reg_data0 | SIM_XMT_STATUS_GPCNT;
	sim_write(reg_data0, sim->ioaddr + XMT_STATUS);

	return;
}

/* Function: sim_char_wait_disable
 *
 * Description: Disable CHAR WAIT in INT_MASK, Disbale CWTEN in CNTL
 * Clear CWT bit in RCV_STATUS
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 * XMT_STATUS in reg_data0
 */

static void sim_char_wait_disable(struct sim_t *sim)
{
	uint32_t reg_data = 0, reg_data1 = 0, rcv_status = 0;

	reg_data1 = sim_read(sim->ioaddr + INT_MASK);
	reg_data1 |= SIM_INT_MASK_CWTM;
	sim_write(reg_data1, sim->ioaddr + INT_MASK);
	reg_data = sim_read(sim->ioaddr + CNTL);
	reg_data &= ~SIM_CNTL_CWTEN;
	sim_write(reg_data, sim->ioaddr + CNTL);
	rcv_status = sim_read(sim->ioaddr + RCV_STATUS);
	pr_debug("In.............. %s state= %x \n", __func__, sim->state);
	pr_debug("rcv_status in disable 0x%x\n", (rcv_status & 0x0f00));
	rcv_status = rcv_status | SIM_RCV_STATUS_CWT;
	sim_write(rcv_status, sim->ioaddr + RCV_STATUS);

	return;
}
/* Function: sim_char_wait_enable
 *
 * Description: set CHAR WAIT counter to value 0x2580
 * Enable CWTEN in CNTL and enable MASK CWTM mask in INT_MASK
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_char_wait_enable(struct sim_t *sim)
{
	uint32_t reg_data = 0, CWI = 10, CWT_ETU = 0;
	int DI = sim_param_D[sim->param_pts.DI] / 64;

	if ((DI == 0) || (sim->xfer.type == SIM_XFER_TYPE_PTS))
		DI = 1;

	if (DI > 0) {
		pr_debug("Entering.. %s state=%x\n", __func__, sim->state);

		if ((sim->state >= 7) && (sim->state <= 11))
			CWT_ETU = (CWI * 960 * DI);
		else
			CWT_ETU = (sim->param_atr.WWT * 960 * DI);

		reg_data = sim_read(sim->ioaddr + INT_MASK);
		reg_data &= ~SIM_INT_MASK_CWTM;
		sim_write(reg_data, sim->ioaddr + INT_MASK);
		sim_write(CWT_ETU, sim->ioaddr + CHAR_WAIT);
		reg_data = sim_read(sim->ioaddr + CNTL);
		reg_data |= SIM_CNTL_CWTEN;
		sim_write(reg_data, sim->ioaddr + CNTL);

		pr_debug("In.. %s state= %x \n", __func__, sim->state);
		pr_debug("RCV_STATUS= %x \n", sim_read(
			sim->ioaddr + RCV_STATUS));
		pr_debug("INT_MASK= %x after \n", sim_read(
			sim->ioaddr + INT_MASK));
		pr_debug("%s CHAR_WAIT=%x CNTL=%x\n", __func__,
			sim_read(sim->ioaddr + CHAR_WAIT),
			sim_read(sim->ioaddr + CNTL));
	} else
		printk(KERN_INFO"DI value is negative / SIM_RFU\n");

	return;
}

/* Function: sim_block_wait_disable
 *
 * Description: Disable BLOCK WAIT BWT in INT_MASK, Disbale BWTEN in CNTL
 * Clear BWT bit in RCV_STATUS
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 * XMT_STATUS in reg_data0
 */

static void sim_block_wait_disable(struct sim_t *sim)
{
	uint32_t reg_data = 0, reg_data1 = 0, rcv_status = 0;

	sim->sim_xmt_last_byte = 0;
	reg_data1 = sim_read(sim->ioaddr + INT_MASK);
	reg_data1 |= SIM_INT_MASK_BWTM;
	sim_write(reg_data1, sim->ioaddr + INT_MASK);
	reg_data = sim_read(sim->ioaddr + CNTL);
	reg_data &= ~SIM_CNTL_BWTEN;
	sim_write(reg_data, sim->ioaddr + CNTL);
	rcv_status = sim_read(sim->ioaddr + RCV_STATUS);
	pr_debug("In.............. %s state= %x \n", __func__, sim->state);
	pr_debug("rcv_status in disable 0x%x\n", (rcv_status & 0x0f00));
	rcv_status = rcv_status | SIM_RCV_STATUS_BWT;
	sim_write(rcv_status, sim->ioaddr + RCV_STATUS);

	return;
}
/* Function: sim_block_wait_enable
 *
 * Description: set BLOCK WAIT counter to value 0x2580 to TC2*960
 * Enable BWTEN in CNTL and enable MASK BWTM mask in INT_MASK
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static void sim_block_wait_enable(struct sim_t *sim)
{
	uint32_t reg_data = 0, WI = 10, BWT_ETU = 0;
	uint16_t BWT_ETU_L = 0, BWT_ETU_H = 0;
	int DI = sim_param_D[sim->param_pts.DI] / 64;

	if ((DI == 0) || (sim->xfer.type == SIM_XFER_TYPE_PTS))
		DI = 1;

	if (DI > 0) {

		if (sim->param_atr.WWT > 10)
			BWT_ETU = (sim->param_atr.WWT * 960 * DI);
		else
			BWT_ETU = (WI * 960 * DI);

		if (BWT_ETU > 65535) {
			BWT_ETU_L = BWT_ETU & 0xffff;
			BWT_ETU_H = ((BWT_ETU >> 16) & 0xffff);
		} else
			BWT_ETU_L = (BWT_ETU & 0xffff);

		pr_debug("Entering.. %s\n", __func__);

		reg_data = sim_read(sim->ioaddr + INT_MASK);
		reg_data &= ~SIM_INT_MASK_BWTM;
		sim_write(reg_data, sim->ioaddr + INT_MASK);
		sim_write(BWT_ETU_L, sim->ioaddr + BWT);
		sim_write(BWT_ETU_H, sim->ioaddr + BWT_H);
		reg_data = sim_read(sim->ioaddr + CNTL);
		reg_data |= SIM_CNTL_BWTEN;
		sim_write(reg_data, sim->ioaddr + CNTL);

		pr_debug("In.. %s state= %x \n", __func__, sim->state);
		pr_debug("RCV_STATUS= %x \n", sim_read(
			sim->ioaddr + RCV_STATUS));
		pr_debug("INT_MASK= %x after \n", sim_read(
			sim->ioaddr + INT_MASK));
		pr_debug("%s BWT_H=%x BWT=%x CNTL=%x\n", __func__,
			sim_read(sim->ioaddr + BWT_H),
			sim_read(sim->ioaddr + BWT),
			sim_read(sim->ioaddr + CNTL));
	} else
		printk(KERN_INFO"DI value is negative /SIM_RFU\n");

	return;
}

/* Function: sim_card_lock
 *
 * Description: physically lock the SIM card.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static int sim_card_lock(struct sim_t *sim)
{
	int errval;

	pr_debug("%s entering.\n", __func__);
	/* place holder for true physcial locking */
	if (sim->present != SIM_PRESENT_REMOVED)
		errval = SIM_OK;
	else
		errval = -SIM_E_NOCARD;
	return errval;
};

/* Function: sim_card_eject
 *
 * Description: physically unlock and eject the SIM card.
 *
 * Parameters:
 * struct sim_t* sim              pointer to SIM device handler
 */

static int sim_card_eject(struct sim_t *sim)
{
	int errval;

	pr_debug("%s entering.\n", __func__);
	/* place holder for true physcial locking */
	if (sim->present != SIM_PRESENT_REMOVED)
		errval = SIM_OK;
	else
		errval = -SIM_E_NOCARD;
	return errval;
};

/* Function: sim_ioctl
 *
 * Description: handle ioctl calls
 *
 * Parameters: OS specific
 */

static int sim_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
	int ret, errval = SIM_OK;
	unsigned long timeout;
	struct sim_t *sim = (struct sim_t *) file->private_data;

	pr_debug("%s entering.\n", __func__);
	switch (cmd) {
	default:
		pr_debug("ioctl cmd %d is issued...\n", cmd);

	case SIM_IOCTL_GET_ATR:
		if (sim->present != SIM_PRESENT_OPERATIONAL) {
			errval = -SIM_E_NOCARD;
			break;
		};
		ret = copy_to_user((sim_atr_t *) arg, &sim->atr,
				   sizeof(sim_atr_t));
		if (ret)
			errval = -SIM_E_ACCESS;
		break;

	case SIM_IOCTL_GET_PARAM_ATR:
		if (sim->present != SIM_PRESENT_OPERATIONAL) {
			errval = -SIM_E_NOCARD;
			break;
		};

		ret = copy_to_user((sim_param_t *) arg, &sim->param_atr,
				   sizeof(sim_param_t));
		if (ret)
			errval = -SIM_E_ACCESS;
		break;

	case SIM_IOCTL_GET_PARAM:
		ret = copy_to_user((sim_param_t *) arg, &sim->param,
				   sizeof(sim_param_t));
		if (ret)
			errval = -SIM_E_ACCESS;
		break;

	case SIM_IOCTL_SET_PARAM:
		ret = copy_from_user(&sim->param, (sim_param_t *) arg,
				     sizeof(sim_param_t));
		if (ret)
			errval = -SIM_E_ACCESS;
		else
			errval = sim_set_param(sim, &sim->param);
		break;

	case SIM_IOCTL_POWER_ON:
		if (sim->power == SIM_POWER_ON) {
			errval = -SIM_E_POWERED_ON;
			break;
		};
		sim_power_on(sim);
		break;

	case SIM_IOCTL_POWER_OFF:
		if (sim->power == SIM_POWER_OFF) {
			errval = -SIM_E_POWERED_OFF;
			break;
		};
		sim_power_off(sim);
		break;

	case SIM_IOCTL_COLD_RESET:
		if (sim->power == SIM_POWER_OFF) {
			errval = -SIM_E_POWERED_OFF;
			break;
		};
		sim_cold_reset(sim);
		break;

	case SIM_IOCTL_WARM_RESET:
		sim_warm_reset(sim);
		if (sim->power == SIM_POWER_OFF) {
			errval = -SIM_E_POWERED_OFF;
			break;
		};
		break;

	case SIM_IOCTL_XFER:
		if (sim->present != SIM_PRESENT_OPERATIONAL) {
			errval = -SIM_E_NOCARD;
			break;
		};

		ret = copy_from_user(&sim->xfer, (sim_xfer_t *) arg,
				     sizeof(sim_xfer_t));
		if (ret) {
			errval = -SIM_E_ACCESS;
			break;
		};

		ret = copy_from_user(sim->xmt_buffer, sim->xfer.xmt_buffer,
				     sim->xfer.xmt_length);
		if (ret) {
			errval = -SIM_E_ACCESS;
			break;
		};

		sim->rcv_count = 0;
		sim->xfer.sw1 = 0;
		sim->xfer.sw2 = 0;

		if (sim->xfer.type == SIM_XFER_TYPE_TPDU) {
			if (sim->xfer.xmt_length < 5) {
				errval = -SIM_E_TPDUSHORT;
				break;
			}
			sim->state = SIM_STATE_OPERATIONAL_COMMAND;
		} else if (sim->xfer.type == SIM_XFER_TYPE_PTS) {
			if (sim->xfer.xmt_length == 0) {
				errval = -SIM_E_PTSEMPTY;
				break;
			}
			sim->state = SIM_STATE_OPERATIONAL_PPSS;
		} else {
			errval = -SIM_E_INVALIDXFERTYPE;
			break;
		};

		if (sim->xfer.xmt_length > SIM_XMT_BUFFER_SIZE) {
			errval = -SIM_E_INVALIDXMTLENGTH;
			break;
		};

		if (sim->xfer.rcv_length > SIM_XMT_BUFFER_SIZE) {
			errval = -SIM_E_INVALIDRCVLENGTH;
			break;
		};

		sim->errval = 0;
		sim->xfer_ongoing = 1;
		init_completion(&sim->xfer_done);
		if (sim->xfer.type == SIM_XFER_TYPE_PTS)
			sim_xmt_start(sim, 0, sim->xfer.xmt_length);
		else
			sim_xmt_start(sim, 0, 5);
		sim->xfer.xmt_position = 5;
		timeout =
		    wait_for_completion_interruptible_timeout(&sim->xfer_done,
							      sim->xfer.
							      timeout);
		sim->xfer_ongoing = 0;

		if (sim->errval) {
			errval = sim->errval;
			break;
		};

		if (timeout == 0) {
			if (sim->xfer.type == SIM_XFER_TYPE_PTS) {
				errval = -SIM_E_TIMEOUT;
				pr_info("timeout breaked and deactivated\n");
				sim_deactivate(sim);
			} else {
				errval = -SIM_E_TIMEOUT;
				pr_debug("timeout break is back\n");
				break;
			}
		}

		ret = copy_to_user(sim->xfer.rcv_buffer, sim->rcv_buffer,
				   sim->xfer.rcv_length);
		if (ret) {
			errval = -SIM_E_ACCESS;
			break;
		};

		ret = copy_to_user((sim_xfer_t *) arg, &sim->xfer,
				   sizeof(sim_xfer_t));
		if (ret)
			errval = -SIM_E_ACCESS;
		break;

	case SIM_IOCTL_GET_PRESENSE:
		if (put_user(sim->present, (int *)arg))
			errval = -SIM_E_ACCESS;
		break;

	case SIM_IOCTL_CARD_LOCK:
		errval = sim_card_lock(sim);
		break;

	case SIM_IOCTL_CARD_EJECT:
		errval = sim_card_eject(sim);
		break;

	};

	return errval;
};

/* Function: sim_fasync
 *
 * Description: async handler
 *
 * Parameters: OS specific
 */

static int sim_fasync(int fd, struct file *file, int mode)
{
	struct sim_t *sim = (struct sim_t *) file->private_data;
	pr_debug("%s entering.\n", __func__);
	return fasync_helper(fd, file, mode, &sim->fasync);
}

/* Function: sim_open
 *
 * Description: ramp up interface when being opened
 *
 * Parameters: OS specific
 */

static int sim_open(struct inode *inode, struct file *file)
{
	int errval = SIM_OK;

	struct sim_t *sim = dev_get_drvdata(sim_dev.parent);
	file->private_data = sim;

	pr_debug("%s entering.\n", __func__);
	if (!sim->ioaddr) {
		errval = -ENOMEM;
		return errval;
	}

	if (!(sim->clk_flag)) {
		pr_debug("\n%s enable the clock\n", __func__);
		sim->sim_clk_ops.clk_enable(sim);
		sim->clk_flag = 1;
	}

	sim_start(sim);

	return errval;
};

/* Function: sim_release
 *
 * Description: shut down interface when being closed
 *
 * Parameters: OS specific
 */

static int sim_release(struct inode *inode, struct file *file)
{
	uint32_t reg_data;

	struct sim_t *sim = (struct sim_t *) file->private_data;

	pr_debug("%s entering.\n", __func__);
	if (sim->clk_flag) {
		pr_debug("\n%s disable the clock\n", __func__);
		sim->sim_clk_ops.clk_disable(sim);
		sim->clk_flag = 0;
	}

	/* disable presense detection */
	reg_data = sim_read(sim->ioaddr + PORT0_DETECT);
	sim_write(reg_data | SIM_PORT_DETECT_SDIM,
		     sim->ioaddr + PORT0_DETECT);

	if (sim->present != SIM_PRESENT_REMOVED) {
		sim_power_off(sim);
		if (sim->fasync)
			kill_fasync(&sim->fasync, SIGIO, POLL_IN);
	};

	sim_stop(sim);

	sim_fasync(-1, file, 0);

	pr_debug("exit\n");
	return 0;
};

static const struct file_operations sim_fops = {
	.open = sim_open,
	.ioctl = sim_ioctl,
	.fasync = sim_fasync,
	.release = sim_release
};

static struct miscdevice sim_dev = {
	MISC_DYNAMIC_MINOR,
	"fsl_sim",
	&sim_fops
};

int sim_device_init(struct of_device *ofdev,
		struct sim_t *sim)
{
	int ret = 0;

	if (sim->ipb_irq)
		ret = request_irq(sim->ipb_irq, sim_irq_handler, 0,
				"fsl_sim_ipb", sim);
	if (ret) {
		dev_err(&ofdev->dev, "%s: Could not register interrupt\n",
				__func__);
		goto out;
	};

	platform_set_drvdata(ofdev, sim);
	sim_dev.parent = &(ofdev->dev);

	misc_register(&sim_dev);

	return ret;

out:
	if (sim->ipb_irq)
		free_irq(sim->ipb_irq, sim);

	return ret;
}
