/*
 * Copyright (c) 2011-2012 Freescale Semiconductor, Inc.
 *
 * Freescale Subscriber Identification Module(SIM) device driver
 *
 * Author: Harninder Rai <harninder.rai@freescale.com>
 * Author: Ashish Kumar  <ashish.kumar@freescale.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/* Transmit and receive buffer sizes */
#define SIM_XMT_BUFFER_SIZE 256
#define SIM_RCV_BUFFER_SIZE 256

#ifdef __BIG_ENDIAN
#define sim_write(x, y)		out_be32(y, x)
#define sim_read(x)		in_be32(x)
#else
/* Following #defines are for ARM platform
#define sim_write(x, y)		__raw_writel(x, y)
#define sim_read(x)		__raw_readl(x) */
#endif

#define PORT0_CONTROL		0x0014
#define SIM_CARD_CLK_ENABLE	(1<<4)

/* ATR parser data (the parser state is stored in the main device structure) */
typedef struct {
	uint8_t T0;		/* ATR T0 */
	uint8_t TS;		/* ATR TS */
	/* ATR TA1, TB1, TC1, TD1, TB1, ... , TD4 */
	uint8_t TXI[16];
	uint8_t THB[15];	/* ATR historical bytes */
	uint8_t TCK;		/* ATR checksum */
	uint16_t ifc_valid;	/* valid interface characters */
	uint8_t ifc_current_valid;	/* calid ifcs in the current batch */
	uint8_t cnt;		/* number of current batch */
	uint8_t num_hb;		/* number of historical bytes */
} sim_atrparser_t;

#ifdef CONFIG_OF
struct clk {
	u32 rate;
	u8 prescaler;
};
#endif

struct sim_t;

/* SIM clock operations structure */
struct sim_clk_operations {
	int (*clk_enable) (struct sim_t *sim);
	void (*clk_disable) (struct sim_t *sim);
	struct clk * (*clk_get) (struct device *dev, const char *id);
	void (*clk_put) (struct sim_t *sim);
};

/* Main SIM driver structure */
struct sim_t{
	/* card inserted = 1, ATR received = 2, card removed = 0 */
	int present;
	/* current ATR or OPS state */
	int state;
	/* current power state */
	int power;
	/* error code occured during transfer */
	int errval;
	struct clk clk;	/* Clock id */
	uint8_t clk_flag;
	struct resource *res;	/* IO map memory */
	void __iomem *ioaddr;	/* Mapped address */
	int ipb_irq;		/* sim ipb IRQ num */
	int dat_irq;		/* sim dat IRQ num */
	/* parser for incoming ATR stream */
	sim_atrparser_t atrparser;
	/* raw ATR stream received */
	sim_atr_t atr;
	/* communication parameters according to ATR */
	sim_param_t param_atr;
	/* communication parameters after pps/pts*/
	sim_param_t param_pts;
	/* current communication parameters */
	sim_param_t param;
	/* current TPDU or PTS transfer */
	sim_xfer_t xfer;
	/* transfer is on the way = 1, idle = 2 */
	int xfer_ongoing;
	/* remaining bytes to transmit for the current transfer */
	int xmt_remaining;
	/* transmit position */
	int xmt_pos;
	/* receive position / number of bytes received */
	int rcv_count;
	uint8_t rcv_buffer[SIM_RCV_BUFFER_SIZE];
	uint8_t xmt_buffer[SIM_XMT_BUFFER_SIZE];
	/* transfer completion notifier */
	struct completion xfer_done;
	/* async notifier for card and ATR detection */
	struct fasync_struct *fasync;
	/* Platform specific data */
	struct fsl_sim_platform_data *plat_data;
	/* Don't know what to write */
	struct sim_clk_operations sim_clk_ops;
	uint8_t sim_pts_pps2_flag;
	uint8_t sim_pts_pps3_flag;
	uint8_t sim_pts_valid_flag;
	uint8_t ins_complimentary_flag;
	uint8_t sim_xmt_last_byte;
};

static struct miscdevice sim_dev;

int sim_device_init(struct of_device *ofdev,
		struct sim_t *sim);

static inline int sim_clk_enable(struct sim_t *sim)
{
	uint32_t reg_data;

	reg_data = sim_read(sim->ioaddr + PORT0_CONTROL);
	reg_data |= SIM_CARD_CLK_ENABLE;
	sim_write(reg_data, sim->ioaddr + PORT0_CONTROL);

	return 0;
}

static inline void sim_clk_disable(struct sim_t *sim)
{
	uint32_t reg_data;

	reg_data = sim_read(sim->ioaddr + PORT0_CONTROL);
	reg_data |= ~SIM_CARD_CLK_ENABLE;
	sim_write(reg_data, sim->ioaddr + PORT0_CONTROL);
}

static inline void sim_clk_put(struct sim_t *sim)
{
	/* TBD */
}
