/*
 * Copyright (c) 2011-2012 Freescale Semiconductor, Inc.
 *
 * Freescale Subscriber Identification Module(SIM) device driver
 *
 * Author: Harninder Rai <harninder.rai@freescale.com>
 * Author: Ashish Kumar  <ashish.kumar@freescale.com>
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/clk.h>
#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/miscdevice.h>
#include <linux/interrupt.h>
#include <linux/io.h>

#ifdef CONFIG_OF
#include <linux/of_platform.h>
/*#else
#include <linux/platform_device.h> */
#endif

#include <linux/fsl_sim_interface.h>
#include <linux/fsl_devices.h>

#include "sim.h"

static int sim_of_resource_init(struct of_device *ofdev,
		struct sim_t *sim)
{
	int ret = 0;
	struct device_node *np = ofdev->node;
	u32 *ptr = NULL;

	sim->clk_flag = 0;

	sim->ioaddr = of_iomap(np, 0);
	if (sim->ioaddr == NULL) {
		dev_err(&ofdev->dev, "%s: of_iomap failed\n", __func__);
		return -ENOMEM;
	}

	/* The first interrupt is the SIM general interrupt */
	sim->ipb_irq = irq_of_parse_and_map(np, 0);
	if (sim->ipb_irq == NO_IRQ) {
		dev_err(&ofdev->dev, "%s: Unable to map SIM general IRQ\n",
				__func__);
		return -ENODEV;
	}

	ptr = (u32 *)of_get_property(np, "clock-prescaler", NULL);
	if (!ptr) {
		dev_err(&ofdev->dev, "%s: Cannot find clock prescaler\n",
				__func__);
		return -EINVAL;
	}

	sim->clk.prescaler = (uint8_t)*ptr;

	sim->sim_clk_ops.clk_enable = sim_clk_enable;
	sim->sim_clk_ops.clk_disable = sim_clk_disable;
	sim->sim_clk_ops.clk_get = NULL;
	sim->sim_clk_ops.clk_put = sim_clk_put;
	return ret;
}

/* Following code is for ARM platform */
#if 0
static int sim_plat_resource_init(struct of_device *ofdev,
		struct sim_t *sim)
{
	int ret = 0;
	struct fsl_sim_platform_data *sim_plat = ofdev->dev.platform_data;

	sim->plat_data = sim_plat;
	sim->clk_flag = 0;

	sim->res = platform_get_resource(ofdev, IORESOURCE_MEM, 0);
	if (!sim->res) {
		dev_err(&ofdev->dev, "%s: Unable to map resource\n", __func__);
		return -ENOMEM;
	}

	/* request the sim clk and sim_serial_clk */
	sim->clk = clk_get(NULL, sim->plat_data->clock_sim);
	if (IS_ERR(sim->clk)) {
		dev_err(&ofdev->dev, "%s: Get CLK ERROR\n", __func__);
		return PTR_ERR(sim->clk);
	}
	pr_debug("sim clock:%lu\n", clk_get_rate(sim->clk));

	sim->ipb_irq = platform_get_irq(ofdev, 0);
	sim->dat_irq = platform_get_irq(ofdev, 1);
	if (!(sim->ipb_irq | sim->dat_irq)) {
		dev_err(&ofdev->dev, "%s: Unable to map irq\n", __func__);
		return -ENOMEM;
	}

	sim->sim_clk_ops->clk_enable = clk_enable;
	sim->sim_clk_ops->clk_disable = clk_disable;
	sim->sim_clk_ops->clk_get = NULL;
	/* sim->sim_clk_ops->clk_get = clk_get; */
	sim->sim_clk_ops->clk_put = clk_put;

	return ret;
}
#endif

static int sim_probe(struct of_device *ofdev,
		const struct of_device_id *match)
{
	int ret = 0;
	struct sim_t *sim = NULL;
	struct device_node *np = ofdev->node;

	sim = kzalloc(sizeof(struct sim_t), GFP_KERNEL);
	if (!sim) {
		dev_err(&ofdev->dev, "%s: No memory for sim\n", __func__);
		ret = -ENOMEM;
		return ret;
	};

	if (np) {
		ret = sim_of_resource_init(ofdev, sim);
		if (ret) {
			dev_err(&ofdev->dev, "%s: SIM initialization failed"
				"with error %d\n", __func__, ret);
			ret = -EINVAL;
			goto out;
		}
	} /*else { Following code is for ARM platform
		ret = sim_plat_resource_init(ofdev, sim);
		if (ret) {
			dev_err(&ofdev->dev, "%s: SIM initialization failed"
				"with error %d\n", __func__, ret);
			ret = -EINVAL;
			goto out;
		}
	}*/

	ret = sim_device_init(ofdev, sim);
	if (ret) {
		dev_err(&ofdev->dev, "%s: SIM initialization failed"
			"with error %d\n", __func__, ret);
		ret = -EINVAL;
		goto out;
	}

	return ret;
out:
	kfree(sim);
	return ret;
}

static int sim_remove(struct of_device *ofdev)
{
	struct sim_t *sim = platform_get_drvdata(ofdev);

	sim->sim_clk_ops.clk_put(sim);

	if (sim->ipb_irq)
		free_irq(sim->ipb_irq, sim);
	if (sim->dat_irq)
		free_irq(sim->dat_irq, sim);

	iounmap(sim->ioaddr);

	kfree(sim);
	/* misc_deregister(&sim_dev);*/

	return 0;
}

static struct of_device_id sim_match[] = {
	{
	 .compatible = "fsl,sim-v1.0",
	},
	{},
};

MODULE_DEVICE_TABLE(of, sim_match);

static struct of_platform_driver sim_driver = {
		.name = "fsl_sim",
		.owner = THIS_MODULE,
		.remove = sim_remove,
		.probe = sim_probe,
#ifdef CONFIG_OF
		.match_table = sim_match,
#endif
};

static int __init sim_drv_init(void)
{
	printk(KERN_INFO "%s: Cleaned up!!!!!!!!!\n", __func__);
	return of_register_platform_driver(&sim_driver);
}

static void __exit sim_drv_exit(void)
{
	of_unregister_platform_driver(&sim_driver);
}

module_init(sim_drv_init);
module_exit(sim_drv_exit);

MODULE_AUTHOR("Freescale Semiconductor, Inc.");
MODULE_DESCRIPTION("FSL SIM Driver");
MODULE_LICENSE("GPL v2");
