/*
 * caam - Freescale Integrated Security Engine (SEC) device driver
 * Support for accessing CAAM block from user space using UIO.
 *
 * Copyright 2011 Freescale Semiconductor, Inc
 */

#include <linux/uio_driver.h>

#include "compat.h"
#include "regs.h"
#include "intern.h"
#include "error.h"
#include "caamuio.h"

#define UIODEV_NAMELEN	20

struct caam_uio_device {
	/* UIO Device Name */
	char uio_dev_name[UIODEV_NAMELEN];

	struct uio_info uio_info;

	/* pointer to controller dev */
	struct caam_drv_private *ctrlpriv;

	/* JobR's register space */
	struct caam_job_ring __iomem *rregs;

	/* Ring id - corresponding to location of job ring node in dts */
	int uio_rid;
};

/* Main user space ring interrupt handler */
static irqreturn_t caam_uio_jr_interrupt(int irq, struct uio_info *dev_info)
{
	struct caam_uio_device *caamuiodev = dev_info->priv;
	struct device *dev = &caamuiodev->ctrlpriv->ofdev->dev;
	u32 irqstate;

	irqstate = rd_reg32(&caamuiodev->rregs->jrintstatus);
	if (!irqstate)
		return IRQ_NONE;

	if (irqstate & JRINT_JR_ERROR)
		dev_info(dev, "uio job ring %d error - irqstate: %08x\n",
			 caamuiodev->uio_rid, irqstate);

	/*mask valid interrupts */
	setbits32(&caamuiodev->rregs->rconfig_lo, JRCFG_IMSK);

	/* Have valid interrupt at this point, just ACK and trigger */
	wr_reg32(&caamuiodev->rregs->jrintstatus, irqstate);

	return IRQ_HANDLED;
}

static int caam_uio_irq_control(struct uio_info *info, s32 irq_control)
{
	struct caam_uio_device *caamuiodev = info->priv;

	switch (irq_control) {
	case SEC_UIO_SIMULATE_IRQ_CMD:
		uio_event_notify(info);
		return 0;

	case SEC_UIO_ENABLE_IRQ_CMD:
		/* Enable Job Ring interrupt */
		clrbits32(&caamuiodev->rregs->rconfig_lo, JRCFG_IMSK);
		return 0;

	case SEC_UIO_DISABLE_IRQ_CMD:
		/* Disable Job Ring interrupt */
		setbits32(&caamuiodev->rregs->rconfig_lo, JRCFG_IMSK);
		return 0;

	case SEC_UIO_RESET_SEC_ENGINE_CMD:
		/* Reset SEC Engine */
		caam_reset(caamuiodev->ctrlpriv);
		return 0;

	default:
		return 0;
	}

	return 0;
}

static int caam_uio_init_jr(struct of_device *ofdev,
			    struct device_node *np, int ridx, int rid)
{
	u32 *jroffset;
	struct resource res;
	struct caam_uio_device *caamuiodev;
	struct device *dev;
	int ret;

	dev = &ofdev->dev;
	jroffset = (u32 *) of_get_property(np, "reg", NULL);
	ret = of_address_to_resource(ofdev->node, 0, &res);
	if (ret) {
		dev_err(dev, "Failed reading reg property from DTS!\n");
		return ret;
	}

	caamuiodev = kzalloc(sizeof(struct caam_uio_device), GFP_KERNEL);
	if (!caamuiodev) {
		dev_err(dev,
			"Failed to allocate memory for caam uio job ring %d\n",
			rid);
		return -ENOMEM;
	}

	snprintf(caamuiodev->uio_dev_name, UIODEV_NAMELEN, "%s%d",
		 SEC_UIO_DEVICE_NAME, rid);
	caamuiodev->uio_info.name = caamuiodev->uio_dev_name;

	caamuiodev->uio_info.mem[0].addr = res.start + jroffset[0];
	caamuiodev->uio_info.mem[0].size = jroffset[1];
	caamuiodev->uio_info.mem[0].memtype = UIO_MEM_PHYS;

	caamuiodev->uio_info.version = SEC_UIO_VERSION;
	caamuiodev->uio_info.priv = caamuiodev;

	caamuiodev->uio_info.irq = of_irq_to_resource(np, 0, NULL);
	caamuiodev->uio_info.irq_flags = IRQF_SHARED;
	caamuiodev->uio_info.handler = caam_uio_jr_interrupt;
	caamuiodev->uio_info.irqcontrol = caam_uio_irq_control;

	caamuiodev->uio_info.mmap = NULL;
	caamuiodev->uio_info.open = NULL;
	caamuiodev->uio_info.release = NULL;

	caamuiodev->uio_rid = rid;
	caamuiodev->ctrlpriv = dev_get_drvdata(dev);
	caamuiodev->rregs =
	    (struct caam_job_ring __iomem *)((void *)caamuiodev->ctrlpriv->
					     ctrl + jroffset[0]);

	ret = uio_register_device(dev, &caamuiodev->uio_info);
	if (ret) {
		dev_err(dev, "uio_register_device() failed for Job Ring %d\n",
			rid);
		kfree(caamuiodev);
		return ret;
	}

	caamuiodev->ctrlpriv->uiodev[ridx] = &caamuiodev->uio_info;

	return 0;
}

static int __init caam_uio_init(void)
{
	struct device_node *dev_node, *np;
	struct of_device *ofdev;
	struct device *ctrldev;
	struct caam_drv_private *ctrlpriv;
	int ridx, rid;

	dev_node = of_find_compatible_node(NULL, NULL, "fsl,sec-v4.0");
	if (!dev_node)
		return -ENODEV;

	ofdev = of_find_device_by_node(dev_node);
	if (!ofdev)
		return -ENODEV;

	ctrldev = &ofdev->dev;
	ctrlpriv = dev_get_drvdata(ctrldev);
	of_node_put(dev_node);

	if (ctrlpriv->nr_uio_jrs) {
		ctrlpriv->uiodev =
		    kzalloc(sizeof(struct uio_info *) * ctrlpriv->nr_uio_jrs,
			    GFP_KERNEL);
		if (ctrlpriv->uiodev == NULL)
			return -ENOMEM;
		dev_info(ctrldev, "Job Rings configured for user space = %d\n",
			 ctrlpriv->nr_uio_jrs);
	} else {
		dev_err(ctrldev, "No Job Ring configured for user space\n");
		return -ENODEV;
	}

	ridx = 0;
	rid = 0;
	for_each_compatible_node(np, NULL, "fsl,sec-v4.0-job-ring") {
		if (of_get_property(np, "user-space-ring", NULL)) {
			caam_uio_init_jr(ofdev, np, ridx, rid);
			ridx++;
		}

		rid++;
	}

	return 0;
}

static void __exit caam_uio_exit(void)
{
	struct device_node *dev_node;
	struct of_device *ofdev;
	struct device *ctrldev;
	struct caam_drv_private *ctrlpriv;
	struct caam_uio_device *caamuiodev;
	struct uio_info *info;
	int i;

	dev_node = of_find_compatible_node(NULL, NULL, "fsl,sec-v4.0");
	if (!dev_node)
		return;

	ofdev = of_find_device_by_node(dev_node);
	if (!ofdev)
		return;

	ctrldev = &ofdev->dev;
	ctrlpriv = dev_get_drvdata(ctrldev);
	of_node_put(dev_node);

	for (i = 0; i < ctrlpriv->nr_uio_jrs; i++) {
		info = ctrlpriv->uiodev[i];
		caamuiodev = info->priv;
		uio_unregister_device(info);
		kfree(caamuiodev);
	}

	kfree(ctrlpriv->uiodev);
}

module_init(caam_uio_init);
module_exit(caam_uio_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("FSL CAAM UIO support");
MODULE_AUTHOR("Freescale Semiconductor - NMG/IDC");
