/*
 * drivers/misc/fsl_fpga.c
 * Freescale FPGA driver
 *
 * Existing FPGA driver is providing minimal functionality for RF BAND
 * configuration
 * Author: bhaskar <bhaskar.upadhaya@freescale.com>
 *
 * Copyright 2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/types.h>
#include <linux/io.h>
#include <linux/of_platform.h>
#include <net/dst.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/fsl_fpga.h>

struct fpga_priv {
	u8 *fpga_addr;
	struct device *dev;
};

static struct fpga_priv *fpga_priv;

/*
** Function: config_rf_regulator()
** Used to configure different bands ie BAND1, BAND7, BAND13 on 9132QDS
** via FPGA
** The FPGA register used to configure the various bands is BRDCFG9
*/

int config_rf_regulator(bool rf_pa, bool rf_lna, int rf_num)
{
	u8 *fpga_addr, reg;
	int rc = 0;

	if (!fpga_priv) {
		printk(KERN_ERR "fpga_priv un-initialized \r\n");
		rc = -EINVAL;
		goto out;
	}

	fpga_addr = fpga_priv->fpga_addr;
	reg = in_8(fpga_addr + BRDCFG9);

	switch (rf_num) {
	case RF1:
		reg = reg & RF1_MASK;
		reg |= (rf_pa ? 0 : 1) << RF1_PA_SHIFT;
		reg |= (rf_lna ? 0 : 1) << RF1_LNA_SHIFT;
		out_8(fpga_addr + BRDCFG9, reg);
		msleep(5000);
		break;
	case RF2:
		reg = reg & RF2_MASK;
		reg |= (rf_pa ? 0 : 1) << RF2_PA_SHIFT;
		reg |= (rf_lna ? 0 : 1) << RF2_LNA_SHIFT;
		out_8(fpga_addr + BRDCFG9, reg);
		msleep(5000);
		break;
	case RF3:
		reg = reg & RF3_MASK;
		reg |= (rf_pa ? 0 : 1) << RF3_PA_SHIFT;
		reg |= (rf_lna ? 0 : 1) << RF3_LNA_SHIFT;
		out_8(fpga_addr + BRDCFG9, reg);
		msleep(5000);
		break;
	default:
		rc = -EINVAL;
	}

out:
	return rc;
}
EXPORT_SYMBOL(config_rf_regulator);

static int fpga_probe(struct of_device *ofdev,
		const struct of_device_id *match)
{
	int rc = 0;
	struct device_node *np = ofdev->node;
	struct device *dev = &ofdev->dev;

	if (!np || !of_device_is_available(np))
		return -ENODEV;

	fpga_priv = kzalloc(sizeof(struct fpga_priv), GFP_KERNEL);
	if (!fpga_priv) {
		dev_err(dev, "Failed to allocate fpga_priv\n");
		return -ENOMEM;
	}

	fpga_priv->fpga_addr = of_iomap(np, 0);
	if (!fpga_priv->fpga_addr) {
		dev_err(dev, "fpga: iomap failed\n");
		rc = -ENOMEM;
		goto out;
	}
	fpga_priv->dev = dev;

out:
	return rc;
}

static int fpga_remove(struct of_device *ofdev)
{
	iounmap(fpga_priv->fpga_addr);
	kfree(fpga_priv);
	return 0;
}

static struct of_device_id fpga_match[] = {
	{
		.compatible = "fsl,9132-fpga",
	},
	{},
};
MODULE_DEVICE_TABLE(of, fpga_match);

static struct of_platform_driver fpga_driver = {
	.name = "fsl,9132-fpga",
	.match_table = fpga_match,
	.probe = fpga_probe,
	.remove = fpga_remove
};

static int __init fpga_mod_init(void)
{
	return of_register_platform_driver(&fpga_driver);
}

static void __exit fpga_exit(void)
{
	of_unregister_platform_driver(&fpga_driver);
}

module_init(fpga_mod_init);
module_exit(fpga_exit);
