/*
 * drivers/net/gianfar_1588.c
 *
 * Copyright 2008-2011 Freescale Semiconductor, Inc.
 * Copyright 2009-2011 IXXAT Automation GmbH
 *
 * Author: Anup Gangwar <anup.gangwar@freescale.com>
 *	   Yashpal Dutta <yashpal.dutta@freescale.com>
 *
 * Gianfar Ethernet Driver -- IEEE 1588 interface functionality
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 */

#include <linux/vmalloc.h>
#include <linux/of.h>
#include <linux/if_vlan.h>
#include <linux/net_tstamp.h>
#include "gianfar.h"
#include "time_sync.h"
#include <linux/of_platform.h>

#if defined(CONFIG_1588_MUX_eTSEC1) || defined(CONFIG_1588_MUX_eTSEC2)
#define MPC85XX_PMUXCR_OFFS 0x60
#if defined(CONFIG_1588_MUX_eTSEC1)
#define PMUXCR_eTSEC1_PPS 0x10000000
#endif
#if defined(CONFIG_1588_MUX_eTSEC2)
#define PMUXCR_eTSEC2_PPS 0x08000000
#endif
#endif
static void gfar_set_1588cnt(struct net_device *dev,
				struct gfar_ptp_time *gfar_time);
static int gfar_ptp_init_circ(struct gfar_ptp_circular *buf, int size);
static int gfar_ptp_is_empty(struct gfar_ptp_circular *buf);
static int gfar_ptp_nelems(struct gfar_ptp_circular *buf);
static int gfar_ptp_is_full(struct gfar_ptp_circular *buf);
static int gfar_ptp_insert(struct gfar_ptp_circular *buf,
				struct gfar_ptp_data *data);
static void gfar_set_fiper_alarm(struct net_device *dev,
				struct gfar_ptp_time *alarm);
static int gfar_ptp_find_and_remove(struct gfar_ptp_circular *buf,
				struct gfar_ptp_ident *ident,
				struct gfar_ptp_time *ts);
static void gfar_get_curr_cnt(struct gfar_regs_1588 __iomem *ptimer,
					struct gfar_ptp_time *curr_time);
static u32 nominal_frequency(u32 sysclock_freq);
static u8 gfar_1588_refclk_detected(struct gfar_private *priv);
static u32 freq_compensation;

static struct proc_dir_entry *gfar_1588_proc_file;
static struct gfar_node_info_t gfar_node;
static struct gfar_1588_data_t gfar_1588_data;

/* 64 bits multiplication and return the result*/
u64 multi64_oper(u32 multiplier, u32 multiplicand)
{
	u64 result = (u64)multiplier;
	result *= multiplicand;
	return result;
}

/* 64 bit division */
u32 div64_oper(u64 dividend, u32 divisor, u32 *quotient)
{
	u32 time_h, time_l;
	u32 result;
	u64 tmp_dividend;
	int i;

	*quotient = 0;

	time_h = (u32)(dividend >> 32);
	time_l = (u32)dividend;
	time_h = time_h % divisor;
	for (i = 1; i <= 32; i++) {
		tmp_dividend = (((u64)time_h << 32) | (u64)time_l);
		tmp_dividend = (tmp_dividend << 1);
		time_h = (u32)(tmp_dividend >> 32);
		time_l = (u32)tmp_dividend;
		result = time_h / divisor;
		time_h = time_h % divisor;
		*quotient += (result << (32 - i));
	}
	return time_h;
}

/*
 * Resource required for accessing 1588 Timer Registers. There are few 1588
 * modules registers which are present in eTSEC1 memory space only. The second
 * reg entry there in denotes the 1588 regs.
 */
int gfar_ptp_init(struct gfar_private *priv)
{
	priv->ptimer = ioremap(priv->timer_resource.start,
			sizeof(struct gfar_regs_1588));
	/* check if hw clock is available */
	if (priv->ptimer == NULL)
		return 1;
	/* initialize circular buffer for tx timestamps */
	if (gfar_ptp_init_circ(&(priv->tx_timestamps),
					(DEFAULT_PTP_TX_BUF_SZ+1)))
		return 1;
	/* initialize circular buffer for rx timestamps */
	if (gfar_ptp_init_circ(&(priv->rx_timestamps),
					(DEFAULT_PTP_RX_BUF_SZ+1)))
		return 1;
	priv->time_sync_priv = NULL;
	return 0;
}


void gfar_ptp_cleanup(struct gfar_private *priv)
{
	if (priv->ptimer != NULL)
		iounmap(priv->ptimer);
	if (priv->tx_timestamps.data_buf)
		vfree(priv->tx_timestamps.data_buf);
	if (priv->rx_timestamps.data_buf)
		vfree(priv->rx_timestamps.data_buf);
}

/**
 * Parse packets if they are PTP.
 * The PTP header can be found in an IPv4, IPv6 or in an IEEE802.3
 * ethernet frame. The function returns the position of the PTP packet
 * or NULL, if no PTP found
 */

u8 *gfar_ptp_parse_packet(struct sk_buff *skb, u16 *eth_type)
{
	u8 *position = skb->data + ETH_ALEN + ETH_ALEN;
	u8 *ptp_loc = NULL;

	*eth_type = *((u16 *)position);
	/* Check if outer vlan tag is here */
	switch (*eth_type) {
		/* Valid outer tags */
	case GFAR_VLAN_QINQ_1:
	case GFAR_VLAN_QINQ_2:
	case GFAR_VLAN_QINQ_3:
	case GFAR_VLAN_QINQ_4:
		{
		position += GFAR_VLAN_TAG_LEN;
		/* get next tag */
		*eth_type = *((u16 *)position);
		}
		break;
	}
	/* Check if inner tag is here */
	if (*eth_type == ETH_P_8021Q) {
		position += GFAR_VLAN_TAG_LEN;
		/* get next tag */
		*eth_type = *((u16 *)position);
	}
	/* set position after ethertype */
	position += GFAR_ETHTYPE_LEN;
	if (ETH_P_1588 == *eth_type) {
		ptp_loc = position;
		if ((ptp_loc[0] & 0xF) <= 3) { /* IEEE1588 event message
						* which needs timestamping */
			if (skb->len >= ((ptp_loc - skb->data) +
						GFAR_PTP_HEADER_SZE))
				return ptp_loc;
		}
	} else if (ETH_P_IP == *eth_type) {
		u8 *ip_header, *prot, *udp_header;
		u8 ip_version, ip_hlen;
		ip_header = position;
		ip_version = ip_header[0] >> 4; /* correct IP version? */
		if (0x04 == ip_version) {
			prot = ip_header + 9;
			if (GFAR_PACKET_TYPE_UDP == *prot) {
				u16 udp_dstPort;
				/* retrieve the size of the ip-header with the
						first byte of the ip-header: */
				/* version ( 4 bits) + Internet header
						length (4 bits) */
				ip_hlen   = (*ip_header & 0xf) * 4;
				udp_header = ip_header + ip_hlen;
				udp_dstPort = *((u16 *)(udp_header + 2));
				/* check the destination port address
					( 319 (0x013F) = PTP event port )*/
				if (udp_dstPort == GFAR_PTP_EVENT_PORT) {
					ptp_loc = udp_header + 8;
					/* long enough ? */
					if (skb->len >= ((ptp_loc - skb->data)
							+ GFAR_PTP_HEADER_SZE))
						return ptp_loc;
				}
			}
		}
	} else if (ETH_P_IPV6 == *eth_type) {
		u8 *ip_header, *udp_header, *prot;
		u8 ip_version;
		ip_header = position;
		ip_version = ip_header[0] >> 4;
		if (0x06 == ip_version) {
			prot = ip_header + 6;
			if (GFAR_PACKET_TYPE_UDP == *prot) {
				u16 udp_dstPort;
				udp_header = ip_header + 40;
				udp_dstPort = *((u16 *)(udp_header + 2));
				/* check the destination port address
				( 319 (0x013F) = PTP event port )*/
				if (udp_dstPort == GFAR_PTP_EVENT_PORT) {
					ptp_loc = udp_header + 8;
					/* long enough ? */
					if (skb->len >= ((ptp_loc - skb->data)
							+ GFAR_PTP_HEADER_SZE))
						return ptp_loc;
				}
			}
		}
	}
	return NULL; /* no PTP frame */
}

/**
 * Store the tx hardware timestamp with additional information
 * in the gfar internal ringbuffer for timestamps.
 */
void gfar_ptp_store_txstamp(struct net_device *dev, struct sk_buff *skb,
						struct gfar_ptp_time *tx_ts)
{
	struct gfar_ptp_data tmp_tx_time;
	struct gfar_private *priv = netdev_priv(dev);
	u8  *ptp_loc;
	u16 eth_type;

	ptp_loc = gfar_ptp_parse_packet(skb, &eth_type);
	if (ptp_loc != NULL) {
		/* store identification data */
		switch (eth_type) {
		case ETH_P_IP:
			tmp_tx_time.ident.netw_prot = GFAR_PTP_PROT_IPV4;
			break;
		case ETH_P_IPV6:
			tmp_tx_time.ident.netw_prot = GFAR_PTP_PROT_IPV6;
			break;
		case ETH_P_1588:
			tmp_tx_time.ident.netw_prot = GFAR_PTP_PROT_802_3;
			break;
		default:
			return;
		}
		tmp_tx_time.ident.version = (*(ptp_loc + 1)) & 0X0F;
		tmp_tx_time.ident.msg_type = (*(ptp_loc)) & 0x0F;
		tmp_tx_time.ident.seq_id = *((u16 *)(ptp_loc +
						GFAR_PTP_HEADER_SEQ_OFFS));
		memcpy(tmp_tx_time.ident.snd_port_id,
		&ptp_loc[GFAR_PTP_SPID_OFFS], GFAR_PTP_SOURCE_PORT_LENGTH);
		/* store tx timestamp */
		tmp_tx_time.ts = *tx_ts;
		/* insert timestamp in circular buffer */
		gfar_ptp_insert(&(priv->tx_timestamps), &tmp_tx_time);
	}
}

/**
 * Store the rx hardware timestamp with additional information
 * in the gfar internal ringbuffer for timestamps.
 */
void gfar_ptp_store_rxstamp(struct net_device *dev, struct sk_buff *skb,
						struct gfar_ptp_time *rx_ts)
{
	struct gfar_ptp_data tmp_rx_time;
	struct gfar_private *priv = netdev_priv(dev);
	u8 *ptp_loc;
	u16 eth_type;

	ptp_loc = gfar_ptp_parse_packet(skb, &eth_type);
	if (ptp_loc != NULL) {
		/* store identification data */
		tmp_rx_time.ident.version = (*(ptp_loc + 1)) & 0X0F;
		tmp_rx_time.ident.msg_type = (*(ptp_loc)) & 0x0F;
		switch (eth_type) {
		case ETH_P_IP:
			tmp_rx_time.ident.netw_prot = GFAR_PTP_PROT_IPV4;
			break;
		case ETH_P_IPV6:
			tmp_rx_time.ident.netw_prot = GFAR_PTP_PROT_IPV6;
			break;
		case ETH_P_1588:
			tmp_rx_time.ident.netw_prot = GFAR_PTP_PROT_802_3;
			break;
		default:
			return;
		}
		tmp_rx_time.ident.seq_id = *((u16 *)(ptp_loc +
						GFAR_PTP_HEADER_SEQ_OFFS));
		memcpy(tmp_rx_time.ident.snd_port_id,
			&ptp_loc[GFAR_PTP_SPID_OFFS],
			GFAR_PTP_SOURCE_PORT_LENGTH);
		/* store rx timestamp */
		tmp_rx_time.ts = *rx_ts;

		/* insert timestamp in circular buffer */
		gfar_ptp_insert(&(priv->rx_timestamps), &tmp_rx_time);
	}
}

int gfar_ptp_init_circ(struct gfar_ptp_circular *buf, int size)
{
	buf->data_buf = (struct gfar_ptp_data *)
		vmalloc(size * sizeof(struct gfar_ptp_data));

	if (!buf->data_buf)
		return 1;
	buf->front = 0;
	buf->end = 0;
	buf->size = size;
	return 0;
}

static inline int gfar_ptp_calc_index(int size, int curr_index, int offset)
{
	return (curr_index + offset) % size;
}

int gfar_ptp_is_empty(struct gfar_ptp_circular *buf)
{
	return (buf->front == buf->end);
}

int gfar_ptp_nelems(struct gfar_ptp_circular *buf)
{
	const int front = buf->front;
	const int end = buf->end;
	const int size = buf->size;
	int n_items;

	if (end > front)
		n_items = end - front;
	else if (end < front)
		n_items = size - (front - end);
	else
		n_items = 0;

	return n_items;
}

int gfar_ptp_is_full(struct gfar_ptp_circular *buf)
{
	if (gfar_ptp_nelems(buf) == (buf->size - 1))
		return 1;
	else
		return 0;
}

int gfar_ptp_insert(struct gfar_ptp_circular *buf,
			struct gfar_ptp_data *data)
{
	struct gfar_ptp_data *tmp;

	if (gfar_ptp_is_full(buf))
		return 1;

	tmp = (buf->data_buf + buf->end);

	memcpy(tmp, data, sizeof(struct gfar_ptp_data));

	buf->end = gfar_ptp_calc_index(buf->size, buf->end, 1);

	return 0;
}

int gfar_ptp_find_and_remove(struct gfar_ptp_circular *buf,
				struct gfar_ptp_ident *ident,
				struct gfar_ptp_time *ts)
{
	int i;
	int size = buf->size, end = buf->end;
	struct gfar_ptp_ident *tmp_ident;

	if (gfar_ptp_is_empty(buf))
		return 1;

	i = buf->front;
	while (i != end) {
		tmp_ident = &(buf->data_buf + i)->ident;
		if (tmp_ident->version == ident->version) {
			if (tmp_ident->msg_type == ident->msg_type) {
				if ((tmp_ident->netw_prot == ident->netw_prot)
					|| (ident->netw_prot ==
					GFAR_PTP_PROT_DONTCARE)) {
					if (tmp_ident->seq_id ==
						ident->seq_id) {
						int ret = memcmp(tmp_ident->
							snd_port_id,
							ident->snd_port_id,
						GFAR_PTP_SOURCE_PORT_LENGTH);
						if (0 == ret)
							break;
					}
				}
			}
		}
		/* get next */
		i = gfar_ptp_calc_index(size, i, 1);
	}
	/* not found ? */
	if (i == end) {
		/* buffer full ? */
		if (gfar_ptp_is_full(buf))
			/* drop one in front */
			buf->front = gfar_ptp_calc_index(size, buf->front, 1);
		return 1;
	}
	*ts = (buf->data_buf + i)->ts;
	/* set front pointer to postion after found */
	buf->front = gfar_ptp_calc_index(size, i, 1);

	return 0;
}

/* Set the 1588 timer counter registers */
static void gfar_set_1588cnt(struct net_device *dev,
				struct gfar_ptp_time *gfar_time)
{
	struct gfar_private *priv = netdev_priv(dev);
	u32 high, low;
	u64 alarm_value = 0, act_time;
	struct gfar_ptp_attr_t ptp_attr;
	struct device_node *np;

	memset(&ptp_attr, 0, sizeof(struct gfar_ptp_attr_t));
	np = of_find_compatible_node(NULL, NULL, "fsl,gianfar-ptp-timer");
	if (np == NULL)	{
		printk(KERN_ERR "1588: Cannot find"
				" gianfar-ptp-timer node \r\n");
		return;
	}
	alarm_value = multi64_oper((u32)gfar_time->sec, (u32)TMR_SEC);

	act_time = alarm_value;
	act_time += (u64)gfar_time->nsec;

	high = (u32)(act_time >> 32);
	low  = (u32)act_time;

	/* We must write the tmr_cnt_l register first */
	gfar_write(&priv->ptimer->tmr_cnt_l, low);
	gfar_write(&priv->ptimer->tmr_cnt_h, high);
	if (gfar_ptp_cal_attr(&ptp_attr))
		return;

	gfar_write(&(priv->ptimer->tmr_fiper1),
		(ptp_attr.tmr_fiper1 - (ptp_attr.tclk_period >> 16)));
	/* restart fiper two second later */
	gfar_time->sec += 2;
	gfar_time->nsec = 0;
	gfar_set_fiper_alarm(dev, gfar_time);
}

void gfar_cnt_to_ptp_time(u32 high, u32 low, struct gfar_ptp_time *gfar_time)
{
	u64 temp;
	u32 sec, nsec;

	temp = high;
	temp <<= 32;
	temp += low;

	nsec = div64_oper(temp, TMR_SEC, &sec);

	gfar_time->nsec = nsec;
	gfar_time->sec = sec;
}

static void gfar_get_curr_cnt(struct gfar_regs_1588 __iomem *ptimer,
					struct gfar_ptp_time *curr_time)
{
	u32 high, high_ctrl, low, low_ctrl;

	high = gfar_read(&ptimer->tmr_cnt_h);
	low  = gfar_read(&ptimer->tmr_cnt_l);
	high_ctrl = gfar_read(&ptimer->tmr_cnt_h);
	low_ctrl = gfar_read(&ptimer->tmr_cnt_l);
	/* do not accept a time at the rollover from low to high */
	if (high != high_ctrl) {
		/* rollover happened, take the 2nd measurement -
		   rollover just happens about all 4 seconds */
		high = high_ctrl;
		low = low_ctrl;
	}
	gfar_cnt_to_ptp_time(high, low, curr_time);
}

/*set Fiper Trigger Alarm */
void gfar_set_fiper_alarm(struct net_device *dev, struct gfar_ptp_time *alarm)
{
	struct gfar_private *priv = netdev_priv(dev);
	struct gfar_ptp_attr_t ptp_attr;
	struct device_node *np;
	u64 act_time;

	memset(&ptp_attr, 0, sizeof(struct gfar_ptp_attr_t));
	np = of_find_compatible_node(NULL, NULL, "fsl,gianfar-ptp-timer");
	if (np == NULL)	{
		printk(KERN_ERR "1588: Cannot find"
				" gianfar-ptp-timer node \r\n");
		return;
	}
	act_time = alarm->sec;
	act_time *= (u32)TMR_SEC;
	act_time += alarm->nsec;
	gfar_write(&(priv->ptimer->tmr_alarm1_l), (u32)act_time);
	gfar_write(&(priv->ptimer->tmr_alarm1_h), (u32)(act_time>>32));
	if (gfar_ptp_cal_attr(&ptp_attr))
		return;
	gfar_write(&(priv->ptimer->tmr_fiper1), ptp_attr.tmr_fiper1
		- (ptp_attr.tclk_period>>16));
}

int gfar_ioctl_1588(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	struct gfar_private *priv = netdev_priv(dev);
	struct gfar_ptp_time act_time;
	struct gfar_ptp_data ptp_ts_data;
	struct gfar_ptp_data *ptp_dat_user;
	struct hwtstamp_config config;
	int retval = 0;
	u32 addend;

	switch (cmd) {
	case SIOCSHWTSTAMP:
		/* check if priv->ptimer is present -
			necessary for HW-timestamping*/
		if (NULL == priv->ptimer)
			return -ENODEV;
		if (copy_from_user(&config, ifr->ifr_data, sizeof(config)))
			return -EFAULT;
		switch (config.tx_type) {
		case HWTSTAMP_TX_OFF:
			priv->hwts_tx_en = 0;
			/* remove RTPE bit - disable timestamp
					insertion on tx packets */
			gfar_write(&(priv->ptimer->tmr_ctrl),
				gfar_read(&(priv->ptimer->tmr_ctrl))
						& (~TMR_RTPE));
			break;
		case HWTSTAMP_TX_ON:
			priv->hwts_tx_en = 1;
			/* add RTPE bit - enable timestamp insertion
							on tx packets */
			gfar_write(&(priv->ptimer->tmr_ctrl),
				gfar_read(&(priv->ptimer->tmr_ctrl))
								| TMR_RTPE);
			break;
		default:
			return -ERANGE;
		}
		switch (config.rx_filter) {
		case HWTSTAMP_FILTER_NONE:
			priv->hwts_rx_en = 0;
			break;
		case HWTSTAMP_FILTER_ALL:
			priv->hwts_rx_en = 1;
			break;
		default:
			return -ERANGE;
		}
		return 0;
		break;
	case PTP_ENBL_TXTS_IOCTL:
		priv->hwts_tx_en_ioctl = 1;
		/* add RTPE bit - enable timestamp insertion on tx packets */
		gfar_write(&(priv->ptimer->tmr_ctrl),
			gfar_read(&(priv->ptimer->tmr_ctrl)) | TMR_RTPE);
		break;
	case PTP_DSBL_TXTS_IOCTL:
		priv->hwts_tx_en_ioctl = 0;
		/* remove RTPE bit - disable timestamp insertion
						on tx packets */
		gfar_write(&(priv->ptimer->tmr_ctrl),
		gfar_read(&(priv->ptimer->tmr_ctrl)) & (~TMR_RTPE));
		break;
	case PTP_ENBL_RXTS_IOCTL:
		priv->hwts_rx_en_ioctl = 1;
		break;
	case PTP_DSBL_RXTS_IOCTL:
		priv->hwts_rx_en_ioctl = 0;
		break;
	case PTP_GET_TX_TIMESTAMP:
		ptp_dat_user = (struct gfar_ptp_data *)ifr->ifr_data;
		if (0 != copy_from_user(&ptp_ts_data.ident,
				&ptp_dat_user->ident,
					sizeof(ptp_ts_data.ident)))
			return -EINVAL;
		if (gfar_ptp_find_and_remove(&(priv->tx_timestamps),
				&ptp_ts_data.ident, &ptp_ts_data.ts) != 0)
			return -EAGAIN;
		if (copy_to_user(&ptp_dat_user->ts, &ptp_ts_data.ts,
						sizeof(ptp_ts_data.ts)) > 0)
			return -EFAULT;
		break;
	case PTP_GET_RX_TIMESTAMP:
		ptp_dat_user = (struct gfar_ptp_data *)ifr->ifr_data;
		if (0 != copy_from_user(&ptp_ts_data.ident,
			&ptp_dat_user->ident, sizeof(ptp_ts_data.ident)))
			return -EINVAL;
		if (gfar_ptp_find_and_remove(&(priv->rx_timestamps),
				&ptp_ts_data.ident, &ptp_ts_data.ts) != 0)
			return -EAGAIN;
		if (copy_to_user(&ptp_dat_user->ts,
				&ptp_ts_data.ts, sizeof(ptp_ts_data.ts)) > 0)
			return -EFAULT;
		break;
	case PTP_GET_TIME:
		gfar_get_curr_cnt(priv->ptimer, &act_time);
		if (copy_to_user(ifr->ifr_data, &act_time,
					sizeof(act_time)) > 0)
			return -EFAULT;
		break;
	case PTP_SET_TIME:
		if (0 != copy_from_user(&act_time, ifr->ifr_data,
							sizeof(act_time)))
			return -EINVAL;
		gfar_set_1588cnt(dev, &act_time);
		break;
	case PTP_SET_FIPER_ALARM:
		if (0 != copy_from_user(&act_time, ifr->ifr_data,
							sizeof(act_time)))
			return -EINVAL;
		gfar_set_fiper_alarm(dev, &act_time);
		break;
	case PTP_SET_ADJ:
		if (0 != copy_from_user(&addend, ifr->ifr_data, sizeof(addend)))
			return -EINVAL;
		/* assign new value directly */
		gfar_write(&priv->ptimer->tmr_add, addend);
		break;
	case PTP_GET_ADJ:
		/* return initial timer add value
		 * to calculate drift correction */
		if (copy_to_user(ifr->ifr_data, &freq_compensation,
					sizeof(freq_compensation)) > 0)
			return -EFAULT;
		break;
	case PTP_CLEANUP_TS:
		/* reset tx-timestamping buffer */
		priv->tx_timestamps.front = 0;
		priv->tx_timestamps.end = 0;
		priv->tx_timestamps.size = (DEFAULT_PTP_TX_BUF_SZ + 1);
		/* reset rx-timestamping buffer */
		priv->rx_timestamps.front = 0;
		priv->rx_timestamps.end = 0;
		priv->rx_timestamps.size = (DEFAULT_PTP_RX_BUF_SZ + 1);
		break;
	case PTP_TIMESYNC_IOCTL:
		return do_time_sync_ioctl(dev, ifr);
	default:
		return -EINVAL;
	}
	return retval;
}

/* 1588 Module intialization and filer table populating routine*/
void gfar_1588_start(struct net_device *dev)
{
	struct gfar_private *priv = netdev_priv(dev);
	struct gfar_ptp_attr_t ptp_attr;
	u32 *tmr_prsc = 0, *cksel = 0;
	struct device_node *np;
	u32 tmr_ctrl;

	priv->hwts_tx_en = 0;
	priv->hwts_rx_en = 0;
	priv->hwts_tx_en_ioctl = 0;
	priv->hwts_rx_en_ioctl = 0;
	memset(&ptp_attr, 0, sizeof(struct gfar_ptp_attr_t));
	if (gfar_ptp_cal_attr(&ptp_attr))
		return;
	np = of_find_compatible_node(NULL, NULL, "fsl,gianfar-ptp-timer");
	if (np == NULL)	{
		printk(KERN_ERR "1588: Cannot find"
				" gianfar-ptp-timer node \r\n");
		return;
	}

	tmr_prsc = (u32 *)of_get_property(np, "tmr-prsc", NULL);
	if (tmr_prsc == NULL)	{
		printk(KERN_ERR "1588: Cannot find tmr-prsc property \r\n");
		return;
	}

	gfar_write(&(priv->ptimer->tmr_prsc), *tmr_prsc);

	gfar_write(&(priv->ptimer->tmr_fiper1), ptp_attr.tmr_fiper1
		 - (ptp_attr.tclk_period>>16));

	gfar_write(&(priv->ptimer->tmr_alarm1_l), TMR_ALARM1_L);
	gfar_write(&(priv->ptimer->tmr_alarm1_h), TMR_ALARM1_H);

	/* Need to mask the TCLK bits as they are initialized with 1 */
	gfar_write(&(priv->ptimer->tmr_ctrl),
		(gfar_read(&(priv->ptimer->tmr_ctrl))
			 & ~TMR_CTRL_TCLK_MASK) | (ptp_attr.tclk_period));

	/* initialize TMR_ADD with the initial frequency compensation value:
	 * freq_compensation = (2^32 / frequency ratio)
	 */
	div64_oper((((u64)2 << 31)*100),
		ptp_attr.freq_div_ratio, &freq_compensation);
	gfar_write(&(priv->ptimer->tmr_add), freq_compensation);

	cksel = (u32 *)of_get_property(np, "cksel", NULL);
	if (cksel == NULL)	{
		printk(KERN_ERR "1588: Cannot find cksel property \r\n");
		return;
	}
	if (EXTERNAL_SRC == *cksel)
		gfar_write(&(priv->ptimer->tmr_ctrl),
			gfar_read(&(priv->ptimer->tmr_ctrl))
				& (~TMR_CTRL_EXT_CLK));
	else if (ETSEC_SYS_CLK == *cksel)
		gfar_write(&(priv->ptimer->tmr_ctrl),
			gfar_read(&(priv->ptimer->tmr_ctrl))
			| TMR_CTRL_SYS_CLK);

	gfar_write(&(priv->ptimer->tmr_ctrl),
		gfar_read(&(priv->ptimer->tmr_ctrl)) |
		TMR_CTRL_ENABLE | TMR_CTRL_FIPER_START);

	if (EXTERNAL_SRC == *cksel) {
		if (FALSE == gfar_1588_refclk_detected(priv)) {
			printk(KERN_ERR "\nIEEE1588: External reference clock"
				" not detected. You may want to"
				" review \n\tthe board's jumper/switch"
				" settings; however, driver will try to"
				" \n\tswitch to system-clock.\n");
			tmr_ctrl = gfar_read(&(priv->ptimer->tmr_ctrl));
			gfar_write(&(priv->ptimer->tmr_ctrl),
					gfar_read(&(priv->ptimer->tmr_ctrl)) |
					TMR_CTRL_ENABLE | 1);
			priv->ieee_1588_refclk_freq =
					ptp_attr.sysclock_freq * MHz;
			if (FALSE == gfar_1588_refclk_detected(priv))
				printk(KERN_ERR "IEEE1588: Even system clock is"
					" not detected. The Etsec/1588"
					" \n\tmodule may not work correctly");
		} else {
			priv->ieee_1588_refclk_freq = EXTERNAL_CLK_FREQ;
		}
	}
}

/* Cleanup routine for 1588 module.
 * When PTP is disabled this routing is called */
void gfar_1588_stop(struct net_device *dev)
{
	struct gfar_private *priv = netdev_priv(dev);

	gfar_write(&priv->ptimer->tmr_ctrl,
		gfar_read(&priv->ptimer->tmr_ctrl)
		& ~TMR_CTRL_ENABLE);
}

void pmuxcr_guts_write(void)
{
#if defined(CONFIG_1588_MUX_eTSEC1) || defined(CONFIG_1588_MUX_eTSEC2)
	struct device_node *np = NULL;
	void __iomem *immap = NULL;
	u32 pmuxcr;

	np =  of_find_compatible_node(NULL, NULL, "fsl,p2020-guts");
	immap = of_iomap(np, 0);
	if (immap) {
		pmuxcr = in_be32(immap + MPC85XX_PMUXCR_OFFS);

#if defined(CONFIG_1588_MUX_eTSEC1)
		pmuxcr |= PMUXCR_eTSEC1_PPS;
#endif

#if defined(CONFIG_1588_MUX_eTSEC2)
		pmuxcr |= PMUXCR_eTSEC2_PPS;
#endif
		out_be32(immap + MPC85XX_PMUXCR_OFFS, pmuxcr);
		iounmap(immap);
	} else {
		printk(KERN_INFO "1588 muxing could not be done,"
				" mapping failed\n");
	}
#endif
}

/**
 * nominal_frequency - This function calculates the nominal frequency.
 * nominal frequency is the desired clock frequency.
 * @sysclock_freq: Timer Oscillator Frequency
 *
 * Description:
 *  Returns the nominal frequency which is calculated on the following
 *  basis.
 *  nominal frequency should be less than the Timer Oscillator frequency.
 *  nominal frequency should be a factor of 1000.
 *
 *  Eg If Timer Oscillator frequency is 400.
 *     then nominal frequency can be 250.
 *
 *     If Timer Oscillator frequency is 600.
 *     then nominal frequency can be 500.
 *
 *     If Timer Oscillator frequency is 333.
 *     then nominal frequency can be 250.
 */
u32 nominal_frequency(u32 sysclock_freq)
{
	u32 remainder = 0;

	remainder = sysclock_freq % 50;
	if (remainder != 0) {
		sysclock_freq = sysclock_freq - remainder;
		sysclock_freq += 50;
	}

	while ((10000 % (sysclock_freq -= 50)) != 0)
		continue;
	return sysclock_freq;
}

int gfar_ptp_cal_attr(struct gfar_ptp_attr_t *ptp_attr)
{
	u32 *sysclock_freq, nominal_freq, tclk_period = 0, *cksel = 0;
	struct device_node *np;

	np = of_find_compatible_node(NULL, NULL, "fsl,gianfar-ptp-timer");
	if (np == NULL)	{
		printk(KERN_ERR "1588: Cannot find gianfar-ptp-timer node \r\n");
		return 1;
	}
	sysclock_freq = (u32 *)of_get_property(np, "timer-frequency", NULL);
	if (sysclock_freq == NULL) {
		printk(KERN_DEBUG "1588: Cannot find timer-frequency property \r\n");
		return 1;
	}

	cksel = (u32 *)of_get_property(np, "cksel", NULL);
	if (cksel == NULL)	{
		printk(KERN_ERR "1588: Cannot find cksel property \r\n");
		return 1;
	}

	nominal_freq = nominal_frequency(DIV_ROUND(*sysclock_freq, 1000)
							/ 1000);

	/* TCLK_PERIOD = 10^9/Nominal_Frequency in MHZ */
	/* The timer clock counter will increment by TCLK_PERIOD every time the
	 * accumulator register overflows. This clock period must be larger
	 * than the clock period of the timer reference clock. For applications
	 * where user does not want the clock period to be added, they can
	 * program this field to 1 to count the clock ticks.
	 * This field defaulted to 1 to count overflow ticks.
	 */
	if (EXTERNAL_SRC == *cksel)
		tclk_period = 1;
	else if (ETSEC_SYS_CLK == *cksel)
		tclk_period = 1000/nominal_freq;

	ptp_attr->tmr_fiper1 = ONE_GIGA;

	/*
	 * left shifting by 16 is done to meet the requirement
	 * of TMR_CTRL[TCLK_PERIOD]
	 */
	tclk_period <<= 16;
	ptp_attr->tclk_period = tclk_period;
	ptp_attr->nominal_freq = nominal_freq;
	ptp_attr->sysclock_freq = DIV_ROUND(*sysclock_freq, 1000) / 1000;

	/*
	 * FreqDivRatio = Timer Oscillator Freq / Nominal Freq
	 * and Timer Oscillator Freq = System Clock Freq
	 */
	ptp_attr->freq_div_ratio = (ptp_attr->sysclock_freq *
					100) / ptp_attr->nominal_freq;

	return 0;
}

static int gfar_1588_proc_read(char *buffer,
		char **buffer_location,
		off_t offset, int buffer_length, int *eof, void *data)
{
	int len;
	struct gfar_1588_data_t *gfar_1588_info = \
				(struct gfar_1588_data_t *)data;

	len = sprintf(buffer, "%s = '%s'\n",
			gfar_1588_info->name, gfar_1588_info->value);
	return len;
}

static int gfar_1588_proc_write(struct file *file, const char *buffer,
			unsigned long count, void *data)
{
	int i, cnt;
	struct gfar_private *priv;
	struct device_node *np = NULL;
	struct of_device *ofdev;
	struct gfar_1588_data_t *gfar_1588_info = \
				(struct gfar_1588_data_t *)data;
	struct gfar __iomem *regs;

	np = of_find_compatible_node(NULL, NULL, "fsl,gianfar-ptp-timer");
	if (np == NULL)	{
		printk(KERN_ERR "1588: Cannot find ptp_timer node. Check"
				" ptp_timer node in dts \r\n");
		return -EINVAL;
	}

	if (count > GFAR_1588_PROCFS_MAX_SIZE)
		count = GFAR_1588_PROCFS_MAX_SIZE;

	if (copy_from_user(gfar_1588_info->value, buffer, count))
		return -EFAULT;
	gfar_1588_info->value[count - 1] = '\0';
	printk(KERN_DEBUG "\r\n  buffer (%s) \r\n", gfar_1588_info->value);
	cnt = gfar_node.match_cnt / sizeof(struct of_device_id);

	/*
	 * Here we are getting the valid index ie "i" for which we have
	 * valid compatible string in the dts file.
	 */
	for (i = 0; i < (cnt - 1); i++) {
		printk(KERN_DEBUG "\r\n gfar-node is (%s)\r\n", \
				gfar_node.gfar_node_match[i].compatible);
		if (!of_find_compatible_node(NULL, NULL, \
				gfar_node.gfar_node_match[i].compatible))
			continue;
		else
			break;
	}

	/*
	 * After getting the valid index "i" from above for loop
	 * we are Enabling/Disabling the 1588 functionality on the
	 * eTSEC controllers.
	 */
	while (1) {
		np = of_find_compatible_node(np, NULL,
				gfar_node.gfar_node_match[i].compatible);
		if (np == NULL) {
			printk(KERN_DEBUG "\r\nUnale to find gfar-node\n");
			break;
		} else {
			ofdev = of_find_device_by_node(np);
			priv = dev_get_drvdata(&ofdev->dev);
			regs = priv->gfargrp[0].regs;
			if (!strcmp(gfar_1588_info->value, "0")) {
				printk(KERN_DEBUG "\r\n PTPD OFF \r\n");
				priv->ptimer_present = 0;
				gfar_write(&priv->ptimer->tmr_ctrl,
					gfar_read(&priv->ptimer->tmr_ctrl)
							& ~TMR_CTRL_ENABLE);
				gfar_write(&regs->rctrl,
					gfar_read(&regs->rctrl)
							& ~RCTRL_TS_ENABLE);
			} else if (!strcmp(gfar_1588_info->value, "1")) {
				printk(KERN_DEBUG "\r\n PTPD ON \r\n");
				priv->ptimer_present = 1;
				gfar_write(&priv->ptimer->tmr_ctrl,
					gfar_read(&priv->ptimer->tmr_ctrl)
							| TMR_CTRL_ENABLE);
				gfar_write(&regs->rctrl,
					gfar_read(&regs->rctrl)
							| RCTRL_TS_ENABLE);
			}
		}
	}
	return count;
}

void gfar_1588_proc_init(struct of_device_id *dev_id, int cnt)
{
	gfar_1588_proc_file = create_proc_entry(GFAR_1588_PROCFS_NAME,
							 0644, NULL);

	if (gfar_1588_proc_file == NULL) {
		remove_proc_entry(GFAR_1588_PROCFS_NAME, NULL);
		printk(KERN_ALERT "Error: Could not initialize /proc/%s\n",
			GFAR_1588_PROCFS_NAME);
		return;
	}
	strcpy(gfar_1588_data.name, GFAR_1588_PROCFS_NAME);
	strcpy(gfar_1588_data.value, "1");
	gfar_1588_proc_file->read_proc  = gfar_1588_proc_read;
	gfar_1588_proc_file->write_proc = gfar_1588_proc_write;
	gfar_1588_proc_file->mode	= S_IFREG | S_IRUGO;
	gfar_1588_proc_file->uid	= 0;
	gfar_1588_proc_file->gid	= 0;
	gfar_1588_proc_file->data	= &gfar_1588_data;
	gfar_node.gfar_node_match = dev_id;
	gfar_node.match_cnt = cnt;
	printk(KERN_INFO "/proc/%s created \r\n", GFAR_1588_PROCFS_NAME);
}

void gfar_1588_proc_exit()
{
	remove_proc_entry(GFAR_1588_PROCFS_NAME, NULL);
	printk(KERN_INFO "/proc/%s removed \r\n", GFAR_1588_PROCFS_NAME);
}

u8 gfar_1588_refclk_detected(struct gfar_private *priv)
{
	u32 tmr_stat;
	u32 cnt = 0;

	tmr_stat = gfar_read(&(priv->ptimer->tmr_stat));
	while ((cnt < RETRY_CNT) &&
			((tmr_stat & TMR_STAT_CLK_RCD) == 0)) {
		tmr_stat = gfar_read(&(priv->ptimer->tmr_stat));
		cnt++;
	}
	if (tmr_stat & TMR_STAT_CLK_RCD)
		return TRUE;
	else
		return FALSE;
}
