/*
 * drivers/net/time_sync.c
 * Freescale time-sync utility
 *
 * Time sync utility provides an interface to correct and synchronize the
 * clock source for an SoC with an atomic clock like GPS, PTP master
 * and NTP server.
 *
 * Author: bhaskar upadhaya <bhaskar.upadhaya@freescale.com>
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/wait.h>
#include <linux/of.h>
#include <linux/of_platform.h>
#include <linux/rfdev_ioctl.h>
#include <net/genetlink.h>
#include "time_sync.h"
#include "gianfar.h"

static irqreturn_t time_sync_isr(int irq, void *arg);
static int time_sync_priv(struct time_sync_priv *time_sync_priv);
static void time_sync_sysfs_exit(struct kobject *time_sync_kobj);
static int time_sync_sysfs_init(struct gfar_private *priv);
static int register_1588_interrupt(struct net_device *dev);
static void time_sync_free(struct gfar_private *priv);
static void time_sync_config_fiper(struct gfar_private *priv);
static void time_sync_config_blackout(struct gfar_private *priv,
			u32 timeout);
static void time_sync_cfg_lte_delay_enable_fiper(struct gfar_private *priv,
			u32 fiper_delay, u32 lte_delay);

static struct time_sync_sysfs attr1 = {
	.attr.name = "black_out_duration",
	.attr.mode = 0444,
	.cmd_type = BLACK_OUT_DURATION,
};

static struct time_sync_sysfs attr2 = {
	.attr.name = "time_sync_source",
	.attr.mode = 0444,
	.cmd_type = TIME_SYNC_SOURCE,
};

static struct time_sync_sysfs attr3 = {
	.attr.name = "lte_delay",
	.attr.mode = 0444,
	.cmd_type = LTE_DELAY,
};

static struct time_sync_sysfs attr4 = {
	.attr.name = "last_black_out_occurence",
	.attr.mode = 0444,
	.cmd_type = LAST_BLACKOUT_OCCURENCE,
};

static struct time_sync_sysfs attr5 = {
	.attr.name = "new_mode_after_blackout",
	.attr.mode = 0444,
	.cmd_type = NEW_MODE_AFTER_BLACKOUT,
};

static struct time_sync_sysfs attr6 = {
	.attr.name = "correction_mode",
	.attr.mode = 0444,
	.cmd_type = CORRECTION_MODE,
};

static struct time_sync_sysfs attr7 = {
	.attr.name = "stats",
	.attr.mode = 0444,
	.cmd_type = STATS,
};

static struct attribute *time_sync_attr[] = {
	&attr1.attr,
	&attr2.attr,
	&attr3.attr,
	&attr4.attr,
	&attr5.attr,
	&attr6.attr,
	&attr7.attr,
	NULL
};

/* NETLINK_GENERIC family definition */
static struct genl_family time_sync_gnl_family = {
	.id = GENL_ID_GENERATE,	/* genetlink should generate an id */
	.hdrsize = 0,
	.name = "TIME_SYNC_GNL",
	.version = VERSION_NR,
	.maxattr = 0,
};

struct time_sync_lock sync_lock;

static void time_sync_free(struct gfar_private *priv)
{
	if (priv->time_sync_priv->time_sync)
		kfree(priv->time_sync_priv->time_sync);
	if (priv->time_sync_priv)
		kfree(priv->time_sync_priv);
	priv->time_sync_priv = NULL;
}

static void time_sync_config_fiper(struct gfar_private *priv)
{
	struct gfar_ptp_attr_t ptp_attr;
	struct device_node *np;
	u32 *tmr_prsc = 0;

	memset(&ptp_attr, 0, sizeof(struct gfar_ptp_attr_t));

	if (gfar_ptp_cal_attr(&ptp_attr))
		return;

	gfar_write(&priv->ptimer->tmr_ctrl,
			gfar_read(&priv->ptimer->tmr_ctrl)
			& ~TMR_CTRL_ENABLE);

	np = of_find_compatible_node(NULL, NULL, "fsl,gianfar-ptp-timer");
	if (np == NULL)	{
		printk(KERN_ERR "1588: Cannot find"
				" gianfar-ptp-timer node\n");
		return;
	}

	tmr_prsc = (u32 *)of_get_property(np, "tmr-prsc", NULL);
	if (tmr_prsc == NULL) {
		printk(KERN_ERR "1588: Cannot find tmr-prsc property\n");
		return;
	}

	gfar_write(&(priv->ptimer->tmr_prsc), *tmr_prsc);
	gfar_write(&(priv->ptimer->tmr_fiper1), priv->ieee_1588_refclk_freq
		 - (ptp_attr.tclk_period >> 16));

	/* Need to mask the TCLK bits as they are initialized with 1 */
	gfar_write(&(priv->ptimer->tmr_ctrl),
		(gfar_read(&(priv->ptimer->tmr_ctrl))
			 & ~TMR_CTRL_TCLK_MASK) | (ptp_attr.tclk_period));
	gfar_write(&(priv->ptimer->tmr_ctrl),
		gfar_read(&(priv->ptimer->tmr_ctrl)) |
		TMR_CTRL_ENABLE | TMR_CTRL_FIPER_START |
		TMR_CTRL_BYPASS_ADDEND);
}

/*
** Function: send_time_stamp()
** send time-stamp, event information over netlink interface
** to the user space listening process on group 1
** How it works:
**     send the time-stamp info ie TMR_ETTS1_H/L and the
**     event info. events can be of type
**	TIME_SYNC_EVT_TIMEINFO_MISS
**	TIME_SYNC_EVT_TIMEINFO_BO
**	TIME_SYNC_EVT_TIMEINFO_RX
*/
static int send_time_stamp(struct time_sync_priv *ts_priv)
{
	struct sk_buff *skb;
	struct nlattr *attr;
	struct time_sync_data *time_sync_data;
	int retval = 0;
	void *msg_head;
	u32 jiffies_delta;

	skb = genlmsg_new(NLMSG_GOODSIZE, GFP_ATOMIC);
	if (!skb) {
		printk(KERN_ERR "Failed to allocate new skb\n");
		retval = -ENOBUFS;
		goto out;
	}

	msg_head = genlmsg_put(skb, 0, 0, &time_sync_gnl_family, 0, 0);
	if (msg_head == NULL) {
		retval = -EINVAL;
		goto out;
	}
	attr = nla_reserve(skb, 0,
			sizeof(struct time_sync_data));
	if (!attr) {
		nlmsg_free(skb);
		retval = -ENOMEM;
		goto out;
	}

	time_sync_data = nla_data(attr);
	if (!time_sync_data) {
		nlmsg_free(skb);
		retval = -ENOMEM;
		goto out;
	}

	/* jiffies_delta is used to track whether timing pulse
	 * is comming for every second or not.
	 * jiffies_delta > HZ means timing pulse is delayed for
	 * more than 1 second and we declare it as
	 * TIME_SYNC_EVT_TIMEINFO_MISS
	 */
	jiffies_delta = jiffies - ts_priv->old_jiffy;
	if (jiffies_delta > HZ) {
		if (ts_priv->time_sync->time_info.event !=
			TIME_SYNC_EVT_TIMEINFO_BO)
			ts_priv->time_sync->time_info.event =
				TIME_SYNC_EVT_TIMEINFO_MISS;
	}
	time_sync_data->time_info.event =
		ts_priv->time_sync->time_info.event;
	time_sync_data->time_info.time_cnt.high =
		ts_priv->time_sync->time_info.time_cnt.high;
	time_sync_data->time_info.time_cnt.low =
		ts_priv->time_sync->time_info.time_cnt.low;

	/* finalize the message */
	genlmsg_end(skb, msg_head);

	ts_priv->old_jiffy = jiffies;

	/* send the message to user space netlink reciever's */
	genlmsg_multicast(skb, 0, TIME_SYNC_NL_GROUP, GFP_ATOMIC);

out:
	return retval;
}

static int register_1588_interrupt(struct net_device *dev)
{
	struct device_node *np;
	struct gfar_private *priv = netdev_priv(dev);
	int retval = -EINVAL;

	np = of_find_compatible_node(NULL, NULL,
				"fsl,gianfar-ptp-timer");
	if (np == NULL) {
		printk(KERN_ERR "unable to find"
				" \"fsl,gianfar-ptp-timer \"\n");
		goto err;
	}

	priv->time_sync_priv->irq = irq_of_parse_and_map(np, 0);
	if (priv->time_sync_priv->irq == NO_IRQ) {
		printk(KERN_ERR "unable to find irq for"
				" fsl,gianfar-ptp-timer\n");
		goto err;
	}
	retval = request_irq(priv->time_sync_priv->irq, time_sync_isr,
				0, "time-sync-1588", priv);
	if (retval < 0) {
		printk(KERN_ERR "unable to register irq handler for 1588\n");
		goto err;
	}
err:
	return retval;
}

int process_time_sync_init(struct net_device *dev,
		struct time_sync_data *time_sync_data)
{
	int retval = 0;
	struct gfar_private *priv = netdev_priv(dev);
	struct time_sync_priv *time_sync_priv = priv->time_sync_priv;

	if (time_sync_priv) {
		printk(KERN_INFO "time sync already initialized \r\n");
		retval = -EEXIST;
		goto err;
	}

	time_sync_priv = kzalloc(sizeof(struct time_sync_priv), GFP_KERNEL);
	if (!time_sync_priv) {
		printk(KERN_ERR "Unable to allocate time_sync_priv\n");
		retval = -ENOMEM;
		goto err;
	}
	init_waitqueue_head(&time_sync_priv->wait_h);
	time_sync_priv->awaiting_first_gps = TRUE;
	time_sync_priv->time_sync = kzalloc(sizeof(struct
				time_sync_data), GFP_KERNEL);
	if (!time_sync_priv->time_sync) {
		printk(KERN_ERR "Unable to allocate time_sync\n");
		retval = -ENOMEM;
		kfree(time_sync_priv);
		goto err;
	}
	memcpy(time_sync_priv->time_sync, time_sync_data,
				sizeof(struct time_sync_data));
	priv->time_sync_priv = time_sync_priv;

	/* FIXME: We cannot disable alarms; so set alarm to a
	   very long time in future */
	gfar_write(&(priv->ptimer->tmr_alarm1_l), 0xFFFFFFFF);
	gfar_write(&(priv->ptimer->tmr_alarm1_h), 0xFFFFFFFF);

	gfar_write(&priv->ptimer->tmr_ctrl,
			gfar_read(&priv->ptimer->tmr_ctrl)
			& ~TMR_CTRL_ENABLE);

	if (time_sync_sysfs_init(priv) != TRUE) {
		printk(KERN_ERR "Unable to initialize sysfs for time-sync\n");
		retval = -EINVAL;
		time_sync_free(priv);
		goto err;
	}

err:
	return retval;
}

int process_time_sync_stop(struct net_device *dev)
{
	int retval = 0;
	struct gfar_private *priv = netdev_priv(dev);

	if (!priv->time_sync_priv) {
		printk(KERN_DEBUG "time sync not initialized \r\n");
		retval = -EINVAL;
		goto err;
	}
	gfar_write(&priv->ptimer->tmr_ctrl,
			gfar_read(&priv->ptimer->tmr_ctrl)
			& ~TMR_CTRL_ENABLE);

	free_irq(priv->time_sync_priv->irq, priv);
	time_sync_sysfs_exit(priv->time_sync_priv->time_sync_kobj);
	time_sync_free(priv);

	/* generic netlink is used for sending the
	 * current time stamps to user process
	 */
	if (genl_unregister_family(&time_sync_gnl_family)) {
		printk(KERN_ERR "Unable to unregister time_sync"
			" with NETLINK_GENERIC framework");
		retval = -EINVAL;
		goto err;
	}
err:
	return retval;
}

/*
** Function: time_sync_enable_fiper()
** configures alarm1 to fire after 1 second wrt to the current time
** How it works:
**     Reads the current time from hw
**     Adds the timeout ie (1 second)
**     Programs the above time (in future) into hw
** Assumption: Fiper is already configured to fire at
** a boundary of 1 second.
*/
void time_sync_enable_fiper(struct gfar_private *priv,
				u32 timeout /* seconds */)
{
	u64 time_counter;
	u32 tmp_lsb_counter;

	tmp_lsb_counter = gfar_read(&(priv->ptimer->tmr_cnt_l));
	time_counter = gfar_read(&(priv->ptimer->tmr_cnt_h));
	time_counter = time_counter << 32;
	time_counter |= tmp_lsb_counter;

	time_counter += multi64_oper(priv->ieee_1588_refclk_freq, timeout);

	gfar_write(&(priv->ptimer->tmr_alarm1_l), (u32) time_counter);
	gfar_write(&(priv->ptimer->tmr_alarm1_h), (u32) (time_counter >> 32));
}

int time_sync_start(struct net_device *dev, struct ifreq *ifr)
{
	int retval = 0;
	struct gfar_private *priv = netdev_priv(dev);
	struct time_sync_data *time_sync;

	if (register_1588_interrupt(dev) != 0) {
		printk(KERN_ERR "1588: Cannot register 1588 ISR\n");
		retval = -EINVAL;
		return retval;
	}
	time_sync = priv->time_sync_priv->time_sync;

	switch (time_sync->sync_source) {
	case PPS_SRC_GPS:
		gfar_write(&(priv->ptimer->tmr_temask),
			TMR_TEMASK_ETS1 | /* enables IRQ for PPS event */
			TMR_TEMASK_ALM1 | /* causes FIPER to start */
			TMR_TEMASK_ALM2 | /* causes blackout timer to fire */
			TMR_TEMASK_PP1);  /* causes PPS1 generation */

		/* required for at least first pulse */
		time_sync_config_blackout(priv, /* configure BO */
			priv->time_sync_priv->time_sync->black_out_duration);
		gfar_write(&priv->ptimer->tmr_ctrl,
			gfar_read(&priv->ptimer->tmr_ctrl)
			| TMR_CTRL_BYPASS_ADDEND | TMR_CTRL_ENABLE);
		break;

	case PPS_SRC_PTP:
		/* timing pulse not supported. Only fiper will be enabled */
		gfar_write(&(priv->ptimer->tmr_temask),
			TMR_TEMASK_ALM1 |
			TMR_TEMASK_PP1);
		time_sync_config_fiper(priv);
		time_sync_enable_fiper(priv, FIPER_START_DELAY);
		gfar_write(&priv->ptimer->tmr_ctrl,
			gfar_read(&priv->ptimer->tmr_ctrl)
			| TMR_CTRL_BYPASS_ADDEND | TMR_CTRL_ENABLE);
		break;

	default:
		printk(KERN_ERR "Unsupported sync-source (%d) in ioctl",
				time_sync->sync_source);
		retval = -EINVAL;
	}

	return retval;
}

int do_time_sync_ioctl(struct net_device *dev, struct ifreq *ifr)
{
	struct time_sync_data time_sync_data;
	int retval = 0;

	memset(&time_sync_data, 0, sizeof(struct time_sync_data));
	mutex_lock_interruptible(&sync_lock.lock);

	if (0 != copy_from_user(&time_sync_data, ifr->ifr_data,
				sizeof(struct time_sync_data))) {
		retval = -EFAULT;
		goto err;
	}

	switch (time_sync_data.ioctl_cmd) {
	case TIMING_SYNC_INIT:
		retval = process_time_sync_init(dev, &time_sync_data);
		break;
	case TIMING_SYNC_STOP:
		retval = process_time_sync_stop(dev);
		break;
	case TIMING_SYNC_START:
		retval = time_sync_start(dev, ifr);
		break;
	default:
		printk(KERN_ERR "Invalid ioctl command"
				" %d\n", time_sync_data.ioctl_cmd);
		retval = -EINVAL;
	}

err:
	mutex_unlock(&sync_lock.lock);
	return retval;
}

/*
** This ISR is invoked on rx of timing pulse (1pps)
** When this ISR executes first time, it does a few special steps:
**    ISR configures FIPER for PPS_OUT to AIC; and hence Alarm1 is also
** configured (so that FIPER can be started at right moment).
** On every regular pulse, it (re)configures/(re)activates blackout
** timer, extracts the time-info (the momen when pps was given to hw),
** gives wakeup to the entities that need 1pps event.
**
*/
static irqreturn_t time_sync_isr(int irq, void *arg)
{
	struct gfar_private *priv = (struct gfar_private *)arg;
	struct time_sync_priv *time_sync_priv = priv->time_sync_priv;
	struct timespec blackout_time;
	if ((priv->ptimer->tmr_tevent & TMR_TEVENT_ETS1) &&
			(priv->ptimer->tmr_temask & TMR_TEMASK_ETS1)) {
		/* PPS event detected */
		gfar_write(&(priv->ptimer->tmr_tevent), TMR_TEVENT_ETS1);
		if (time_sync_priv->awaiting_first_gps == TRUE) {
			time_sync_priv->awaiting_first_gps = FALSE;
			time_sync_config_fiper(priv);
			time_sync_cfg_lte_delay_enable_fiper(priv,
				FIPER_START_DELAY,
				time_sync_priv->time_sync->lte_delay);
		}
		/* FIXME: we can avoid blackout cfg at every PPS */
		time_sync_config_blackout(priv,
			time_sync_priv->time_sync->black_out_duration);
		time_sync_priv->time_sync->time_info.time_cnt.low =
				gfar_read(&(priv->ptimer->tmr_etts1_l));
		time_sync_priv->time_sync->time_info.time_cnt.high =
				gfar_read(&(priv->ptimer->tmr_etts1_h));
		time_sync_priv->time_sync->time_info.event =
						TIME_SYNC_EVT_TIMEINFO_RX;
		time_sync_priv->stats.gps_cnt++;
		send_time_stamp(time_sync_priv);
	} else if ((priv->ptimer->tmr_tevent & TMR_TEVENT_ALM2) &&
			(priv->ptimer->tmr_temask & TMR_TEMASK_ALM2)) {
		/* Blackout event detected */
		gfar_write(&(priv->ptimer->tmr_tevent), TMR_TEVENT_ALM2);
		time_sync_priv->time_sync->time_info.event =
				TIME_SYNC_EVT_TIMEINFO_BO;
		time_sync_priv->stats.alarm2_cnt++;
		getnstimeofday(&blackout_time);
		memcpy(&time_sync_priv->blackout_time, &blackout_time,
						sizeof(struct timespec));
		send_time_stamp(time_sync_priv);
	}

	if ((priv->ptimer->tmr_tevent & TMR_TEVENT_ALM1) &&
			(priv->ptimer->tmr_temask & TMR_TEMASK_ALM1)) {
		/* this event is only for debug */
		gfar_write(&(priv->ptimer->tmr_tevent), TMR_TEVENT_ALM1);
		time_sync_priv->stats.alarm1_cnt++;
	}

	if ((priv->ptimer->tmr_tevent & TMR_TEVENT_PP1) &&
			(priv->ptimer->tmr_temask & TMR_TEMASK_PP1)) {
		/* this event is only for debug */
		gfar_write(&(priv->ptimer->tmr_tevent), TMR_TEVENT_PP1);
		time_sync_priv->stats.fiper1_cnt++;
	}
	return IRQ_HANDLED;
}

static ssize_t time_sync_show_sysfs(struct kobject *kobj,
				struct attribute *attr, char *buf)
{
	struct time_sync_sysfs *a = container_of(attr,
				struct time_sync_sysfs, attr);
	struct time_sync_priv priv;
	int retval = 0;

	memset(&priv, 0, sizeof(struct time_sync_priv));

	retval = time_sync_priv(&priv);
	if (retval != 0) {
		printk(KERN_ERR "Unable to found time_sync priv\n");
		return -EINVAL;
	}

	switch (a->cmd_type) {
	case BLACK_OUT_DURATION:
		return scnprintf(buf, BUFFER_SIZE, "%d %s\n",
				priv.time_sync->black_out_duration, "seconds");
	case TIME_SYNC_SOURCE:
		switch (priv.time_sync->sync_source) {
		case PPS_SRC_GPS:
			return scnprintf(buf, BUFFER_SIZE,
					"%s\n", GPS_PPS);
		case PPS_SRC_RAW_GPS:
			return scnprintf(buf, BUFFER_SIZE,
					"%s\n", RAW_GPS);
		case PPS_SRC_PTP:
			return scnprintf(buf, BUFFER_SIZE,
					"%s\n", PTP_PPS);
		default:
			return -EINVAL;
		}
	case CORRECTION_MODE:
		if (priv.time_sync->correction_mode == TMSYN_CORR_MODE_DAC) {
			return scnprintf(buf, BUFFER_SIZE, "%s\n",
						DAC_CORRECTION);
			}
		else if (priv.time_sync->correction_mode ==
					TMSYN_CORR_MODE_ADDEND)
			return scnprintf(buf, BUFFER_SIZE, "%s\n",
						ADDEND_CORRECTION);
		else if (priv.time_sync->correction_mode ==
					TMSYN_CORR_MODE_DAC_ADDEND)
			return scnprintf(buf, BUFFER_SIZE, "%s\n",
						DAC_ADDEND_CORRECTION);
		else
			return scnprintf(buf, BUFFER_SIZE, "%s\n",
						"undefined mode");
	case LTE_DELAY:
		return scnprintf(buf, BUFFER_SIZE, "%d %s\n",
				priv.time_sync->lte_delay % 1000000,
				"micro seconds");
	case LAST_BLACKOUT_OCCURENCE:
		return scnprintf(buf, BUFFER_SIZE, "%lu %s %lu %s\n",
			(unsigned long)priv.blackout_time.tv_sec, "secs",
			(unsigned long)priv.blackout_time.tv_nsec / 1000,
					"usec");
	case NEW_MODE_AFTER_BLACKOUT:
		/*TBD*/
		break;
	case STATS:
		return scnprintf(buf, BUFFER_SIZE,
				"%s = %d\n%s = %d\n%s = %d\n%s = %d\n",
				"gps_cnt", priv.stats.gps_cnt, "fiper1_cnt",
				priv.stats.fiper1_cnt, "alarm1_cnt",
				priv.stats.alarm1_cnt, "alarm2_cnt",
				priv.stats.alarm2_cnt);
		break;
	default:
		return -EINVAL;
	}
	return retval;
}

static ssize_t time_sync_show_store(struct kobject *kobj,
			struct attribute *attr,
			const char *buf, size_t len)
{
	struct time_sync_sysfs *a = container_of(attr,
			struct time_sync_sysfs, attr);
	int bytes_written;

	bytes_written = snprintf(a->data, BUFFER_SIZE, "%s", buf);
	return bytes_written;
}

static struct sysfs_ops time_sync_ops = {
	.show = time_sync_show_sysfs,
	.store = time_sync_show_store,
};

static struct kobj_type time_sync_type = {
	.sysfs_ops = &time_sync_ops,
	.default_attrs = time_sync_attr,
};

static int time_sync_sysfs_init(struct gfar_private *priv)
{
	struct kobject *time_sync_kobj;
	int err = TRUE;
	time_sync_kobj = kzalloc(sizeof(*time_sync_kobj), GFP_KERNEL);
	if (time_sync_kobj) {
		kobject_init(time_sync_kobj, &time_sync_type);
		if (kobject_add(time_sync_kobj, NULL, "%s", "time_sync")) {
			printk(KERN_ERR "Sysfs creation failed\n");
			kobject_put(time_sync_kobj);
			err = -EINVAL;
			goto out;
		}
	} else {
		printk(KERN_ERR "Unable to allocate time_sync_kobj\n");
		err = -ENOMEM;
		goto out;
	}
	priv->time_sync_priv->time_sync_kobj = time_sync_kobj;

	err = genl_register_family(&time_sync_gnl_family);
	if (err)
		printk(KERN_ERR "time_sync registration with"
			" NTELINK_GENERIC framework failed\n");

out:
	return err;
}

static void time_sync_sysfs_exit(struct kobject *time_sync_kobj)
{
	if (time_sync_kobj) {
		kobject_put(time_sync_kobj);
		kfree(time_sync_kobj);
	}
}

int time_sync_priv(struct time_sync_priv *time_sync_priv)
{
	struct gfar_private *priv;
	struct of_device *ofdev;
	struct device_node *np = NULL;
	struct time_sync_priv *time_sync;

	while (1) {
		np = of_find_compatible_node(NULL, NULL, ETSEC_NODE);
		if (np == NULL)	{
			printk(KERN_ERR "1588: Cannot find eTSEC. Check"
				" eTSEC node in dts \r\n");
			return -EINVAL;
		}
		ofdev = of_find_device_by_node(np);
		priv = dev_get_drvdata(&ofdev->dev);
		time_sync = priv->time_sync_priv;
		if (time_sync)
			break;
	}
	memcpy(time_sync_priv, time_sync, sizeof(struct time_sync_priv));
	return 0;
}

/*
** Function: time_sync_config_blackout()
** (re)configures and (re)activates blackout timer
** How it works:
**     Reads the current time from hw
**     Adds the timeout
**     Programs the above time (in future) into hw
** Assumption: TEMASK already has ALM2 enabled
*/
void time_sync_config_blackout(struct gfar_private *priv,
				u32 timeout /* seconds */)
{
	u64 time_counter;
	u32 tmp_lsb_counter;

	tmp_lsb_counter = gfar_read(&(priv->ptimer->tmr_cnt_l));
	time_counter = gfar_read(&(priv->ptimer->tmr_cnt_h));
	time_counter = time_counter << 32;
	time_counter |= tmp_lsb_counter;

	time_counter += multi64_oper(priv->ieee_1588_refclk_freq, timeout);

	gfar_write(&(priv->ptimer->tmr_alarm2_l), (u32) time_counter);
	gfar_write(&(priv->ptimer->tmr_alarm2_h), (u32) (time_counter >> 32));
}

/*
** Function: time_sync_cfg_lte_delay_enable_fiper()
** Configures LTE delay and enables FIPER
*/
void time_sync_cfg_lte_delay_enable_fiper(struct gfar_private *priv,
				u32 fiper_delay /* seconds */,
				u32 lte_delay /* us */)
{
	u64 time_counter, etts_time = 0ll, temp;
	u32 tmp_lsb_counter, quotient; /*We are not using quotient varible*/

loop:
	tmp_lsb_counter = gfar_read(&(priv->ptimer->tmr_etts1_l));
	etts_time = gfar_read(&(priv->ptimer->tmr_etts1_h));
	etts_time = etts_time << 32;
	etts_time |= tmp_lsb_counter;
	while (gfar_read(&(priv->ptimer->tmr_stat)) & TMR_STAT_ETS1_VLD) {
		/* ignoring stale info */
		mdelay(100);
		goto loop;
	}

	time_counter = etts_time + fiper_delay * priv->ieee_1588_refclk_freq;
	temp = multi64_oper(priv->ieee_1588_refclk_freq, lte_delay);
	time_counter += div64_oper(temp, 1000000, &quotient);

	gfar_write(&(priv->ptimer->tmr_alarm1_l), (u32) time_counter);
	gfar_write(&(priv->ptimer->tmr_alarm1_h), (u32) (time_counter >> 32));

	return;
}

void gfar_1588_time_sync_init()
{
	mutex_init(&sync_lock.lock);
}
