/*
 * drivers/rf/controllers/fsl_aic.h
 * Freescale AIC (Antenna Controller Interface) driver
 *
 * AIC is the antenna interface controller found in bsc913x
 * family of SOCs. It supports NCDMA, WCDMA-FDD, LTE-FDD, LTE-TDD
 * and GSM-SNIFF network modes. AIC has 6 lanes on which RFICs
 * can be connected. And AIC supports 4 RFICs working simultaneously.
 * This driver provides only configuration path for all other modes
 * except SNIFF because data from RFIC through AIC goes directly
 * to a hardware accelerator from where it is taken out by DSP core
 * in bsc913x.
 *
 * Author: pankaj chauhan <pankaj.chauhan@freescale.com>
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#ifndef __FSL_AIC_H__
#define __FSL_AIC_H__

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/device.h>
#include <linux/spinlock.h>

struct aic_common_regs {

	u32	aic_ip_version;
	u32	aic_interconnect;
	u32	pll_cntl_0;
	u32	pll_cntl_1;
	u32	aic_sniff0_refcnt;
	u32	aic_sniff0_refcnt_off;
	u32	aic_sniff0_capt_off;
	u32	aic_sniff0_capt_dur;
	u32	aic_sniff_bcar;
	u32	aic_sniff_bcbr;
	u32	aic_sniff_brm_cntrl;
	u32	aic_sniff0_cntrl;
	u32	aic_sniff0_inten;
	u32	aic_sniff0_int_stat;
	u32	aic_sniff0_refcnt_val;
	u32	aic_sniff0_frmint;
	u32	reserved[6];
	u32	aic_sniff1_refcnt;
	u32	aic_sniff1_refcnt_off;
	u32	aic_sniff1_capt_off;
	u32	aic_sniff1_capt_dur;
	u32	reserved1[3];
	u32	aic_sniff1_cntrl;
	u32	aic_sniff1_inten;
	u32	aic_sniff1_int_stat;
	u32	aic_sniff1_refcnt_val;
	u32	aic_sniff1_frmint;
	u32	reserved2[52];
	u32	maxim_rxcntl;
	u32	maxim_cntl;
	u32	maxim_rssi;
	u32	maxim_txcoeff_0_1;
	u32	maxim_txcoeff_2_3;
	u32	maxim_txcoeff_4_5;
	u32	maxim_txcoeff_6_7;
	u32	maxim_txcoeff_8_9;
	u32	maxim_txcoeff_10_11;
	u32	maxim_txcoeff_12_13;
	u32	maxim_txcoeff_14_15;
	u32	maxim_txcoeff_16_17;
	u32	maxim_status;
	u32	maxim_dlfrac_delay;
	u32	maxim_spi_rxgain_1;
	u32	maxim_spi_rxgain_2;
	u32	reserved3[6];
	u32	jesd0_cntl;
	u32	jesd0_status;
	u32	jesd1_cntl;
	u32	jesd1_status;
	u32	jesd2_cntl;
	u32	jesd2_status;
	u32	aic_tm_cntrl1;
	u32	aic_interrupt_mux_ctrl_reg;
	u32	reserved4[4];
	u32	ppc_interrupt_ctrl_reg;
	u32	ppc_interrupt_status_reg;
	u32	dsp_lte_interrupt_ctrl_reg;
	u32	dsp_lte_interrupt_status_reg;
	u32	reserved5[4];
};

/*INTERCONNECT*/
#define GPS_SNIFF_MASK			0x000000c0
#define GPS_SNIFF_GPS_PPS_IN		0x00000000
#define GPS_SNIFF_GPS_PPS_REFCLK_IN	0x00000040
#define GPS_SNIFF_GPS_ETSEC_PULSE	0x00000080
#define GPS_SNIFF_RESVD			0x000000c0
#define AGC_STROBE_MASK			0x3
#define AGC_STROBE_1_SHIFT		8
#define AGC_STROBE_2_SHIFT		10

/*JESDCNTL*/
#define SINGLE_PORT		0x00000001
#define HALF_DUPLEX		0x00000002
#define SINGLE_DATA_RATE	0x00000004
#define INVERT_CLK		0x00000020
#define INVERT_RXFRAME		0x00000040
#define RXF_NEGATIVE_EDGE_EN	0x00000400
#define TXF_NEGATIVE_EDGE_EN	0x00000800
#define RXF_TIMEOUT_EN		0x00010000
#define ADI_MCLK_EN		0x00100000
#define JESD_10BITS_WIDTH_EN	0x40000000
#define JESD_MODE_EN		0x80000000
#define	RDELAY_SHIFT		7
#define RDELAY_MASK		(0x3 << RDELAY_SHIFT)
#define RDELAY_0_CYCL		(0x0 << RXDATA_DELAY_SHIFT)
#define RDELAY_1_by_2_CYCL	(0x1 << RXDATA_DELAY_SHIFT)
#define RDELAY_1_CYCL		(0x2 << RXDATA_DELAY_SHIFT)
#define RDELAY_3_by_2_CYCL	(0x3 << RXDATA_DELAY_SHIFT)
#define TX_RX_MODE_SHIFT	3
#define TXRX_MODE_MASK		(0x3 << TX_RX_MODE_SHIFT)
#define TXRX_MODE_1T1R		(0x0 << TX_RX_MODE_SHIFT)
#define TXRX_MODE_1T2R		(0x2 << TX_RX_MODE_SHIFT)
#define TXRX_MODE_2T2R		(0x1 << TX_RX_MODE_SHIFT)

/*AICTMCNTRL1*/
/*These values will be shifted right
 * by lane id before writing to AICTMCNTRL1
 */
#define TMCTRL_DL_EN	0x00100000
#define TMCTRL_UL_EN	0x04000000

/*PPC_INTERRUPT_CNTL_REG*/
#define ADILANE0_DL_TTI		0x10
#define ADILANE_DL_TTI_SHIFT	5
#define IE_PPC_SNIFF0_CAPT_DONE_INT	0x04000000
#define IE_PPC_SNIFF1_CAPT_DONE_INT	0x08000000

/* PPC INTERRUPT STATUS REGISTER */
#define PPC_ISR_SNIFF0_CAPT_DONE_INT	0x04000000
#define PPC_ISR_SNIFF1_CAPT_DONE_INT	0x08000000

struct aic_lane_regs {
	u32	aic_dma_riqmts;
	u32	aic_dma_tiqmts;
	u32	aic_dma_riqmpl1;
	u32	aic_dma_riqmpl23;
	u32	aic_dma_tiqmpl1;
	u32	aic_dma_tiqmpl23;
	u32	aic_dma_mss;
	u32	aic_dma_riqba0;
	u32	aic_dma_riqba1;
	u32	reserved[7];
	u32	aic_dma_riqbs;
	u32	aic_dma_rmba;
	u32	aic_dma_rmbs;
	u32	aic_dma_tiqba0;
	u32	aic_dma_tiqba1;
	u32	reserved1[8];
	u32	aic_dma_tiqbs;
	u32	aic_dma_dmsr;
	u32	aic_dma_dcr;
	u32	aic_dma_riqt;
	u32	aic_dma_riqft;
	u32	aic_dma_riqst;
	u32	aic_dma_tiqt;
	u32	aic_dma_tiqft;
	u32	aic_dma_tiqst;
	u32	aic_dma_ier;
	u32	reserved2[19];
	u32	aic_frame_conf;
	u32	aic_netw_conf1;
	u32	aic_netw_conf2;
	u32	aic_netw_conf3;
	u32	aic_dlul_delay;
	u32	aic_lane_dlslot;
	u32	aic_lane_ulslot;
	u32	aic_lane_spslot;
	u32	aic_lane_frame_count;
	u32	aic_lane_tmctrl;
	u32	reserved6[3];
	u32	aic_symconf0;
	u32	aic_symconf1;
	u32	aic_symconf2;
	u32	aic_symconf3;
	u32	aic_symconf4;
	u32	aic_symconf5;
	u32	aic_symconf6;
	u32	aic_symconf7;
	u32	aic_int_pretime;
	u32	aic_ref_framconf;
	u32	aic_lane_frame_rollover;
	u32	reserved3;
	u32	aic_dma_isr;
	u32	aic_dma_miar;
	u32	aic_dma_dsr;
	u32	aic_dma_riqbdr;
	u32	aic_dma_tiqbdr;
	u32	reserved4[4];
	u32	rftimer_intr_ctrl;
	u32	rftimer_isr;
	u32	reserved5[34];
};

/*AICFRAMCONF*/
#define SUBFRM_PER_FRM_SHIFT	21
#define SUBFRM_PER_FRM_MASK	0x3f
#define SLOT_PER_SUBFRM_SHIFT	15
#define SLOT_PER_SUBFRM_MASK	0x1f
#define CHIPS_PER_SLOT_SHIFT	0
#define CHIPS_PER_SLOT_MASK	0x7fff

/*AICNETWCONF*/
#define TOTAL_LEN_SHIFT		16
#define	TOTAL_LEN_MASK		0xffff
#define SYM_PER_SLOT_SHIFT	0
#define SYM_PER_SLOT_MASK	0xff
#define PPS_TRIG_PPS		0x00000000
#define PPS_TRIG_PTP		0x00001000
#define PPS_TRIG_NLM		0x00002000
#define PPS_TRIG_RAW_PPS	0x00003000
#define DL_EN			0x00008000
#define UL_EN			0x00004000

/*AICNETWCONF2*/
#define DWPTS_SHIFT		0
#define DWPTS_MASK		0x7fff
#define UPPTS_SHIFT		16
#define UPPTS_MASK		0xfff

/*AICNETWCONF3*/
#define GP_SHIFT		0
#define GP_MASK			0x7fff

/*DLULDELAY*/
#define DL_DELAY_SHIFT		16
#define DL_DELAY_MASK		0xffff
#define UL_DELAY_SHIFT		0
#define UL_DELAY_MASK		0xffff

/*AICLANETMCTRL*/
#define GPS_EN			0x00000001
#define SYNC_EN			0x00000002
#define LTE_EN			0x00000004
#define REFCLK_SEL		0x01000000
#define AGC_STRB_SEL_MASK	0xc0000000
#define AGC_STRB_SEL_SLOT	0x00000000
#define AGC_STRB_SEL_TTI	0x40000000
#define AGC_STRB_SEL_TTI_RLD	0x80000000
#define AGC_STRB_SEL_NEXT_SYM	0xc0000000

/* AICDMAnDMSR */
#define RDD_UNI_SYS_MEM		0x00000000
#define RDD_UNI_MAPLE		0x00000080
#define RDD_MULTICAST		0x00000100
#define RDD_MASK		0xfffffe7f
#define DBL_BUF_MODE		0x00000200
#define RX_OVERSAMPLING		0x00000020
#define TNA_2ANT		0x00010000
#define RNA_2ANT		0x00000001
#define RSW_16B			0x00000010
#define RXNOACK			0x00000008
#define RDNSZ_TRUNC		0x00000400
#define RXENSEL_SNIFFER		0x00000800
#define TXNOACK			0x00080000
#define TX_FLW_CTRL_DIS		0x00100000
#define TDNSZ_TRUNC		0x04000000

#define RIQBS_MASK		0x00ffffff
#define RIQBA_MASK		0xfffffff0
#define RIQT_MASK		0x00ffffff

/* AICDMAnDCR */
#define TIQE			0x00010000
#define RIQE			0x00000001

/* AICDMAnDSR */
#define TIQS			0x00010000
#define RIQS			0x00000001

/* AICDMAnRIQMTS */
#define RIQMTS_MASK		0x00000003
#define RIQMTS_64B		0x00000000
#define RIQMTS_128B		0x00000001
#define RIQMTS_256B		0x00000002

/* AICDMAnIER */
#define ANT0_RIQTIE		0x00000001
#define ANT1_RIQTIE		0x00000002
#define RIQFTIE			0x00002000
#define RIQSTIE			0x00004000
#define RIQOIE			0x00008000
#define ANT0_TIQTIE		0x00010000
#define ANT1_TIQTIE		0x00020000
#define TIQFTIE			0x20000000
#define TIQSTIE			0x40000000
#define TIQUIE			0x80000000

/* AICDMAnISR */
#define ANT0_RIQTI		0x00000001
#define ANT1_RIQTI		0x00000002
#define RIQFTI			0x00002000
#define RIQSTI			0x00004000
#define RIQOI			0x00008000
#define ANT0_TIQTI		0x00010000
#define ANT1_TIQTI		0x00020000
#define TIQFTI			0x20000000
#define TIQSTI			0x40000000
#define TIQUI			0x80000000

/*AICDMADSR*/
#define RIQS			0x00000001
#define TIQS			0x00010000

/* CHIPRATE_REFCLK is chiprate in refclk terms, it is calculated as:
 * chiprate_refclk = chips_per_slot * (REFCLK/SAMPLING_FREQUENCY)
 *
 * REFCLK - 19.2 Mhz
 * chips_per_slot, and SAMPLING frequency changes with bandwidth
 * and the ration remains same. so for all bandwidths chiprate_refclk
 * remains constant
 */
#define CHIPRATE_REFCLK		9600

/* Default rollover value is taken as 19200000
 * so as to cover a period of 1 second */
#define DEFAULT_ROLLOVER	19200000
#define DEFAULT_CAPT_OFF		0

#define MAX_LANE_COUNT		6
#define LANE_REGDUMP_OFFSET	0x200
#define SYNC_WAIT		20

#define DEFAULT_DL_DELAY	0x15DC
#define DEFAULT_UL_DELAY	0x15DC

#define AIC_NUM_SNIFFER_BLKS	2
#define AIC_MAX_NUM_CHANNELS	2

struct aic_sniff_regs {
	u32 aic_sniff_refcnt;
	u32 aic_sniff_refcnt_off;
	u32 aic_sniff_capt_off;
	u32 aic_sniff_capt_dur;
	u32 reserved[3];
	u32 aic_sniff_ctrl;
	u32 aic_sniff_inten;
	u32 aic_sniff_int_stat;
	u32 aic_sniff_refcnt_val;
	u32 aic_sniff_frmint;
};

/*SNIFFCTRL*/
#define REF_SEL_MASK	0x00000003
#define REF_SEL_ADI	0x00000000
#define REF_SEL_MAXPHY	0x00000001
#define REF_SEL_PREMUL	0x00000002
#define CHP_SEL_MASK	0x0000000c
#define CHP_SEL_SHIFT	2
#define CHP_SEL_MAXPHY	0x0000000c
#define REFCNT_EN	0x00000010
#define	GPS_SYNC	0x00000020
#define LD_OFF		0x00000040
#define CAP_EN		0x00000080
#define PRE_CAP_EN	0x00000100

/* AICSNIFFnINTEN */
#define CAP_INT_EN	0x00000001
#define SNIF_FRMINT_EN	0x00000002

/* AICSNIFFnINT_STAT */
#define CAP_INT_STAT		0x00000001
#define SNIF_FRMINT_STAT	0x00000002

struct aic_sniffer {
	struct aic_lane *lane;
	atomic_t in_use;
	unsigned int id;
	unsigned int intialized;
	unsigned int gps_synced;
	unsigned int capt_dur_frms;
	struct aic_sniff_regs *regs;
	spinlock_t lock;
	int capture_done;
};

struct aic_dev {
	unsigned int sniff_irq;
	unsigned int dsp_general_irq;
	atomic_t sniffers_in_use;
	enum rf_timer_src sniff_timing_src;
	struct aic_common_regs *regs;
	struct aic_lane *lanes[6];
	struct aic_sniffer sniffers[2];
	spinlock_t lock;		/*Protects aic_dev->regs*/
	struct device *dev;
	u32 *gir_ant_en_mpic;
	u32 *gir_ant;
};

/* GIR_ANT_EN_MPIC */
#define GIR_ANT_OFFSET		0x140
#define GIR_ANT_EN_MPIC_OFFSET	0x144
#define AIC_IRQ_ANT29			0x10000000

#define DMA_STATUS_RETRIES	2

struct aic_lane_stats {
	int dl_tti_count;
	int ul_tti_count;
	int rx_ant0_frames;
	int rx_ant1_frames;
};

enum aic_lane_type {
	AIC_LANE_ADI,
	AIC_LANE_MAXIM
};
struct aic_lane {
	struct aic_dev *aic;
	struct rf_ctrl_dev *rf_dev;
	struct aic_sniffer *sniffer;
	enum aic_lane_type type;
	unsigned int id;
	unsigned int irq;
	struct aic_lane_regs *regs;
	struct aic_lane_stats stats;
	unsigned int num_chan;
	u32	rx_dma_done;
	u32	rx_dma_done_mask;
	int rx_ant0_idx;
	int rx_ant1_idx;
	spinlock_t lock;		/*Protects aic_lane->regs*/
};

#define AIC_DMA_DONE	0x1
#endif
