/*
 * drivers/rf/core/rf_channel.c
 * RF Data Channel support
 *
 * RF data channels: Each RF device has per antenna RF
 * data channel. An RF data channel is a full duplex channel
 * (supports both Tx and RX), which supports Tx and Rx of IQ
 * data to/from a RF device. Current Implementation only
 * supports Rx/Tx data between rf device and user space but
 * this framework can be extended to support in kernel users
 * of IQ data as well.
 *
 * Author: pankaj chauhan <pankaj.chauhan@freescale.com>
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/bitops.h>
#include <linux/rfdev.h>
#include <linux/rfdev_ioctl.h>
#include <linux/rf_channel.h>

int rf_channel_open(struct rf_ctrl_dev *rf_dev,
	struct rf_channel_params *chan_params)
{

	struct rf_channel *chan;
	struct rf_mmaped_bufs *bufs;
	struct rf_ctrl_ops *ops;
	int ant, rc = 0;

	ant = chan_params->ant;
	if (ant >= MAX_NUM_ANTENNAS) {
		pr_err("%s: Ant[%d] invalid\n", rf_dev->name,
			ant);
		rc = -EINVAL;
		goto out;
	}
	chan = rf_dev->channels[ant];
	if (chan) {
		pr_err("%s: Ant[%d] channel is already open\n", rf_dev->name,
			ant);
		rc = -EEXIST;
		goto out;
	}

	chan = kzalloc(sizeof(struct rf_channel), GFP_KERNEL);
	if (!chan) {
		rc = -ENOMEM;
		goto out;
	}

	raw_spin_lock_init(&chan->lock);
	chan->ant = ant;
	chan->dev = rf_dev;
	chan->flags = chan_params->flags;
	chan->tx_buf_size = chan_params->tx_buf_size;
	chan->rx_buf_size = chan_params->rx_buf_size;

	if (chan_params->flags & RF_CHAN_MMAPED) {
		bufs = &chan->bufs.mmap_bufs;
		/* We don't touch the buffers in kernel space
		 * since we don't have any in kernel users of
		 * channel, so virtual address of bufs is being
		 * kept NULL. If we ever have users who can touch
		 * channel buffer in kernel then an ioremap() is
		 * required here.
		 */
		bufs->tx_base = 0;
		bufs->rx_base = 0;
		bufs->tx_base_phys = chan_params->tx_base_phys;
		bufs->rx_base_phys = chan_params->rx_base_phys;
		bufs->tx_curr_idx = 0;
		bufs->rx_curr_idx = 0;
	} else {
		pr_err("%s: Only MMAPED channels are supported, flags %x\n",
			rf_dev->name, chan_params->flags);
		kfree(chan);
		rc = -EINVAL;
		goto out;
	}

	ops = rf_dev->ops;
	rc = ops->channel_open(rf_dev, chan);
	if (rc) {
		pr_err("%s: Ant[%d] Ctrl dev failed to open channel\n",
			rf_dev->name, ant);
		kfree(chan);
		goto out;
	}

	rf_dev->channels[ant] = chan;
out:

	return rc;
}


int rf_channel_close(struct rf_ctrl_dev *rf_dev, int ant)
{
	struct rf_ctrl_ops *ops = rf_dev->ops;
	struct rf_channel *chan;
	int rc = 0;

	if (ant >= MAX_NUM_ANTENNAS) {
		pr_err("%s: Ant[%d] invalid\n", rf_dev->name,
			ant);
		rc = -EINVAL;
		goto out;
	}
	chan = rf_dev->channels[ant];

	if (!chan)
		goto out;

	rc = ops->channel_close(rf_dev, chan);
	if (rc) {
		pr_err("%s: Ctrl dev failed to close chan %d\n",
			rf_dev->name, ant);
		goto out;
	}

	kfree(chan);
	rf_dev->channels[ant] = NULL;

out:
	return rc;
}


int rf_rx_frame(struct rf_ctrl_dev *rf_dev, struct rf_frame *frame)
{
	struct rf_event_handler *event_handler;
	struct rf_channel *chan;
	struct rf_channel_stats *stats;
	u16 evt;
	int rc = 0;

	chan = rf_dev->channels[frame->ant];

	if (!(chan->flags & RF_CHAN_MMAPED)) {
		rc = -ENOSYS;
		pr_err("%s: Non MMAPED rx frame not supported\n",
			rf_dev->name);
		goto out;
	}

	stats = &chan->stats;
	stats->rx_frame_count++;
	raw_spin_lock(&rf_dev->event_handler_lock);

	list_for_each_entry(event_handler, &rf_dev->event_handler_list, list) {

		if (event_handler->evt_mask & RF_EVT_RX_FRAME) {

			evt = RF_MAKE_EVENT(RF_EVT_RX_FRAME, frame->ant,
				frame->buf.buffer_idx);
			rc = rf_send_event(rf_dev, event_handler, evt);
			if (rc) {
				pr_err("%s: Failed to send evt sig\n",
					rf_dev->name);
				break;
			}
		}
	}

	raw_spin_unlock(&rf_dev->event_handler_lock);
out:

	return rc;
}

int rf_sniff_done(struct rf_ctrl_dev *rf_dev, int ant)
{
	struct rf_event_handler *event_handler;
	u16 evt;
	int rc = 0;

	list_for_each_entry(event_handler, &rf_dev->event_handler_list, list) {

		if (event_handler->evt_mask & RF_EVT_SNIFF_DONE) {

			evt = RF_MAKE_EVENT(RF_EVT_SNIFF_DONE, ant, 0);
			rc = rf_send_event(rf_dev, event_handler, evt);
			if (rc) {
				pr_err("%s: Failed to send evt sig\n",
					rf_dev->name);
				break;
			}

		}
	}

	return rc;
}

