/*
 * drivers/rf/core.c
 * RF device framework
 *
 * Radio interface device framework exposes the combination of a RF interface
 * controller and RFIC as a RF device (eg. rf0, rf1) to user space for
 * configuration. This interface is for RF PHYs for LTE/CDMA systems. RF
 * interface controller driver and RFIC driver registers respective devices
 * with this framework, and thier combination is exposed to user space as
 * a unified rfdev interface 'rf0'.
 *
 * Author: pankaj chauhan <pankaj.chauhan@freescale.com>
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/types.h>
#include <linux/list.h>
#include <linux/bitops.h>
#include <linux/fs.h>
#include <linux/wait.h>
#include <linux/rfdev.h>
#include <linux/cdev.h>
#include <linux/uaccess.h>
#include <linux/signal.h>
#include <linux/rfdev_ioctl.h>
#include <linux/rf_channel.h>

struct rf_priv {
	struct list_head ctrl_list;
	struct list_head phy_list;
	spinlock_t phy_lock;
	spinlock_t ctrl_lock;
	unsigned long	name_idx_bitmap;
	struct cdev cdev;
	dev_t	dev;
};

static struct rf_priv *rf_priv;
static int rf_change_state(struct rf_ctrl_dev *rf_dev, unsigned int state);
static int rf_attach_phy(struct rf_ctrl_dev *rf_dev, struct rf_phy_dev *phy);
static int rf_open(struct inode *inode, struct file *filep);
static int rf_release(struct inode *inode, struct file *filep);
static ssize_t rf_read(struct file *, char __user *, size_t, loff_t *);
static int rf_ioctl(struct inode *, struct file *, unsigned int,
		unsigned long);

const static struct file_operations rf_fops = {
	.open 	= rf_open,
	.release = rf_release,
	.read 	= rf_read,
	.ioctl 	= rf_ioctl,
};

struct rf_ctrl_dev *allocate_rf_ctrl_dev(size_t priv_size,
		unsigned long flags)
{
	struct rf_ctrl_dev *rf_dev;
	size_t size;

	size = sizeof(struct rf_ctrl_dev) + priv_size;
	rf_dev = kzalloc(size, flags);

	if (!rf_dev)
		return rf_dev;

	atomic_set(&rf_dev->ref, 1);
	mutex_init(&rf_dev->lock);
	init_waitqueue_head(&rf_dev->wait_q);
	INIT_LIST_HEAD(&rf_dev->event_handler_list);
	raw_spin_lock_init(&rf_dev->event_handler_lock);
	raw_spin_lock_init(&rf_dev->wait_q_lock);
	rf_dev->priv = (unsigned char *) rf_dev + sizeof(struct rf_ctrl_dev);
	rf_dev->dev_idx = ~0;

	return rf_dev;
}
EXPORT_SYMBOL(allocate_rf_ctrl_dev);

int free_rf_ctrl_dev(struct rf_ctrl_dev *rf_dev)
{
	struct rf_event_handler *evt_handler;
	struct list_head *list, *next;

	if (!atomic_dec_and_test(&rf_dev->ref))
		return -EBUSY;

	list_for_each_safe(list, next, &rf_dev->event_handler_list) {
		evt_handler = list_entry(list, struct rf_event_handler, list);
		kfree(evt_handler);
	}

	kfree(rf_dev);

	return 0;
}
EXPORT_SYMBOL(free_rf_ctrl_dev);

int register_rf_ctrl_dev(struct rf_ctrl_dev *rf_dev)
{
	struct rf_ctrl_ops *ctrl_ops;
	unsigned int dev_idx;
	unsigned long *devid_map;
	int rc = 0;

	ctrl_ops = rf_dev->ops;
	if (!ctrl_ops || !ctrl_ops->init
		|| !ctrl_ops->start || !ctrl_ops->stop)
		return -EINVAL;

	rf_dev->state = RF_CREATED;
	spin_lock(&rf_priv->ctrl_lock);

	devid_map = (unsigned long *)&rf_priv->name_idx_bitmap;
	dev_idx = find_first_zero_bit(devid_map, 32);

	set_bit(dev_idx, devid_map);
	list_add_tail(&rf_dev->list, &rf_priv->ctrl_list);

	spin_unlock(&rf_priv->ctrl_lock);
	sprintf(&rf_dev->name[0], "rf%d", dev_idx);
	rf_dev->dev_idx = dev_idx;
	if (rf_attach_phy(rf_dev, NULL))
		rc = rf_change_state(rf_dev, RF_PHY_ATTACHED);

	return rc;
}
EXPORT_SYMBOL(register_rf_ctrl_dev);

int unregister_rf_ctrl_dev(struct rf_ctrl_dev *rf_dev)
{
	struct rf_phy_dev *phy;
	int rc = 0;

	/* Notify PHY that we are going down */
	phy = rf_dev->phy;

	if (phy->ops->stop) {
		rc = phy->ops->stop(phy);
		if (rc) {
			pr_debug("%s: unregister failed, phy busy\n",
					rf_dev->name);
			return -EBUSY;
		}
	}
	rf_change_state(rf_dev, RF_DOWN);

	spin_lock(&rf_priv->ctrl_lock);

	clear_bit(rf_dev->dev_idx, &rf_priv->name_idx_bitmap);
	list_del(&rf_dev->list);

	spin_unlock(&rf_priv->ctrl_lock);

	return rc;
}
EXPORT_SYMBOL(unregister_rf_ctrl_dev);

struct rf_phy_dev *allocate_rf_phy_dev(size_t priv_size, unsigned long flags)
{
	struct rf_phy_dev *phy;
	size_t size;

	size = sizeof(struct rf_phy_dev) + priv_size;
	phy = kzalloc(size, flags);
	if (!phy)
		return NULL;

	phy->priv = ((unsigned char *) phy) + sizeof(struct rf_phy_dev);

	return phy;
}
EXPORT_SYMBOL(allocate_rf_phy_dev);

int free_rf_phy_dev(struct rf_phy_dev *phy)
{
	kfree(phy);

	return 0;
}
EXPORT_SYMBOL(free_rf_phy_dev);

int register_rf_phy_dev(struct rf_phy_dev *phy)
{
	struct rf_ctrl_dev *rf_ctrl_dev;
	struct rf_phy_ops *phy_ops;

	phy_ops = phy->ops;
	if (!phy_ops || !phy_ops->init
		|| !phy_ops->start || !phy_ops->stop)
		return -EINVAL;

	spin_lock(&rf_priv->phy_lock);

	list_add_tail(&phy->list, &rf_priv->phy_list);

	spin_unlock(&rf_priv->phy_lock);

	if (rf_attach_phy(NULL, phy)) {
		rf_ctrl_dev = phy->ctrl_dev;
		rf_change_state(rf_ctrl_dev, RF_PHY_ATTACHED);
	}

	return 0;
}
EXPORT_SYMBOL(register_rf_phy_dev);

int unregister_rf_phy_dev(struct rf_phy_dev *phy)
{
	struct rf_ctrl_dev *rf_dev;
	int rc;

	rf_dev = phy->ctrl_dev;
	if (rf_dev->ops->phy_detach) {
		rc = rf_dev->ops->phy_detach(rf_dev);
		if (rc)
			return -EBUSY;
		rf_dev->phy = NULL;
		rf_change_state(rf_dev, RF_DOWN);
	}
	spin_lock(&rf_priv->phy_lock);

	list_del(&phy->list);

	spin_unlock(&rf_priv->phy_lock);

	return 0;
}
EXPORT_SYMBOL(unregister_rf_phy_dev);

int rf_change_state(struct rf_ctrl_dev *rf_dev, unsigned int new_state)
{
	int err = 0;

	/* Save old state, so that if any operation during
	 * state change fails then we can go back to previous
	 * state
	 */
	rf_dev->old_state = rf_dev->state;

	switch (new_state) {

	case RF_PHY_ATTACHED:
		if (rf_dev->state != RF_CREATED) {
			err = -EINVAL;
			goto out;
		}
		break;

	case RF_PHY_INITIALIZED:
		if (rf_dev->state < RF_PHY_ATTACHED) {
			err = -EINVAL;
			goto out;
		}
		break;

	case RF_INITIALIZED:
	case RF_SNIFFING:
		if (rf_dev->state < RF_PHY_INITIALIZED) {
			err = -EINVAL;
			goto out;
		}
		break;

	case RF_READY:
		if ((rf_dev->state < RF_INITIALIZED) ||
			(rf_dev->state == RF_READY)) {
			err = -EINVAL;
			goto out;
		}
		break;
	/* By default all other state changes are allowed */
	case RF_CREATED:
	case RF_TIMER_SYNC_AWAITED:
	case RF_TIMER_SYNC_FAILED:
	case RF_STOPPED:
	case RF_DOWN:
			/*Fall through*/
	default:
		break;
	}

	rf_dev->state = new_state;
out:
	return err;
}

int rf_attach_phy(struct rf_ctrl_dev *rf_dev, struct rf_phy_dev *phy)
{
	int match = 0;
	struct rf_ctrl_dev *rf_dev_itr = NULL;
	struct rf_phy_dev *phy_itr = NULL;

	/* Search for a phy for given rf ctrl dev */
	if (rf_dev) {
		spin_lock(&rf_priv->phy_lock);

		list_for_each_entry(phy_itr, &rf_priv->phy_list, list) {
			if (rf_dev->phy_id == phy_itr->phy_id) {
				match = 1;
				rf_dev->phy = phy_itr;
				phy_itr->ctrl_dev = rf_dev;
				break;
			}
		}

		spin_unlock(&rf_priv->phy_lock);
		return match;
	}

	/* Search for a rf ctrl dev for a given phy */
	if (phy) {
		spin_lock(&rf_priv->ctrl_lock);

		list_for_each_entry(rf_dev_itr, &rf_priv->ctrl_list, list) {
			if (phy->phy_id == rf_dev_itr->phy_id) {
				match = 1;
				phy->ctrl_dev = rf_dev_itr;
				rf_dev_itr->phy = phy;
				break;
			}
		}

		spin_unlock(&rf_priv->ctrl_lock);
	}

	return match;
}

int rf_notify_dl_tti(struct rf_ctrl_dev *rf_dev)
{
	struct rf_stats *stats;

	stats = &rf_dev->stats;
	stats->tti_count++;
	raw_spin_lock(&rf_dev->wait_q_lock);
	wake_up_locked(&rf_dev->wait_q);
	raw_spin_unlock(&rf_dev->wait_q_lock);

	return 0;
}

static int rf_open(struct inode *inode, struct file *filep)
{
	int minor, match = 0;
	struct rf_ctrl_dev *rf_dev = NULL;

	minor = iminor(inode);
	spin_lock(&rf_priv->ctrl_lock);

	list_for_each_entry(rf_dev, &rf_priv->ctrl_list, list) {
		if (rf_dev->dev_idx == minor) {
			match = 1;
			break;
		}
	}

	spin_unlock(&rf_priv->ctrl_lock);
	if (!match) {
		return -ENODEV;
	} else {
		filep->private_data = rf_dev;
		atomic_inc(&rf_dev->ref);
	}

	return 0;
}

int rf_release(struct inode *inode, struct file *filep)
{
	struct rf_ctrl_dev *rf_dev;

	rf_dev = filep->private_data;
	atomic_dec(&rf_dev->ref);

	return 0;
}
static ssize_t rf_read(struct file *filep, char __user *buf, size_t size,
			loff_t *offset)
{
	struct rf_ctrl_dev *rf_dev;
	struct rf_stats *stats;
	wait_queue_t 	wait;
	unsigned long flags;
	int rc;

	rf_dev = filep->private_data;
	stats = &rf_dev->stats;
	init_waitqueue_entry(&wait, current);

	/*
	 * Spin_locks are changed to mutexes if PREEMPT_RT is enabled,
	 * i.e they can sleep. This fact is problem for us because
	 * add_wait_queue()/wake_up_all() takes wait queue spin lock.
	 * Since spin lock can sleep with PREEMPT_RT, wake_up_all() can not be
	 * called from rf_notify_dl_tti (which is called in interrupt context).
	 * As a workaround, wait_q_lock is used for protecting the wait_q and
	 * add_wait_queue_locked()/ wake_up_locked() functions of wait queues
	 * are used.
	 */
	raw_spin_lock_irqsave(&rf_dev->wait_q_lock, flags);
	add_wait_queue_exclusive_locked(&rf_dev->wait_q, &wait);
	raw_spin_unlock_irqrestore(&rf_dev->wait_q_lock, flags);
	set_current_state(TASK_INTERRUPTIBLE);
	/*Now wait here, tti notificaion will wake us up*/
	schedule();
	set_current_state(TASK_RUNNING);
	raw_spin_lock_irqsave(&rf_dev->wait_q_lock, flags);
	remove_wait_queue_locked(&rf_dev->wait_q, &wait);
	raw_spin_unlock_irqrestore(&rf_dev->wait_q_lock, flags);

	rc = put_user(stats->tti_count, (int *)buf);
	if (!rc)
		rc = sizeof(stats->tti_count);

	return rc;
}

int rf_fill_dev_info(struct rf_ctrl_dev *rf_dev, struct rf_dev_info *dev_info)
{
	int i;
	struct rf_phy_dev *phy;
	struct rf_dev_params *dev_params;
	struct rf_dev_sniffer *sniffer;

	dev_params = &rf_dev->dev_params;

	dev_info->state = rf_dev->state;

	strncpy(dev_info->controller, rf_dev->name, RIF_NAME_SIZE);
	if (rf_dev->state >= RF_PHY_ATTACHED) {
		phy = rf_dev->phy;
		strncpy(dev_info->phy, phy->name, RIF_NAME_SIZE);
		for (i = 0; i < dev_params->ants; i++)
			dev_info->tx_atten[i] = phy->tx_atten[i];
	} else {
		memset(dev_info->phy, 0, RIF_NAME_SIZE);
	}

	dev_info->net_mode = rf_dev->net_mode;
	dev_info->tx_rxmode = rf_dev->tx_rxmode;
	dev_info->bw = rf_dev->bw;
	dev_info->timing_src = rf_dev->timing_src;

	dev_info->ants = dev_params->ants;
	dev_info->symbol_len = dev_params->symbol_len;
	dev_info->long_cp = dev_params->long_cp;
	dev_info->cp0_len = dev_params->cp0_len;
	dev_info->cp1_len = dev_params->cp1_len;

	dev_info->sniff_enabled = rf_dev->sniff_enabled;

	if (dev_info->sniff_enabled) {
		sniffer = &rf_dev->sniffer;
		dev_info->sniff_net_mode = sniffer->net_mode;
		dev_info->sniff_tx_rxmode = sniffer->tx_rxmode;
		dev_info->sniff_timing_src = sniffer->timing_src;
		dev_info->sniff_bw = sniffer->bw;
		dev_info->sniff_carrier_freq = sniffer->carrier_freq;
	}

	return 0;
}

void rf_sniffer_enabled(struct rf_ctrl_dev *rf_dev,
	struct rf_sniff_params *sniff_params)
{
	struct rf_dev_sniffer *sniffer = &rf_dev->sniffer;
	struct rf_init_params *sniff_dev_params;

	rf_dev->sniff_enabled = 1;
	sniff_dev_params = &sniff_params->dev_params;
	sniffer->net_mode = sniff_dev_params->mode;
	sniffer->tx_rxmode = sniff_dev_params->tx_rxmode;
	sniffer->timing_src = sniff_params->timing_src;
	sniffer->bw = sniff_dev_params->bw;
	sniffer->carrier_freq = sniff_params->carrier_freq;
}

void rf_sniffer_disabled(struct rf_ctrl_dev *rf_dev)
{

	rf_dev->sniff_enabled = 0;
}

int rf_send_event(struct rf_ctrl_dev *rf_dev,
	struct rf_event_handler *evt_handler, u16 evt)
{
	struct task_struct *dst_task;
	struct siginfo *siginfo = &evt_handler->siginfo;
	int rc = 0;

	dst_task = pid_task(evt_handler->pid, PIDTYPE_PID);
	if (!dst_task) {
		pr_err("%s: No listener task, Failed to send evt sig\n",
			rf_dev->name);
		goto out;
	}
	siginfo->si_int = evt;
	rc = send_sig_info(siginfo->si_signo, siginfo, dst_task);
	if (rc) {
		pr_err("%s: Failed to send evt sig\n",
			rf_dev->name);
		goto out;
	}

out:
	return rc;
}

int rf_register_event_handler(struct rf_ctrl_dev *rf_dev,
	struct rf_event_listener *listener)
{
	struct rf_event_handler *evt_handler;
	int rc = 0;

	if (!(listener->evt_mask & RF_EVT_ALL)) {
		pr_err("%s: Unkown event %x, registration failed\n",
			rf_dev->name, listener->evt_mask);
		rc = -EINVAL;
		goto out;
	}

	evt_handler = kzalloc(sizeof(struct rf_event_handler), GFP_KERNEL);
	if (!evt_handler) {
		rc = -ENOMEM;
		goto out;
	}

	evt_handler->evt_mask = listener->evt_mask;
	evt_handler->pid = task_pid(current);
	evt_handler->siginfo.si_signo = listener->signal;
	evt_handler->siginfo.si_errno = 0;
	evt_handler->siginfo.si_code = SI_QUEUE;
	raw_spin_lock(&rf_dev->event_handler_lock);
	list_add_tail(&evt_handler->list,
		&rf_dev->event_handler_list);
	raw_spin_unlock(&rf_dev->event_handler_lock);

out:
	return rc;
}

int rf_unregister_event_handler(struct rf_ctrl_dev *rf_dev)
{
	struct rf_event_handler *evt_handler = NULL;
	struct task_struct *listener_task;
	struct list_head *list, *next;
	int rc = 0, found_evt_handler = 0;

	raw_spin_lock(&rf_dev->event_handler_lock);

	list_for_each_safe(list, next, &rf_dev->event_handler_list) {
		evt_handler = list_entry(list, struct rf_event_handler, list);
		listener_task = pid_task(evt_handler->pid, PIDTYPE_PID);
		if (listener_task == current) {
			list_del(&evt_handler->list);
			found_evt_handler = 1;
			break;
		}
	}
	raw_spin_unlock(&rf_dev->event_handler_lock);

	if (found_evt_handler)
		kfree(evt_handler);

	return rc;
}

static int rf_ioctl(struct inode *inode, struct file *filep,
		unsigned int cmd, unsigned long arg)
{
	struct rf_ctrl_dev *rf_dev;
	struct rf_phy_dev *phy;
	struct rf_init_params init_params;
	struct rif_phy_cmd_set cmd_set;
	struct rif_reg_buf reg_buf;
	struct rif_write_reg_buf write_reg_buf;
	struct rif_dac_params dac_params;
	struct rif_dac_buf dac_buff;
	struct rf_dev_info dev_info;
	struct rf_tx_buf tx_buf;
	struct rf_tx_en_dis tx_en_dis;
	struct rf_rssi rssi;
	struct rf_rx_gain rx_gain;
	struct rf_gain_ctrl gain_ctrl;
	struct rf_sniff_params sniff_params;
	struct rf_synth_table synth_table;
	struct rf_channel_params chan_params;
	struct rf_event_listener listener;
	unsigned long long (*params_buf)[NUM_SYNTH_PARAMS];
	u8 (*reg_vals_buf)[NUM_SYNTH_REGS];
	u32	*buf;
	u32 u32arg;
	int rc = -ENOSYS, size;

	rf_dev = filep->private_data;
	phy = rf_dev->phy;

	mutex_lock_interruptible(&rf_dev->lock);
	switch (cmd) {

	case RIF_DEV_INIT:

		if (rf_change_state(rf_dev, RF_INITIALIZED)) {
			rc = -EAGAIN;
			goto out;
		}
		if (copy_from_user(&init_params,
			(struct rf_init_params *)arg, sizeof(init_params))) {
			rc = -EFAULT;
			goto out;
		}

		if (phy) {
			rc = phy->ops->init(phy, &init_params);
			if (rc)
				goto out;
		}

		rc = rf_dev->ops->init(rf_dev, &init_params);
		break;

	case RIF_CONFIG_SNIFF:

		if (rf_change_state(rf_dev, RF_SNIFFING)) {
			rc = -EAGAIN;
			goto out;
		}
		if (copy_from_user(&sniff_params,
			(struct rf_sniff_params *)arg, sizeof(sniff_params))) {
			rc = -EFAULT;
			goto out;
		}

		if (phy && phy->ops->config_sniff) {
			rc = phy->ops->config_sniff(phy, &sniff_params);
			if (rc)
				goto out;
		}
		if (rf_dev->ops->config_sniff)
			rc = rf_dev->ops->config_sniff(rf_dev, &sniff_params);
		if (!rc)
			rf_sniffer_enabled(rf_dev, &sniff_params);
		break;

	case RIF_SET_TIMER_SOURCE:

		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!get_user(u32arg, (int *)arg))
			rc = rf_dev->ops->set_timing_source(rf_dev, u32arg);
		else
			rc = -EFAULT;
		break;

	case RIF_SET_TIMER_CORRECTION:

		if (rf_dev->state < RF_PHY_ATTACHED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!copy_from_user(&dac_params, (struct rif_dac_params *)arg,
			sizeof(struct rif_dac_params))) {

			if (phy && phy->ops->set_timer_correction) {
				rc = phy->ops->set_timer_correction(phy,
						&dac_params);
			}
		} else {
			rc = -EFAULT;
		}

		break;

	case RIF_GET_DAC_VALUE:

		if (rf_dev->state < RF_PHY_ATTACHED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!copy_from_user(&dac_buff, (struct rif_dac_buf *)arg,
			sizeof(struct rif_dac_buf))) {

			if (phy && phy->ops->read_dac_value) {
				rc = phy->ops->read_dac_value(phy,
					dac_buff.correction_type, &u32arg);
			}
			if (copy_to_user(dac_buff.buf, &u32arg, sizeof(u32arg)))
				rc = -EFAULT;
		} else {
			rc = -EFAULT;
		}

		break;

	case RIF_RUN_PHY_CMDS:

		if (rf_change_state(rf_dev, RF_PHY_INITIALIZED)) {
			rc = -EAGAIN;
			goto out;
		}

		if (!phy || !phy->ops->run_cmds)
			goto out;

		rc = -EFAULT;
		if (!copy_from_user(&cmd_set, (struct rif_phy_cmd_set *)arg,
			sizeof(struct rif_phy_cmd_set))) {

			size = sizeof(struct rif_phy_cmd) * cmd_set.count;
			buf = kzalloc(size, GFP_KERNEL);
			if (buf) {
				if (!copy_from_user(buf,
						(u32 *) cmd_set.cmds, size))
					rc = phy->ops->run_cmds(phy,
						(struct rif_phy_cmd *)buf,
						cmd_set.count);
				kfree(buf);
			} else {
				rc = -ENOMEM;
			}
		}
		break;

	case RIF_EN_DIS_TX:
		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!copy_from_user(&tx_en_dis, (struct rf_tx_en_dis *)arg,
				sizeof(struct rf_tx_en_dis))) {
			if (phy && phy->ops->en_dis_tx) {
				rc = phy->ops->en_dis_tx(phy,
						tx_en_dis.tx_if,
						tx_en_dis.tx_cmd);
			}
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_READ_RSSI:

		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}
		if (!copy_from_user(&rssi, (struct rf_rssi *)arg,
			sizeof(struct rf_rssi))) {

			if (phy && phy->ops->read_rssi)
				rc = phy->ops->read_rssi(phy, &rssi);
			if (!rc && copy_to_user((struct rf_rssi *)arg,
						&rssi, sizeof(struct rf_rssi)))
				rc = -EFAULT;
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_READ_PHY_REGS:
	case RIF_READ_CTRL_REGS:

		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!copy_from_user(&reg_buf, (struct rif_reg_buf *)arg,
			sizeof(struct rif_reg_buf))) {

			size = 4 * reg_buf.count;
			buf = kzalloc(size, GFP_KERNEL);
			if (!buf) {
				rc = -ENOMEM;
				goto out;
			}
			if (cmd == RIF_READ_PHY_REGS) {

				if (phy && phy->ops->read_regs) {
					rc = phy->ops->read_regs(phy,
						reg_buf.addr,
						reg_buf.count, buf);
				}
			} else if (cmd == RIF_READ_CTRL_REGS) {

				if (rf_dev->ops->read_regs) {
					rc = rf_dev->ops->read_regs(rf_dev,
						reg_buf.addr,
						reg_buf.count, buf);
				}
			}
			if (copy_to_user((u32 *)reg_buf.buf,
					buf, size))
				rc = -EFAULT;
			kfree(buf);
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_WRITE_PHY_REGS:

		if (rf_dev->state < RF_PHY_ATTACHED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!copy_from_user(&write_reg_buf,
			(struct rif_write_reg_buf *)arg,
			sizeof(struct rif_write_reg_buf))) {

			if (phy->ops->write_reg) {
				rc = phy->ops->write_reg(phy,
						write_reg_buf.addr,
						write_reg_buf.data);
			}
		} else {
			rc = -EFAULT;
		}

		break;

	case RIF_WRITE_CTRL_REGS:

		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!copy_from_user(&write_reg_buf,
			(struct rif_write_reg_buf *)arg,
			sizeof(struct rif_write_reg_buf))) {

			if (rf_dev->ops->write_reg) {
				rc = rf_dev->ops->write_reg(rf_dev,
						write_reg_buf.addr,
						write_reg_buf.data);
			}
		} else {
			rc = -EFAULT;
		}

		break;

	case RIF_START:
		if (rf_change_state(rf_dev, RF_READY)) {
			rc = -EAGAIN;
			goto out;
		}
		if (phy) {
			rc = phy->ops->start(phy);
			if (rc)
				goto out;
		}
		rc = rf_dev->ops->start(rf_dev);
		break;

	case RIF_STOP:

		if (rf_change_state(rf_dev, RF_STOPPED)) {
			rc = -EAGAIN;
			goto out;
		}

		rc = rf_dev->ops->stop(rf_dev);
		if (!rc && rf_dev->sniff_enabled)
			rf_sniffer_disabled(rf_dev);

		if (phy) {
			rc = phy->ops->stop(phy);
			if (rc)
				goto out;
		}
		break;

	case RIF_GET_DEV_INFO:
		rc = rf_fill_dev_info(rf_dev, &dev_info);
		if (copy_to_user((struct rf_dev_info *)arg,
					&dev_info, sizeof(dev_info)))
			rc = -EFAULT;
		break;
	case RIF_SET_TX_ATTEN:
		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}

		if (!copy_from_user(&tx_buf, (struct rf_tx_buf *)arg,
				sizeof(struct rf_tx_buf))) {
			if (phy && phy->ops->set_tx_atten) {
				rc = phy->ops->set_tx_atten(phy,
						tx_buf.tx_if,
						tx_buf.tx_atten);
			}
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_READ_RX_GAIN:

		if (rf_dev->state < RF_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}
		if (!copy_from_user(&rx_gain, (struct rf_rx_gain *)arg,
			sizeof(struct rf_rx_gain))) {

			if (phy && phy->ops->get_rx_gain)
				rc = phy->ops->get_rx_gain(phy, &rx_gain);

			if (!rc && copy_to_user((struct rf_rx_gain *)arg,
						&rx_gain,
						sizeof(struct rf_rx_gain)))
				rc = -EFAULT;
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_WRITE_RX_GAIN:

		if (rf_dev->state < RF_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}
		if (!copy_from_user(&rx_gain, (struct rf_rx_gain *)arg,
			sizeof(struct rf_rx_gain))) {

			if (phy && phy->ops->set_rx_gain)
				rc = phy->ops->set_rx_gain(phy, &rx_gain);
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_SET_GAIN_CTRL_MODE:

		if (rf_dev->state < RF_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}
		if (!copy_from_user(&gain_ctrl, (struct rf_gain_ctrl *)arg,
			sizeof(struct rf_gain_ctrl))) {

			if (phy && phy->ops->set_gain_ctrl_mode)
				rc = phy->ops->set_gain_ctrl_mode(phy,
						&gain_ctrl);
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_INIT_SYNTH_TABLE:
		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EAGAIN;
			goto out;
		}
		if (!phy->ops->save_synth_table)
			goto out;

		if (!copy_from_user(&synth_table, (struct rf_synth_table *)arg,
					sizeof(struct rf_synth_table))) {
			size = sizeof(unsigned long long) *
				NUM_SYNTH_PARAMS * synth_table.count;
			params_buf = (unsigned long long (*)[NUM_SYNTH_PARAMS])
					kzalloc(size, GFP_KERNEL);
			rc = 0;
			if (params_buf) {
				if (copy_from_user(params_buf, (unsigned long
					long *)synth_table.params, size))
					rc = -EFAULT;
			} else
				rc = -ENOMEM;
			if (rc) {
				kfree(params_buf);
				goto out;
			}

			size = sizeof(u8) * NUM_SYNTH_REGS * synth_table.count;
			reg_vals_buf = (u8 (*)[NUM_SYNTH_REGS])kzalloc(size,
					GFP_KERNEL);
			if (reg_vals_buf) {
				if (copy_from_user(reg_vals_buf, (u8 *)
						synth_table.reg_vals, size))
					rc = -EFAULT;
			} else
				rc = -ENOMEM;
			if (rc) {
				kfree(reg_vals_buf);
				kfree(params_buf);
				goto out;
			}

			synth_table.params = params_buf;
			synth_table.reg_vals = reg_vals_buf;
			rc = phy->ops->save_synth_table(phy, &synth_table);
			kfree(params_buf);
			kfree(reg_vals_buf);
		} else {
			rc = -EFAULT;
		}
		break;

	case RIF_CHANNEL_OPEN:
		if (rf_dev->state < RF_PHY_INITIALIZED) {
			rc = -EINVAL;
			goto out;
		}
		if (copy_from_user(&chan_params,
			(struct rf_chan_params *)arg, sizeof(chan_params))) {
			rc = -EFAULT;
			goto out;
		}
		rc = rf_channel_open(rf_dev, &chan_params);

		break;

	case RIF_CHANNEL_CLOSE:

		if (!get_user(u32arg, (int *)arg))
			rc = rf_channel_close(rf_dev, u32arg);
		else
			rc = -EFAULT;
		break;

	case RIF_REGISTER_EVENT:

		if (copy_from_user(&listener, (struct rf_event_listener *)arg,
			sizeof(listener))) {
			rc = -EFAULT;
			goto out;
		}

		rc = rf_register_event_handler(rf_dev, &listener);
		break;

	case RIF_UNREGISTER_EVENT:

		rc = rf_unregister_event_handler(rf_dev);
		break;

	default:
		rc = -ENOSYS;
	}

	if (!rc)
		rf_dev->old_state = rf_dev->state;
out:
	if (rc)
		rf_dev->state = rf_dev->old_state;
	mutex_unlock(&rf_dev->lock);
	return rc;
}

static int __init rf_init(void)
{
	int rc;

	rf_priv = kzalloc(sizeof(struct rf_priv), GFP_KERNEL);
	if (!rf_priv) {
		pr_debug("rfdev: Unable to allocate rf_priv\n");
		return -ENOMEM;
	}

	rc = alloc_chrdev_region(&rf_priv->dev, 0, RF_MAX_DEVS, "rfdev");
	if (rc) {
		pr_debug("rfdev:Failed to register rf chardev,err %d\n",
				rc);
		return rc;
	}

	cdev_init(&rf_priv->cdev, &rf_fops);
	cdev_add(&rf_priv->cdev, rf_priv->dev, RF_MAX_DEVS);
	INIT_LIST_HEAD(&rf_priv->phy_list);
	INIT_LIST_HEAD(&rf_priv->ctrl_list);

	return 0;
}

static void __exit rf_cleanup(void)
{
	if (rf_priv) {
		cdev_del(&rf_priv->cdev);
		unregister_chrdev_region(rf_priv->dev, RF_MAX_DEVS);
		kfree(rf_priv);
	}
}

MODULE_LICENSE("GPL v2");
module_init(rf_init);
module_exit(rf_cleanup);
