/*
 *
 * File: drivers/rf/phy/ad9361.c
 *
 * Freescale AD9361 Phy driver.
 * AD9361 is a RF phy device connected to one of the AIC lane.
 * Control path of AD9361 is through SPI interface whereas data path
 * is through AIC. Its driver register it as a spi driver for the configuration
 * of ad device via spi and also registers itself as rf_phy_dev that binds
 * phy device to AIC lane, for data path(UL/DL).
 *
 * Author: Shaveta Leekha <shaveta@freescale.com>
 *
 * Copyright (C) 2011-2012 Freescale Semiconductor, Inc. All rights reserved.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 */

#include <linux/module.h>
#include <linux/device.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/spi/spi.h>
#include <linux/param.h>
#include <linux/delay.h>
#include <linux/of.h>
#include <linux/io.h>
#include <linux/of_gpio.h>
#include <linux/gpio.h>
#include <linux/rfdev.h>
#include <linux/string.h>
#include <linux/fsl_fpga.h>
#include "ad9361.h"

#define DRV_DESC "FREESCALE DEVELOPED AD9361 PHY DRIVER"
#define DRV_NAME "ad9361_phy"
#define DEBUG

/* To check where reset gpio has been toggled to
 * reset AD PHY device or not. */
bool reset_status;

MODULE_LICENSE("GPL v2");
MODULE_DESCRIPTION(DRV_DESC);

static int ad_phy_run_cmds(struct rf_phy_dev *ad_phy, struct rif_phy_cmd *cmds,
		int count);
static int ad_phy_read(struct rf_phy_dev *ad_phy, u32 start, u32 count,
		u32 *buff);
static int ad_phy_write(struct rf_phy_dev *ad_phy, u32 reg, u32 data);
static int ad_set_tx_atten(struct rf_phy_dev *ad_phy, u32 tx_if, u32 tx_atten);
static int ad_en_dis_tx(struct rf_phy_dev *ad_phy, u32 tx_if, u32 cmd);
static int ad_en_dis_rx(struct rf_phy_dev *ad_phy, u32 rx_if, u32 rx_cmd);
static int ad_read_rssi(struct rf_phy_dev *ad_phy, struct rf_rssi *rssi);
static int ad_phy_dac_correction(struct rf_phy_dev *ad_phy,
					struct rif_dac_params *params);
static int ad_phy_get_dac_value(struct rf_phy_dev *ad_phy,
					u32 correction_type, u32 *buff);
static int ad_phy_start(struct rf_phy_dev *ad_phy);
static int ad_init(struct rf_phy_dev *phy, struct rf_init_params *params);
static int ad_phy_stop(struct rf_phy_dev *phy);
static int ad_get_rx_gain(struct rf_phy_dev *ad_phy,
		struct rf_rx_gain *rx_gain);
static int ad_set_rx_gain(struct rf_phy_dev *ad_phy,
		struct rf_rx_gain *rx_gain);
static int ad_config_sniff(struct rf_phy_dev *ad_phy,
		struct rf_sniff_params *sniff_params);
static int ad_set_gain_ctrl_mode(struct rf_phy_dev *ad_phy,
		struct rf_gain_ctrl *gain_ctrl);
static int ad_set_band(struct rf_phy_dev *ad_phy, unsigned int band);
static int ad_save_synth_table(struct rf_phy_dev *ad_phy,
		struct rf_synth_table *synth_table);

static struct rf_phy_ops ad_phy_ops = {
	.init = ad_init,
	.set_timer_correction = ad_phy_dac_correction,
	.read_dac_value = ad_phy_get_dac_value,
	.run_cmds = ad_phy_run_cmds,
	.read_regs = ad_phy_read,
	.write_reg = ad_phy_write,
	.set_tx_atten = ad_set_tx_atten,
	.en_dis_tx = ad_en_dis_tx,
	.read_rssi = ad_read_rssi,
	.start = ad_phy_start,
	.stop = ad_phy_stop,
	.get_rx_gain = ad_get_rx_gain,
	.set_rx_gain = ad_set_rx_gain,
	.config_sniff = ad_config_sniff,
	.set_gain_ctrl_mode = ad_set_gain_ctrl_mode,
	.save_synth_table = ad_save_synth_table,
};


u16 create_ad_phy_cmd(u16 reg, int opcode, int bytes_transfer)
{
	u16 cmd_word = 0;

	switch (opcode) {
	case SPI_WRITE:
		cmd_word = OPCODE_WRITE
			|(((bytes_transfer - 1) << SHIFT_BYTES_TRANSFER)
						& (BYTES_TRANSFER_MASK))
			| (reg & REG_ADDRESS_MASK);
		break;

	case SPI_READ:
		cmd_word = OPCODE_READ
			|(((bytes_transfer - 1) << SHIFT_BYTES_TRANSFER)
						& (BYTES_TRANSFER_MASK))
			| (reg & REG_ADDRESS_MASK);
		break;

	}

	return cmd_word;
}

static int spi_write_transaction(struct rf_phy_dev *ad_phy, u8 *buf, int len)
{
	u8 *tx_buf;
	int status;
	struct spi_message spi_msg;
	struct spi_transfer ad_tx;
	struct ad_dev_info *phy_info = ad_phy->priv;

	spi_message_init(&spi_msg);
	memset(&ad_tx, 0, sizeof(ad_tx));

	tx_buf = buf;
	ad_tx.len = len;
	ad_tx.tx_buf = &tx_buf[0];
	ad_tx.rx_buf = &(phy_info->rx_buf[0]);

	spi_message_add_tail(&ad_tx, &spi_msg);
	status = spi_sync(phy_info->ad_spi, &spi_msg);

	return status;
}

static int spi_read_transaction(struct rf_phy_dev *ad_phy, u8 *buf, int len)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	u8 *tx_buf;
	struct spi_message spi_msg;
	struct spi_transfer ad_tx;
	int status;

	spi_message_init(&spi_msg);
	memset(&ad_tx, 0, sizeof(ad_tx));

	tx_buf = buf;
	ad_tx.len = len;
	ad_tx.tx_buf = &tx_buf[0];
	ad_tx.rx_buf = &(phy_info->rx_buf[0]);

	spi_message_add_tail(&ad_tx, &spi_msg);
	status = spi_sync(phy_info->ad_spi, &spi_msg);

	return status;
}

static int check_cal_done(struct rf_phy_dev *ad_phy, u32 reg, u32 mask,
		u32 bitval)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	u32 val;
	int rc = 0;
	ad_phy_read(ad_phy, reg, 1, &val);
	switch (bitval) {
	case 0:
		if (!(val & mask))
			rc = 1;
		break;
	case 1:
		if (val & mask)
			rc = 1;
		break;
	default:
		dev_err(dev, "Invalid bit value\n");
		return -EINVAL;
	}
	return rc;
}

int save_tx_atten(struct rf_phy_dev *ad_phy)
{
	u32 buf[4];
	u32 msb;
	ad_phy_read(ad_phy, TX1_ATTEN0, sizeof(buf)/sizeof(u32), buf);
	msb = buf[1] & ATTEN_MSB_BIT_MASK;
	ad_phy->tx_atten[0] = (msb << MSB_SHIFT) + buf[0];
	msb = buf[3] & ATTEN_MSB_BIT_MASK;
	ad_phy->tx_atten[1] = (msb << MSB_SHIFT) + buf[2];
	return 1;
}

int ad_phy_run_cmds(struct rf_phy_dev *ad_phy,
		struct rif_phy_cmd *cmds, int count)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	int i, elapsed_time;
	u32 val;

	for (i = 0; i < count; i++) {
		switch (cmds[i].cmd) {
		case SPI_WRITE:
			if (cmds[i].param1 == REG_RX_CP_CONFIG) {
				ad_phy_read(ad_phy, REG_RX_CAL_STATUS, 1, &val);
				if (val & MASK_RX_CP_CAL_VALID)
					break;
			} else if (cmds[i].param1 == REG_TX_CP_CONFIG) {
				ad_phy_read(ad_phy, REG_TX_CAL_STATUS, 1, &val);
				if (val & MASK_TX_CP_CAL_VALID)
					break;
			}
			ad_phy_write(ad_phy, cmds[i].param1, cmds[i].param2);

			dev_dbg(dev, "Write: %x %x\n", cmds[i].param1,
							cmds[i].param2);
			break;

		case SPI_READ:

			/* XXX: This READ is not returnning anything from
			 * here so can be removed safely. But not removing
			 * it completely from here because these reads may
			 * impact timing between reg updates up RFIC.
			 *
			 * This needs to be fixed, right way is to return
			 * read values, and let user space decide what to
			 * do this value.
			 */
			ad_phy_read(ad_phy, cmds[i].param1, 1, &val);
			dev_dbg(dev, "Read from reg: %x, val %x\n",
				cmds[i].param1, val);
			break;

		case SPI_WAIT:
			msleep_interruptible(cmds[i].param3);
			break;

		case SPI_WAIT_BBPLL_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy, REG_CH1_OVERFLOW,
							MASK_BBPLL_LOCK, 1))
					break;
			}
			break;

		case SPI_WAIT_RXCP_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy, REG_RX_CAL_STATUS,
						MASK_RX_CP_CAL_VALID, 1))
					break;
			}
			break;

		case SPI_WAIT_TXCP_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy,
						REG_TX_CAL_STATUS,
						MASK_TX_CP_CAL_VALID, 1))
					break;
			}
			break;

		case SPI_WAIT_RXFILTER_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy,
						REG_CALIBRATION_CONTROL,
						MASK_RX_BB_TUNE, 0))
					break;
			}
			break;

		case SPI_WAIT_TXFILTER_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy,
						REG_CALIBRATION_CONTROL,
						MASK_TX_BB_TUNE, 0))
					break;
			}
			break;

		case SPI_WAIT_BBDC_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy,
						REG_CALIBRATION_CONTROL,
						MASK_DC_CAL_BBSTART, 0))
					break;
			}
			break;

		case SPI_WAIT_RFDC_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy,
						REG_CALIBRATION_CONTROL,
						MASK_DC_CAL_RFSTART, 0))
					break;
			}
			break;

		case SPI_WAIT_TXQUAD_CAL:
			for (elapsed_time = 1; elapsed_time < cmds[i].param3;
					elapsed_time++) {
				msleep_interruptible(1);
				if (check_cal_done(ad_phy,
						REG_CALIBRATION_CONTROL,
						MASK_TXQUAD_CAL, 0))
					break;
			}
			break;

		case SPI_WAIT_CALDONE:
			dev_info(dev, "Waiting for unknown CALDONE.\n");
			msleep_interruptible(cmds[i].param3);
			break;

		default:
			dev_dbg(dev, "Not a valid AD_PHY command\n");
			return -EINVAL;
		}
	}

	save_tx_atten(ad_phy);
	return 0;
}

static int ad_save_synth_table(struct rf_phy_dev *ad_phy,
		struct rf_synth_table *synth_table)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	int size = sizeof(unsigned long long) * synth_table->count *
					NUM_SYNTH_PARAMS;

	kfree(phy_info->synth_table.params);
	kfree(phy_info->synth_table.reg_vals);
	phy_info->synth_table.params =
		(unsigned long long (*)[NUM_SYNTH_PARAMS])kzalloc
			(size, GFP_KERNEL);
	if (!phy_info->synth_table.params) {
		dev_dbg(dev, "Failed to allocate rf_synth_table\n");
		return -ENOMEM;
	}
	memcpy(phy_info->synth_table.params, synth_table->params, size);

	size = sizeof(u8) * synth_table->count * NUM_SYNTH_REGS;
	phy_info->synth_table.reg_vals = (u8 (*)[NUM_SYNTH_REGS])kzalloc
						(size, GFP_KERNEL);
	if (!phy_info->synth_table.reg_vals) {
		dev_dbg(dev, "Failed to allocate rf_synth_table\n");
		kfree(phy_info->synth_table.params);
		return -ENOMEM;
	}
	memcpy(phy_info->synth_table.reg_vals, synth_table->reg_vals, size);
	phy_info->synth_table.count = synth_table->count;

	return 0;
}

static int ad_get_rx_gain(struct rf_phy_dev *ad_phy,
		struct rf_rx_gain *rx_gain)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	enum rx_gain_table_name tbl;
	struct rx_gain_info *gain_info;
	u32 val, idx_reg;
	u8 gain_ctl_shift, rx_enable_mask;
	u8 fast_atk_shift;
	int rc = 0, rx_gain_db;

	if (rx_gain->ant == 1) {
		gain_ctl_shift = RX1_GAIN_CTL_SHIFT;
		idx_reg = REG_RX1_FULL_TBL_IDX;
		rx_enable_mask = RX1_EN;
		fast_atk_shift = RX1_FAST_ATK_SHIFT;

	} else if (rx_gain->ant == 2) {
		gain_ctl_shift = RX2_GAIN_CTL_SHIFT;
		idx_reg = REG_RX2_FULL_TBL_IDX;
		rx_enable_mask = RX2_EN;
		fast_atk_shift = RX2_FAST_ATK_SHIFT;
	} else {
		dev_err(dev, "Unkown Rx path %d\n", rx_gain->ant);
		rc = -EINVAL;
		goto out;

	}

	rc = ad_phy_read(ad_phy, REG_RXEN_N_FILTER_CTRL, 1, &val);
	if (rc) {
		dev_err(dev, "Unable to read REG_RXEN_N_FILTER_CTRL\n");
		goto out;
	}

	if (!(val & rx_enable_mask)) {
		dev_err(dev, "Rx%d is not enabled\n", rx_gain->ant);
		rc = -EAGAIN;
		goto out;
	}

	rc = ad_phy_read(ad_phy, REG_AGC_CONF2, 1, &val);
	if (rc) {
		dev_err(dev, "Unable to read AGC_CONF2\n");
		goto out;
	}

	if (!(val & FULL_GAIN_TBL)) {
		dev_err(dev, "RX gain read for split table not supported\n");
		rc = -ENOSYS;
		goto out;
	}

	rc = ad_phy_read(ad_phy, REG_AGC_CONF1, 1, &val);
	if (rc) {
		dev_err(dev, "Unable to read AGC_CONF1\n");
		goto out;
	}

	val = (val >> gain_ctl_shift) & RX_GAIN_CTL_MASK;

	if (val == RX_GAIN_CTL_AGC_FAST_ATK) {
		/* In fast attack mode check whether Fast attack state machine
		 * has locked gain, if not then we can not read gain.
		 */
		rc = ad_phy_read(ad_phy, REG_FAST_ATK_STATE, 1, &val);
		if (rc) {
			dev_err(dev, "Unable to read REG_FAST_ATK_STATE\n");
			goto out;
		}

		val = (val >> fast_atk_shift) & FAST_ATK_MASK;
		if (val != FAST_ATK_GAIN_LOCKED) {
			dev_err(dev, "Failed to read gain, state m/c at %x\n",
				val);
			rc = -EAGAIN;
			goto out;
		}
	}

	switch (phy_info->freq_band) {
	case 1:
	case 7:
		tbl = FULL_TBL_1300_4000_MHZ;
		break;
	case 13:
		tbl = FULL_TBL_200_1300_MHZ;
		break;
	default:
		dev_err(dev, "unkown band for rx gain read, %d\n",
			phy_info->freq_band);
		rc = -EINVAL;
		goto out;
	}

	rc = ad_phy_read(ad_phy, idx_reg, 1, &val);
	if (rc) {
		dev_err(dev, "Unable to read gain tbl idx reg: %d\n", idx_reg);
		goto out;
	}

	val = val & FULL_TBL_IDX_MASK;
	gain_info = &phy_info->rx_gain[tbl];
	if (val > gain_info->idx_step_offset) {
		val = val - gain_info->idx_step_offset;
		rx_gain_db = gain_info->starting_gain_db +
			((val) * gain_info->gain_step_db);
	} else {
		rx_gain_db = gain_info->starting_gain_db;
	}

	rx_gain->gain_db = rx_gain_db;
out:
	return rc;
}

static void ad_ensm_force_state(struct rf_phy_dev *ad_phy, u8 ensm_state)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	u8 dev_ensm_state;
	int rc;
	u32 val;

	rc = ad_phy_read(ad_phy, REG_DEV_STATE, 1, &val);
	if (rc)
		goto out;

	dev_ensm_state = (val & ENSM_STATE_MASK) >> ENSM_STATE_SHIFT;

	if (dev_ensm_state == ensm_state) {
		dev_err(dev, "Nothing to do, device is already in %d state\n",
			ensm_state);
		goto out;
	}

	phy_info->prev_ensm_state = dev_ensm_state;
	dev_err(dev, "Device is in %x state, forcing to %x\n", dev_ensm_state,
			ensm_state);

	rc = ad_phy_read(ad_phy, REG_ENSM_CONF1, 1, &val);
	if (rc)
		goto out;

	/* Enable control through SPI writes, and take out from
	 * Alert
	 */
	if (val & ENSM_CONF1_ENSM_PIN_CTL_EN) {
		val &= ~ENSM_CONF1_ENSM_PIN_CTL_EN;
		phy_info->ensm_pin_ctl_en = 1;
	} else {
		phy_info->ensm_pin_ctl_en = 0;
	}

	if (dev_ensm_state & dev_ensm_state)
		val &= ~(ENSM_CONF1_TO_ALERT);

	switch (ensm_state) {

	case ENSM_STATE_TX:
		val |= ENSM_CONF1_FORCE_TX_ON;
		break;
	case ENSM_STATE_RX:
		val |= ENSM_CONF1_FORCE_RX_ON;
		break;
	case ENSM_STATE_FDD:
		val |= (ENSM_CONF1_FORCE_TX_ON | ENSM_CONF1_FORCE_RX_ON);
		break;
	default:
		dev_err(dev, "No handleing for forcing %d ensm state\n",
		ensm_state);
		goto out;
	}

	rc = ad_phy_write(ad_phy, REG_ENSM_CONF1, val);
	if (rc)
		dev_err(dev, "Failed to restore state\n");

out:
	return;

}

static void ad_ensm_restore_prev_state(struct rf_phy_dev *ad_phy)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	int rc;
	u32 val;

	rc = ad_phy_read(ad_phy, REG_ENSM_CONF1, 1, &val);
	if (rc) {
		dev_err(dev, "Read REG_ENSM_CONF1 Failed, restore failed\n");
		goto out;
	}

	/* We are restoring state only, so clear State bits first
	 * which might have set while forcing a particular state
	 */
	val &= ~(ENSM_CONF1_FORCE_TX_ON | ENSM_CONF1_FORCE_RX_ON |
			ENSM_CONF1_TO_ALERT);

	switch (phy_info->prev_ensm_state) {

	case ENSM_STATE_TX:
		val |= ENSM_CONF1_FORCE_TX_ON;
		break;
	case ENSM_STATE_RX:
		val |= ENSM_CONF1_FORCE_RX_ON;
		break;
	case ENSM_STATE_FDD:
		val |= (ENSM_CONF1_FORCE_TX_ON | ENSM_CONF1_FORCE_RX_ON);
		break;
	case ENSM_STATE_ALERT:
		val |= ENSM_CONF1_TO_ALERT;
		break;
	case ENSM_STATE_INVALID:
		dev_dbg(dev, "No need to restore, ENSM state wasn't saved\n");
		goto out;
	default:
		dev_dbg(dev, "Could not restore to %d ENSM state\n",
		phy_info->prev_ensm_state);
		goto out;
	}

	rc = ad_phy_write(ad_phy, REG_ENSM_CONF1, val);
	if (rc) {
		dev_err(dev, "Failed to write REG_ENSM_CONF1");
		goto out;
	}

	if (phy_info->ensm_pin_ctl_en) {
		val |= ENSM_CONF1_ENSM_PIN_CTL_EN;
		rc = ad_phy_write(ad_phy, REG_ENSM_CONF1, val);
		if (rc)
			dev_err(dev, "Failed to write REG_ENSM_CONF1");
	}

out:
	return;
}

static int ad_set_rx_gain(struct rf_phy_dev *ad_phy,
		struct rf_rx_gain *rx_gain)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	enum rx_gain_table_name tbl;
	struct rx_gain_info *gain_info;
	u32 val, idx_reg;
	u8 gain_ctl_shift, ensm_state;
	int rc = 0;

	if (rx_gain->ant == 1) {
		gain_ctl_shift = RX1_GAIN_CTL_SHIFT;
		idx_reg = REG_RX1_MGC_FULL_TBL_IDX;

	} else if (rx_gain->ant == 2) {
		gain_ctl_shift = RX2_GAIN_CTL_SHIFT;
		idx_reg = REG_RX2_MGC_FULL_TBL_IDX;
	} else {
		dev_err(dev, "Unkown Rx path %d\n", rx_gain->ant);
		rc = -EINVAL;
		goto out;

	}

	rc = ad_phy_read(ad_phy, REG_AGC_CONF1, 1, &val);
	if (rc) {
		dev_err(dev, "Unable to read REG_AGC_CONF1\n");
		goto out;
	}

	val = (val >> gain_ctl_shift) & RX_GAIN_CTL_MASK;

	if (val != RX_GAIN_CTL_MGC) {
		dev_err(dev, "Rx gain can be set in MGC mode only\n");
		goto out;
	}

	switch (phy_info->freq_band) {
	case 1:
	case 7:
		tbl = FULL_TBL_1300_4000_MHZ;
		break;
	case 13:
		tbl = FULL_TBL_200_1300_MHZ;
		break;
	default:
		dev_err(dev, "unkown band for rx gain read, %d\n",
			phy_info->freq_band);
		rc = -EINVAL;
		goto out;
	}

	gain_info = &phy_info->rx_gain[tbl];
	if ((rx_gain->gain_db < gain_info->starting_gain_db) ||
		(rx_gain->gain_db > gain_info->max_gain_db)) {

		dev_err(dev, "Invalid gain %d, supported range [%d - %d]\n",
			rx_gain->gain_db, gain_info->starting_gain_db,
			gain_info->max_gain_db);
		rc = -EINVAL;
		goto out;

	}
	rc = ad_phy_read(ad_phy, idx_reg, 1, &val);
	if (rc) {
		dev_err(dev, "Unable to read gain tbl idx reg: %d\n", idx_reg);
		goto out;
	}

	if (phy_info->net_mode == LTE_FDD)
		ensm_state = ENSM_STATE_FDD;
	else
		ensm_state = ENSM_STATE_RX;

	/* RX must be enabled while changing Gain */
	ad_ensm_force_state(ad_phy, ensm_state);

	val = val & ~FULL_TBL_IDX_MASK;
	val |= ((rx_gain->gain_db - gain_info->starting_gain_db) /
		gain_info->gain_step_db) + gain_info->idx_step_offset;
	rc = ad_phy_write(ad_phy, idx_reg, val);

	/* Restore is done intentionally before checking rc, because
	 * we need to restore PHY to previous state even if write failed
	 */
	ad_ensm_restore_prev_state(ad_phy);

	if (rc) {
		dev_err(dev, "Unable to write gain tbl idx reg: %d\n", idx_reg);
		goto out;
	}

out:
	return rc;

}

void ad_init_gain_info(struct rx_gain_info *rx_gain,
	enum rx_gain_table_type type, int starting_gain,
	int max_gain, int gain_step, int max_idx, int idx_offset)
{
	rx_gain->tbl_type = type;
	rx_gain->starting_gain_db = starting_gain;
	rx_gain->max_gain_db = max_gain;
	rx_gain->gain_step_db = gain_step;
	rx_gain->max_idx = max_idx;
	rx_gain->idx_step_offset = idx_offset;
}

int ad_init_gain_tables(struct rf_phy_dev *ad_phy)
{
	struct rx_gain_info *rx_gain;
	struct ad_dev_info *phy_info = ad_phy->priv;

	/* Intialize Meta data according to default gain tables
	 * of AD9631. Changing/Writing of gain tables is not
	 * supported yet.
	 */
	rx_gain = &phy_info->rx_gain[FULL_TBL_200_1300_MHZ];
	ad_init_gain_info(rx_gain, RXGAIN_FULL_TBL, 1, 77, 1,
		RXGAIN_FULL_TBL_MAX_IDX, 0);

	rx_gain = &phy_info->rx_gain[FULL_TBL_1300_4000_MHZ];
	ad_init_gain_info(rx_gain, RXGAIN_FULL_TBL, -4, 71, 1,
		RXGAIN_FULL_TBL_MAX_IDX, 1);

	rx_gain = &phy_info->rx_gain[FULL_TBL_4000_6000_MHZ];
	ad_init_gain_info(rx_gain, RXGAIN_FULL_TBL, -10, 62, 1,
		RXGAIN_FULL_TBL_MAX_IDX, 4);

	return 0;
}

int ad_set_gain_ctrl_mode(struct rf_phy_dev *ad_phy,
		struct rf_gain_ctrl *gain_ctrl)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	int rc = 0;
	u32 val, gain_ctl_shift, mode;

	rc = ad_phy_read(ad_phy, REG_AGC_CONF1, 1, &val);
	if (rc) {
		dev_err(dev, "Unable to read AGC config1 register: %x\n",
				REG_AGC_CONF1);
		goto out;
	}

	switch (gain_ctrl->mode) {
	case MGC:
		mode = RX_GAIN_CTL_MGC;
		break;
	case FASTATTACK_AGC:
		mode = RX_GAIN_CTL_AGC_FAST_ATK;
		break;
	case SLOWATTACK_AGC:
		mode = RX_GAIN_CTL_AGC_SLOW_ATK;
		break;
	case HYBRID_AGC:
		mode = RX_GAIN_CTL_AGC_SLOW_ATK_HYBD;
		break;
	default:
		rc = -EINVAL;
		goto out;
	}

	if (gain_ctrl->ant == 1) {
		gain_ctl_shift = RX1_GAIN_CTL_SHIFT;
	} else if (gain_ctrl->ant == 2) {
		gain_ctl_shift = RX2_GAIN_CTL_SHIFT;
	} else {
		dev_err(dev, "Unknown Rx path %d\n", gain_ctrl->ant);
		rc = -EINVAL;
		goto out;
	}

	rc = ad_en_dis_rx(ad_phy, gain_ctrl->ant, RX_DISABLE);
	if (rc) {
		dev_err(dev, "Unable to disable rx%d\n", gain_ctrl->ant);
		goto out;
	}

	val &= ~(RX_GAIN_CTL_MASK << gain_ctl_shift);
	val |= mode << gain_ctl_shift;
	if (mode == RX_GAIN_CTL_AGC_SLOW_ATK_HYBD)
		val |= SLOW_ATK_HYBD_BIT_EN;
	else
		val &= ~SLOW_ATK_HYBD_BIT_EN;
	rc = ad_phy_write(ad_phy, REG_AGC_CONF1, val);
	if (rc) {
		dev_err(dev, "Unable to write AGC config1 register: %x\n",
				REG_AGC_CONF1);
		goto out;
	}

	rc = ad_en_dis_rx(ad_phy, gain_ctrl->ant, RX_ENABLE);
out:
	return rc;
}

int ad_program_synthesizer(struct rf_phy_dev *ad_phy,
		struct rf_sniff_params *sniff_params)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	int ret = 0, i = 1;
	unsigned int block_scalar, vco_divider_shift, reg_offset;
	unsigned long f_rfpll, f_ref, n_int, n_frac;
	unsigned long freq = MAX_CARRIER_FREQ_KHZ;
	unsigned long long f_rfpll_hz;
	u8 divider_mask, vco_divider = i - 1;
	u32 val[2], temp;

	if (sniff_params->carrier_freq > MAX_CARRIER_FREQ_KHZ ||
			sniff_params->carrier_freq < MIN_CARRIER_FREQ_KHZ) {
		dev_err(dev, "Invalid carrier freq for sniffing %lu KHz\n",
			sniff_params->carrier_freq);
		ret = -EINVAL;
		goto out;
	}

	/*
	 * RFPLL works in freq range of 6-12 Ghz, So to convert carrier freq
	 * in this range vco divider is programmed accordingly.
	 */
	while (freq > MIN_CARRIER_FREQ_KHZ) {
		if (sniff_params->carrier_freq > (freq >> 1)) {
			vco_divider = (u8) i - 1;
			break;
		} else {
			freq = freq >> 1;
			i++;
		}
	}

	/*
	 * To get the f_rfpll from carrier freq and appropriate vco divider:
	 * f_rfpll = carrier_freq * (2^(vco_divider + 1))
	 */
	f_rfpll = sniff_params->carrier_freq << (vco_divider + 1);

	ad_phy_read(ad_phy, REG_REF_DIVIDE_CONFIG1, 2, val);
	if (sniff_params->rf_synth == RF_RX) {
		/* For RX synthesizer block scalar is represented by
		 * REG_REF_DIVIDE_CONFIG1[D0] and REG_REF_DIVIDE_CONFIG2[D7] */
		block_scalar = ((val[0] & RX_SCALER_MASK1) << 1) +
				((val[1] & RX_SCALER_MASK2) >> 7);
		reg_offset = RX_REG_OFFSET;
		/* For RX synthesizer, VCO divider is programmed in
		 * REG_RFPLL_DIVIDER[D3:D0] */
		vco_divider_shift = 0;
		divider_mask = 0xF0;
	} else if (sniff_params->rf_synth == RF_TX) {
		/* For TX synthesizer block scalar is represented by
		 * REG_REF_DIVIDE_CONFIG2[D3:D2] */
		block_scalar = (val[1] & TX_SCALER_MASK) >> 2;
		reg_offset = TX_REG_OFFSET;
		/* For TX synthesizer, VCO divider is programmed in
		 * REG_RFPLL_DIVIDER[D7:D4] */
		vco_divider_shift = 4;
		divider_mask = 0x0F;
	} else {
		dev_err(dev, "RF_BBPLL synth programming not supported\n");
		ret = -EINVAL;
		goto out;
	}

	switch (block_scalar) {
	case 0:
		f_ref = DEVICE_REF_FREQ_KHZ;
		break;
	case 1:
		f_ref = DEVICE_REF_FREQ_KHZ / 2;
		break;
	case 2:
		f_ref = DEVICE_REF_FREQ_KHZ / 4;
		break;
	case 3:
		f_ref = DEVICE_REF_FREQ_KHZ * 2;
		break;
	default:
		f_ref = DEVICE_REF_FREQ_KHZ * 2;
	}

	f_rfpll_hz = (unsigned long long)f_rfpll * 1000;
	i = 0;
	while (i < phy_info->synth_table.count &&
		(phy_info->synth_table.params[i][VCO_Frequency] > f_rfpll_hz))
		i++;

	/* n_int = floor(f_rfpll / f_ref) */
	n_int = f_rfpll / f_ref;

	/* Divide f_rfpll and f_ref by 100 to avoid 64 bit operations */
	f_rfpll /= 100;
	f_ref /= 100;

	/*
	 * n_frac = Round(8388593 * ((f_rfpll/f_ref - n_int)))
	 * 8388593 is a constant for RX and TX synthesizer
	 * and it is denoted as rfpll modulus
	 */
	n_frac = (RFPLL_MODULUS * (f_rfpll - (n_int * f_ref)) +
			(f_ref / 2)) / f_ref;

	/* Program various registers according to the synthesizer look up
	 * table and f_rfpll(vco frequency) */

	ad_phy_read(ad_phy, REG_VCO_OUTPUT + reg_offset, 1, &temp);
	temp = (temp & MASK_VCO_OUTPUT) |
		phy_info->synth_table.reg_vals[i][VCO_Output_Level];
	ad_phy_write(ad_phy, REG_VCO_OUTPUT + reg_offset, temp);

	ad_phy_read(ad_phy, REG_VCO_VARACTOR + reg_offset, 1, &temp);
	temp = (temp & MASK_VCO_VARACTOR) |
		phy_info->synth_table.reg_vals[i][VCO_Varactor];
	ad_phy_write(ad_phy, REG_VCO_VARACTOR + reg_offset, temp);

	ad_phy_read(ad_phy, REG_VCO_BIAS1 + reg_offset, 1, &temp);
	temp = (temp & MASK_VCO_BIAS1) |
		phy_info->synth_table.reg_vals[i][VCO_Bias_Ref];
	temp |= (phy_info->synth_table.reg_vals[i][VCO_Bias_Tcf] <<
		SHIFT_VCO_BIAS_TCF);
	ad_phy_write(ad_phy, REG_VCO_BIAS1 + reg_offset, temp);

	ad_phy_read(ad_phy, REG_FORCE_VCO_TUNE1 + reg_offset, 1, &temp);
	temp = (temp & MASK_FORCE_VCO_TUNE1) |
		(phy_info->synth_table.reg_vals[i][VCO_Cal_Offset] <<
		SHIFT_VCO_CAL_OFFSET);
	ad_phy_write(ad_phy, REG_FORCE_VCO_TUNE1 + reg_offset, temp);

	ad_phy_read(ad_phy, REG_VCO_VARACTOR_CONTROL1 + reg_offset, 1, &temp);
	temp = (temp & MASK_VCO_VARACTOR_CONTROL1) |
		phy_info->synth_table.reg_vals[i][VCO_Varactor_Reference];
	ad_phy_write(ad_phy, REG_VCO_VARACTOR_CONTROL1 + reg_offset, temp);

	ad_phy_read(ad_phy, REG_CP_CURRENT + reg_offset, 1, &temp);
	temp = (temp & MASK_CP_CURRENT) |
		phy_info->synth_table.reg_vals[i][Charge_Pump_Current];
	ad_phy_write(ad_phy, REG_CP_CURRENT + reg_offset, temp);

	temp =  (phy_info->synth_table.reg_vals[i][Loop_Filter_C2] <<
			SHIFT_LOOP_FILTER_C2);
	temp |= phy_info->synth_table.reg_vals[i][Loop_Filter_C1];
	ad_phy_write(ad_phy, REG_LOOP_FILTER1 + reg_offset, temp);

	temp =  (phy_info->synth_table.reg_vals[i][Loop_Filter_R1] <<
			SHIFT_LOOP_FILTER_R1);
	temp |= phy_info->synth_table.reg_vals[i][Loop_Filter_C3];
	ad_phy_write(ad_phy, REG_LOOP_FILTER2 + reg_offset, temp);

	ad_phy_read(ad_phy, REG_LOOP_FILTER3 + reg_offset, 1, &temp);
	temp = (temp & MASK_LOOP_FILTER3) |
		phy_info->synth_table.reg_vals[i][Loop_Filter_R3];
	ad_phy_write(ad_phy, REG_LOOP_FILTER3 + reg_offset, temp);

	temp = (n_frac & MASK_FRAC_B0) >> SHIFT_FRAC_B0;
	ad_phy_write(ad_phy, REG_FRACTIONAL_BYTE0 + reg_offset, temp);
	temp = (n_frac & MASK_FRAC_B1) >> SHIFT_FRAC_B1;
	ad_phy_write(ad_phy, REG_FRACTIONAL_BYTE1 + reg_offset, temp);
	temp = (n_frac & MASK_FRAC_B2) >> SHIFT_FRAC_B2;
	ad_phy_write(ad_phy, REG_FRACTIONAL_BYTE2 + reg_offset, temp);

	/* n_integer is programmed in the last so that calibration is
	 * triggered*/
	ad_phy_read(ad_phy, REG_INTEGER_BYTE1 + reg_offset, 1, &temp);
	temp = (temp & MASK_INTEGER_REG1) |
		((n_int & MASK_INT_B1) >> SHIFT_INT_B1);
	ad_phy_write(ad_phy, REG_INTEGER_BYTE1 + reg_offset, temp);

	temp = (n_int & MASK_INT_B0);
	ad_phy_write(ad_phy, REG_INTEGER_BYTE0 + reg_offset, temp);

	ad_phy_read(ad_phy, REG_RFPLL_DIVIDER, 1, &temp);
	temp = (temp & divider_mask) | (vco_divider << vco_divider_shift);
	ad_phy_write(ad_phy, REG_RFPLL_DIVIDER, temp);

out:
	return ret;
}

int ad_config_sniff(struct rf_phy_dev *ad_phy,
	struct rf_sniff_params *sniff_params)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	struct rf_init_params *dev_params = &sniff_params->dev_params;
	int ret = 0;

	switch (dev_params->fq_band) {
	case 1:
	case 7:
		/* For High band Pull up PA_EN in order to avoid
		 * Tx noise leakage into Rx1C reciever of high band.
		 */
		if (ad_phy->ctrl_dev->soc == SOC_9132)
			config_rf_regulator(1, 0, ad_phy->ctrl_dev->dev_idx);
		else {
			gpio_set_value(phy_info->pa_en_gpio, 1);
			gpio_set_value(phy_info->lna_en_gpio, 0);
		}
		break;
	case 13:
		/* For Low band Pull down PA_EN in order to avoid
		 * Tx noise leakage into Rx1C reciever of low band.
		 */
		if (ad_phy->ctrl_dev->soc == SOC_9132)
			config_rf_regulator(0, 1, ad_phy->ctrl_dev->dev_idx);
		else {
			gpio_set_value(phy_info->pa_en_gpio, 0);
			gpio_set_value(phy_info->lna_en_gpio, 1);
		}
		break;
	default:
		dev_dbg(dev, "Invalid band for sniffing %d\n",
			dev_params->fq_band);
		ret = -EINVAL;
		goto out;
	}

	phy_info->freq_band = dev_params->fq_band;
	phy_info->net_mode = dev_params->mode;
	ret = ad_program_synthesizer(ad_phy, sniff_params);
out:
	return ret;
}

static int ad_set_band(struct rf_phy_dev *ad_phy, unsigned int band)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	int rc = 0;

	switch (band) {
	case 1:
	case 7:
		if (ad_phy->ctrl_dev->soc == SOC_9132)
			config_rf_regulator(0, 0, ad_phy->ctrl_dev->dev_idx);
		else {
			gpio_set_value(phy_info->pa_en_gpio, 0);
			gpio_set_value(phy_info->lna_en_gpio, 0);
		}
		break;

	case 13:
	case 40:
		if (ad_phy->ctrl_dev->soc == SOC_9132)
			config_rf_regulator(1, 1, ad_phy->ctrl_dev->dev_idx);
		else {
			gpio_set_value(phy_info->pa_en_gpio, 1);
			gpio_set_value(phy_info->lna_en_gpio, 1);
		}
		break;

	case 41:
		if (ad_phy->ctrl_dev->soc == SOC_9132)
			config_rf_regulator(0, 1, ad_phy->ctrl_dev->dev_idx);
		else {
			gpio_set_value(phy_info->pa_en_gpio, 0);
			gpio_set_value(phy_info->lna_en_gpio, 1);
		}
		break;

	default:
		rc = -EINVAL;
		goto out;
	}

	phy_info->freq_band = band;

out:
	return rc;
}

int ad_init(struct rf_phy_dev *ad_phy,
		struct rf_init_params *params)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	int rc = 0;
	u32 default_rssi_delay[4] = {RSSI_DELAY_5MHZ,
					RSSI_DELAY_10MHZ,
					RSSI_DELAY_15MHZ,
					RSSI_DELAY_20MHZ};

	rc = ad_set_band(ad_phy, params->fq_band);
	if (rc) {
		dev_err(dev, "Failed to set band %d\n", rc);
		goto out;
	}

	rc = ad_phy_write(ad_phy, RSSI_DELAY_REG,
			default_rssi_delay[params->bw]);
	if (rc) {
		dev_err(dev, "Unable to write rssi_delay reg: %d\n",
				RSSI_DELAY_REG);
		goto out;
	}

	phy_info->net_mode = params->mode;
	phy_info->prev_ensm_state = ENSM_STATE_INVALID;

out:
	return rc;
}

int ad_phy_stop(struct rf_phy_dev *ad_phy)
{
	struct ad_dev_info *phy_info = ad_phy->priv;

	ad_en_dis_tx(ad_phy, 1, TX_DISABLE);
	ad_en_dis_tx(ad_phy, 2, TX_DISABLE);

	kfree(phy_info->synth_table.params);
	kfree(phy_info->synth_table.reg_vals);
	phy_info->synth_table.params = NULL;
	phy_info->synth_table.reg_vals = NULL;

	return 0;
}

int ad_phy_start(struct rf_phy_dev *ad_phy)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;

	dev_dbg(dev, "Reading BBPLL locked status.\n");

	if (check_cal_done(ad_phy, REG_CH1_OVERFLOW,
			MASK_BBPLL_LOCK, 1))
		return 0;
	else {
		dev_err(dev, "BBPLL not locked.\n");
		return -EBUSY;
	}
}

int ad_phy_read(struct rf_phy_dev *ad_phy, u32 start,
		u32 count, u32 *buff)
{
	u32 data;
	u16 rx_cmd, read_addr;
	int i = 0, j = 0, rc;
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;

	/* RFdev f/w provides start address as u32, but
	 * ADI has only 12 bits regs, so downsize it to 16 bits
	 */
	read_addr =  (u16)(start + count - 1);
	j = count - 1;

	while (count > MAX_READ_TRANSACTION_SIZE) {
		rx_cmd = create_ad_phy_cmd(read_addr, SPI_READ,
			MAX_READ_TRANSACTION_SIZE);

		rc = spi_read_transaction(ad_phy, (u8 *)&rx_cmd,
				RXBUF_SIZE);
		if (rc)
			goto out;

		for (i = COMMAND_LEN; i < RXBUF_SIZE ; i++) {
			data = phy_info->rx_buf[i];
			buff[j--] = data;
		}

		count -= MAX_READ_TRANSACTION_SIZE;
		read_addr = read_addr - MAX_READ_TRANSACTION_SIZE;
	}

	rx_cmd = create_ad_phy_cmd(read_addr, SPI_READ, count);

	rc = spi_read_transaction(ad_phy, (u8 *)&rx_cmd, (COMMAND_LEN + count));
	if (rc)
		goto out;

	for (i = COMMAND_LEN; i < (COMMAND_LEN + count); i++) {
		data = phy_info->rx_buf[i];
		buff[j--] = data;
	}

out:
	dev_dbg(dev, "Failed to read %d regs, start addr %x\n",
		count, start);
	return rc;
}

int ad_phy_write(struct rf_phy_dev *ad_phy, u32 reg,
		u32 data)
{
	int rc;
	u32 cmd = 0;
	u16 ad_reg;
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;

	dev_dbg(dev, "AD's register write call.\n");
	/* Rfdev f/w sends reg as 32 bit address, but ADI has
	 * only 12 bit register addresses, thus downsize to
	 * u16
	 */
	ad_reg = (u16) reg;
	cmd = create_ad_phy_cmd(ad_reg, SPI_WRITE, 1);
	cmd = (cmd << 16) & COMMAND_MASK;
	cmd = cmd | (data << 8);
	rc = spi_write_transaction(ad_phy, (u8 *)&cmd, TRANSACTION_BYTES);

	return rc;
}

int ad_phy_get_dac_value(struct rf_phy_dev *ad_phy,
			u32 correction_type , u32 *buff)
{
	u32 reg = 0;
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;

	dev_dbg(dev, "DAC's register read call.\n");
	switch (correction_type) {
	case COARSE_CORRECTION:
		reg = DAC1_CONFIG;
		break;
	case FINE_CORRECTION:
		reg = DAC1_WORD;
		break;
	default:
		return -EINVAL;
	}

	return ad_phy_read(ad_phy, reg, 1, buff);
}

int ad_phy_dac_correction(struct rf_phy_dev *ad_phy,
			struct rif_dac_params *dac_params)
{
	u32 reg = 0;
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;

	dev_dbg(dev, "DAC's register write call.\n");
	switch (dac_params->correction_type) {
	case COARSE_CORRECTION:
		reg = DAC1_CONFIG;
		break;
	case FINE_CORRECTION:
		reg = DAC1_WORD;
		break;
	default:
		return -EINVAL;
	}

	return ad_phy_write(ad_phy, reg, dac_params->correction_value);
}

int ad_set_tx_atten(struct rf_phy_dev *ad_phy, u32 tx_if, u32 tx_atten)
{
	u32 reg_addr1, reg_addr2, reg_val;
	int rc = 0;

	if (tx_if == 1) {
		reg_addr1 = TX1_ATTEN0;
		reg_addr2 = TX1_ATTEN1;
	} else if (tx_if == 2) {
		reg_addr1 = TX2_ATTEN0;
		reg_addr2 = TX2_ATTEN1;
	} else {
		rc = -EINVAL;
		goto out;
	}

	reg_val = tx_atten & ATTEN0_MASK;
	ad_phy_write(ad_phy, reg_addr1, reg_val);
	reg_val = tx_atten & ATTEN1_MASK;
	reg_val = reg_val >> MSB_SHIFT;
	ad_phy_write(ad_phy, reg_addr2, reg_val);
	ad_phy->tx_atten[tx_if - 1] = tx_atten;

out:
	return rc;
}

int ad_en_dis_tx(struct rf_phy_dev *ad_phy, u32 tx_if, u32 tx_cmd)
{
	u32 mask, reg_val;
	ad_phy_read(ad_phy, TX_ENABLE_REG, 1,  &reg_val);
	if (tx_if == 1)
		mask = TX1_ENABLE_MASK;
	else if (tx_if == 2)
		mask = TX2_ENABLE_MASK;
	else
		return -EFAULT;

	if (tx_cmd == TX_ENABLE)
		reg_val |= mask;
	else if (tx_cmd == TX_DISABLE)
		reg_val &= ~mask;
	else
		return -EFAULT;
	return  ad_phy_write(ad_phy, TX_ENABLE_REG, reg_val);
}

int ad_en_dis_rx(struct rf_phy_dev *ad_phy, u32 rx_if, u32 rx_cmd)
{
	u32 mask, reg_val;
	ad_phy_read(ad_phy, RX_ENABLE_REG, 1,  &reg_val);
	if (rx_if == 1)
		mask = RX1_ENABLE_MASK;
	else if (rx_if == 2)
		mask = RX2_ENABLE_MASK;
	else
		return -EFAULT;

	if (rx_cmd == RX_ENABLE)
		reg_val |= mask;
	else if (rx_cmd == RX_DISABLE)
		reg_val &= ~mask;
	else
		return -EFAULT;
	return  ad_phy_write(ad_phy, RX_ENABLE_REG, reg_val);
}

static int ad_read_rssi(struct rf_phy_dev *ad_phy, struct rf_rssi *rssi)
{
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &phy_info->ad_spi->dev;
	u32 reg_val_buf[6], weight[4], total_dur, val, temp;
	u32 subframe_size[4] = {SUBFRAME_SIZE_5MHZ,
				SUBFRAME_SIZE_10MHZ,
				SUBFRAME_SIZE_15MHZ,
				SUBFRAME_SIZE_20MHZ};
	u8 dur_buf[4] = {0}, i, j;
	int rc;

	if (rssi->duration > 1) {
		dev_err(dev, "RSSI measurement duration > 1ms not supported\n");
		rc = -EINVAL;
		goto out;
	}

	rc = ad_phy_read(ad_phy, RSSI_CONFIG_REG, 1, &val);
	val &= RSSI_MEAS_MODE_MASK;
	val |= RSSI_GAIN_CHANGE_EN_AGC_MODE;
	rc |= ad_phy_write(ad_phy, RSSI_CONFIG_REG, val);
	if (rc) {
		dev_err(dev, "Unable to read/write rssi config reg\n");
		goto out;
	}

	temp = subframe_size[ad_phy->ctrl_dev->bw] * rssi->duration;
	for (i = 0, j = 0; temp != 0 && j < 4; i++) {
		if (temp & 0x01)
			dur_buf[j++] = i;
		temp = temp >> 1;
	}

	total_dur = (1 << dur_buf[0]) + (1 << dur_buf[1]) +
			(1 << dur_buf[2]) + (1 << dur_buf[3]);
	weight[0] = RSSI_MAX_WEIGHT * (1 << dur_buf[0]) / total_dur;
	weight[1] = RSSI_MAX_WEIGHT * (1 << dur_buf[1]) / total_dur;
	weight[2] = RSSI_MAX_WEIGHT * (1 << dur_buf[2]) / total_dur;
	weight[3] = RSSI_MAX_WEIGHT * (1 << dur_buf[3]) / total_dur;
	rc = ad_phy_write(ad_phy, RSSI_MEAS_DUR_10_REG,
			((dur_buf[1] << 4) | dur_buf[0])) ||
		ad_phy_write(ad_phy, RSSI_MEAS_DUR_32_REG,
			((dur_buf[3] << 4) | dur_buf[2])) ||
		ad_phy_write(ad_phy, RSSI_WEIGHT0_REG, weight[0]) ||
		ad_phy_write(ad_phy, RSSI_WEIGHT1_REG, weight[1]) ||
		ad_phy_write(ad_phy, RSSI_WEIGHT2_REG, weight[2]) ||
		ad_phy_write(ad_phy, RSSI_WEIGHT3_REG, weight[3]);
	if (rc) {
		dev_err(dev, "Unable to write rssi measurement duration\n");
		goto out;
	}

	rc = ad_phy_read(ad_phy, RSSI_READBACK_REG,
			sizeof(reg_val_buf) / sizeof(u32), reg_val_buf);

	if (rssi->ant == 1) {
		rssi->symbol = RSSI_RESOLUTION *
				((reg_val_buf[0] << LSB_SHIFT) +
				 (reg_val_buf[4] & RSSI_LSB_MASK1));
		rssi->preamble = RSSI_RESOLUTION *
				((reg_val_buf[1] << LSB_SHIFT) +
				 (reg_val_buf[5] & RSSI_LSB_MASK1));
	} else if (rssi->ant == 2) {
		rssi->symbol = RSSI_RESOLUTION *
				((reg_val_buf[2] << LSB_SHIFT) +
				 ((reg_val_buf[4] & RSSI_LSB_MASK2) >> 1));
		rssi->preamble = RSSI_RESOLUTION *
				((reg_val_buf[3] << LSB_SHIFT) +
				 ((reg_val_buf[5] & RSSI_LSB_MASK2) >> 1));
	} else
		rc = -EFAULT;

	rssi->multiplier = RSSI_MULTIPLIER;

out:
	return rc;
}

static int ad_phy_remove(struct spi_device *spi)
{
	int ret = 0;
	struct device *dev = &spi->dev;
	struct rf_phy_dev *ad_phy = dev_get_drvdata(&spi->dev);
	struct ad_dev_info *phy_info = ad_phy->priv;

	dev_dbg(dev, "AD9361 PHY module uninstalled\n");

	gpio_free(phy_info->reset_gpio);
	gpio_free(phy_info->pa_en_gpio);
	gpio_free(phy_info->lna_en_gpio);

	ret = unregister_rf_phy_dev(ad_phy);
	if (ret < 0) {
		dev_dbg(dev, "unregister_rf_phy_dev failed.\n");
		return ret;
	}

	ret = free_rf_phy_dev(ad_phy);

	return ret;
}

static int check_revision(struct rf_phy_dev *ad_phy)
{
	u32 val;
	u8 product_id, rev;
	int rc = 0;
	struct ad_dev_info *phy_info = ad_phy->priv;
	struct device *dev = &(phy_info->ad_spi->dev);

	ad_phy_read(ad_phy, PRODUCT_CODE_REG, 1, &val);

	product_id = (u8) val & PRODUCT_ID_MASK;
	if (product_id != PRODUCT_ID_9361) {
		dev_err(dev, "Unsuported product id 0x%x\n", product_id);
		rc = -ENODEV;
		goto out;
	}

	rev = (u8) val & REV_MASK;
	if (rev != 2) {
		dev_err(dev, "Unsuported AD9361 revision 0x%x\n", rev);
		rc = -ENODEV;
		goto out;
	}

	dev_info(dev, "Detected AD9361 Rev 0x%x\n", rev);
out:
	return rc;
}

static int ad_phy_probe(struct spi_device *spi)
{
	int ret = 0;
	int gpio, gpio_inval = ~0;
	static struct rf_phy_dev *ad_phy;
	struct device *dev = &spi->dev;
	struct ad_dev_info *phy_info;
	struct device_node *np = spi->dev.archdata.of_node, *child;
	u32 flags;
	char *gpio_name;

	ad_phy = allocate_rf_phy_dev(sizeof(struct ad_dev_info), GFP_KERNEL);
	if (!ad_phy) {
		dev_dbg(dev, "Failed to allocate rf_phy_dev\n");
		return -ENOMEM;
	}

	phy_info = (struct ad_dev_info *) ad_phy->priv;

	phy_info->ad_spi = spi;
	spi->bits_per_word = 8;

	ad_phy->ops = &ad_phy_ops;

	strncpy(&ad_phy->name[0], "ad9361", sizeof(ad_phy->name));
	ad_phy->phy_id = (u32) np;

	phy_info->reset_gpio = gpio_inval;
	phy_info->pa_en_gpio = gpio_inval;
	phy_info->lna_en_gpio = gpio_inval;

	for_each_child_of_node(np, child) {

		gpio_name = (char *) of_get_property(child, "label", NULL);
		if (!gpio_name)
			gpio_name = (char *) child->name;

		gpio = of_get_gpio_flags(child, 0, &flags);
		if (gpio < 0) {
			dev_err(dev, "Could not get gpio for %s, err %d\n",
				gpio_name, gpio);
			ret = gpio;
			goto out;
		}

		if (!strncmp(gpio_name, "reset", RIF_NAME_SIZE))
			phy_info->reset_gpio = gpio;
		else if (!strncmp(gpio_name, "pa_en", RIF_NAME_SIZE))
			phy_info->pa_en_gpio = gpio;
		else if (!strncmp(gpio_name, "lna_en", RIF_NAME_SIZE))
			phy_info->lna_en_gpio = gpio;
		else
			continue;

		ret = gpio_request(gpio, dev_name(dev));
		if (ret) {
			dev_dbg(dev, "gpio_request failed, gpio %s[%d]",
					gpio_name, gpio);
			goto out;
		}

		ret = gpio_direction_output(gpio, 0);
		if (ret) {
			dev_dbg(dev,
				"gpio_direction_output failed, gpio %s[%d]",
				gpio_name, gpio);
			goto out;
		}

	}

	if ((phy_info->reset_gpio == gpio_inval) ||
		(phy_info->lna_en_gpio == gpio_inval) ||
		(phy_info->pa_en_gpio == gpio_inval)) {

		dev_err(dev, "Control GPIOs not intialized in dtb\n");
		ret = -EINVAL;
		goto out;
	}

	if (!reset_status) {
		/* Toggle reset gpio to reset AD9361 */
		gpio_set_value(phy_info->reset_gpio, 0);
		gpio_set_value(phy_info->reset_gpio, 1);
		reset_status = true;
	}

	dev_set_drvdata(&spi->dev, ad_phy);

	ret = check_revision(ad_phy);
	if (ret)
		goto out;

	ad_init_gain_tables(ad_phy);

	/* If revision code is correct, Registering phy as a rf phy device. */
	ret = register_rf_phy_dev(ad_phy);
	if (ret) {
		dev_dbg(dev, "register_rf_phy_dev failed.\n");
		goto out;
	}

	return ret;

out:
	if (phy_info->reset_gpio != gpio_inval)
		gpio_free(phy_info->reset_gpio);

	if (phy_info->pa_en_gpio != gpio_inval)
		gpio_free(phy_info->pa_en_gpio);

	if (phy_info->lna_en_gpio != gpio_inval)
		gpio_free(phy_info->lna_en_gpio);

	free_rf_phy_dev(ad_phy);
	return ret;
}

static struct spi_driver ad_phy_driver = {
	.driver = {
		.name = DRV_NAME,
		.bus = &spi_bus_type,
		.owner = THIS_MODULE,
		},
	.probe = ad_phy_probe,
	.remove = __devexit_p(ad_phy_remove),

};

static int __init ad_phy_init(void)
{
	int ret;


	ret = spi_register_driver(&ad_phy_driver);
	if (ret != 0) {
		pr_err("%s spi_register_driver failed with err %x\n",
			__func__, ret);
	}

	return ret;
}

static void __exit ad_phy_exit(void)
{
	spi_unregister_driver(&ad_phy_driver);
}

module_init(ad_phy_init);
module_exit(ad_phy_exit);
