/*
 * RF device framework
 *
 * Author: pankaj chauhan <pankaj.chauhan@freescale.com>
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#ifndef __RFDEV_H__
#define __RFDEV_H__

#include <linux/time.h>
#include <linux/rfdev_ioctl.h>
#include <linux/list.h>
#include <asm/atomic.h>

#define TTI_LOG_SIZE	10

struct rf_stats {
	unsigned int tti_count;
	struct timeval tti_ts[TTI_LOG_SIZE];
};

struct rf_dev_params {
	unsigned int ants;
	unsigned int symbol_len;
	unsigned int long_cp;
	unsigned int cp0_len;
	unsigned int cp1_len;
	/*Frame structure*/
	unsigned int subfrm_per_frm;
	unsigned int slots_per_subfrm;
	unsigned int chips_per_slot;
	unsigned int symbols_per_slot;
	/*TDD frame structure*/
	char tdd_dl_ul_conf[LTE_SUBFRM_PER_FRM];
	unsigned int tdd_dwpts;
	unsigned int tdd_uppts;
};

struct rf_dev_sniffer {
	enum rf_network_mode 	net_mode;
	enum rf_txrxmode 	tx_rxmode;
	enum rf_timer_src	timing_src;
	enum rf_band_width 	bw;
	unsigned long carrier_freq;
};

struct rf_channel_stats {
	int tx_frame_count;
	int tx_err;
	int rx_frame_count;
	int rx_err;
};


/**
 * struct rf_mmaped_bufs - Memory mapped channel buffers,
 * They are used when channel is opened using RF_CHANNEL_MMAPED
 * flag. User space will provide mmaped buffer in this case.
 *
 * @ tx/rx_base : Virtual address of tx/rx buffer base. User
 *   space will provide the physical address while opening channel
 *   and rfdev frame work may ioremap it to store virtual address
 *   in tx/rx_base.
 * @ tx/rx_phys : Physical address of tx/rx buffer base.
 * @ tx_curr_idx : Current index at which tx data has been written.
 * @ rx_curr_idx : Current index at which rx data has been written.
 *
 */
struct rf_mmaped_bufs {
	u32 tx_base;
	u32 rx_base;
	u32 tx_base_phys;
	u32 rx_base_phys;
	unsigned int tx_curr_idx;
	unsigned int rx_curr_idx;
};

struct rf_channel {
	unsigned int ant;
	struct rf_ctrl_dev *dev;
	void *priv;
	raw_spinlock_t	lock;
	u32 flags;
	unsigned int tx_buf_size;
	unsigned int rx_buf_size;
	union {
		struct rf_mmaped_bufs mmap_bufs;
		/* For non mmaped bufs, buf queues
		 * should be added here in this union
		 * only
		 */
	} bufs;
	struct rf_channel_stats stats;
};

struct rf_event_handler {
	struct list_head list;
	u16 evt_mask;
	struct pid *pid;
	struct siginfo siginfo;
};

enum soc_type {
	SOC_9131,
	SOC_9132,
	SOC_INVALID
};

struct rf_ctrl_dev {
	char			name[RIF_NAME_SIZE];
	enum rf_state		state;
	enum rf_state		old_state;
	enum rf_network_mode 	net_mode;
	enum rf_txrxmode 	tx_rxmode;
	enum rf_timer_src	timing_src;
	enum rf_band_width 	bw;
	enum soc_type 		soc;
	struct rf_dev_params	dev_params;
	struct rf_channel	*channels[MAX_NUM_ANTENNAS];
	struct list_head	event_handler_list;
	raw_spinlock_t		event_handler_lock;
	struct rf_dev_sniffer   sniffer;
	int			sniff_enabled;
	int			data_chans_enabled;
	int			dev_idx;
	u32			phy_id;
	void 			*priv;
	struct rf_ctrl_ops	*ops;
	struct rf_phy_dev	*phy;
	atomic_t		ref;
	raw_spinlock_t		wait_q_lock;
	wait_queue_head_t	wait_q;
	struct list_head	list;
	struct mutex		lock;
	struct rf_stats		stats;
	unsigned int		frame_count0;
	unsigned int		frame_count1;
	int			rftimer_started;
};

struct rf_phy_dev {
	char	name[RIF_NAME_SIZE];
	void	*priv;
	u32	phy_id;
	u32	tx_atten[MAX_NUM_ANTENNAS];
	struct	rf_phy_ops *ops;
	struct  rf_ctrl_dev *ctrl_dev;
	struct	list_head list;
};

struct rf_phy_ops {
	int	(*init)(struct rf_phy_dev *phy,
				struct rf_init_params *params);

	int	(*set_timer_correction)(struct rf_phy_dev *phy,
			 struct rif_dac_params *params);

	int	(*read_dac_value)(struct rf_phy_dev *phy,
				u32 correction_type, u32 *buf);

	int	(*run_cmds)(struct rf_phy_dev *phy,
				struct rif_phy_cmd *cmds, int count);

	int	(*read_regs)(struct rf_phy_dev *phy, u32 start,
				u32 count, u32 *buff);

	int 	(*write_reg)(struct rf_phy_dev *phy, u32 reg,u32 data);
	int	(*set_tx_atten)(struct rf_phy_dev *phy, u32 reg, u32 data);
	int	(*en_dis_tx)(struct rf_phy_dev *phy, u32 tx_if, u32 cmd);
	int	(*read_rssi)(struct rf_phy_dev *phy, struct rf_rssi *rssi);
	int	(*get_rx_gain)(struct rf_phy_dev *phy,
			struct rf_rx_gain *rx_gain);
	int	(*set_rx_gain)(struct rf_phy_dev *phy,
			struct rf_rx_gain *rx_gain);
	int	(*start)(struct rf_phy_dev *phy);
	int	(*stop)(struct rf_phy_dev *phy);
	int	(*config_sniff) (struct rf_phy_dev *ad_phy,
			struct rf_sniff_params *sniff_params);
	int	(*set_gain_ctrl_mode)(struct rf_phy_dev *phy,
			struct rf_gain_ctrl *gain_ctrl);
	int	(*save_synth_table)(struct rf_phy_dev *phy,
			struct rf_synth_table *synth_table);
};

struct rf_ctrl_ops {

	int	(*init)(struct rf_ctrl_dev *ctrl_dev,
				struct rf_init_params *params);

	int	(*set_timing_source)(struct rf_ctrl_dev *ctrl_dev,
			unsigned int pps_src);

	int	(*read_regs)(struct rf_ctrl_dev *ctrl_dev, u32 start,
				u32 count, u32 *buff);

	int 	(*write_reg)(struct rf_ctrl_dev *ctrl_dev, u32 reg, u32 data);
	int	(*phy_detach)(struct rf_ctrl_dev *ctrl_dev);
	int	(*start)(struct rf_ctrl_dev *ctrl_dev);
	int	(*stop)(struct rf_ctrl_dev *ctrl_dev);
	int	(*config_sniff) (struct rf_ctrl_dev *ctrl_dev,
			struct rf_sniff_params *sniff_params);
	int	(*channel_open) (struct rf_ctrl_dev *ctrl_dev,
			struct rf_channel *chan);
	int	(*channel_close) (struct rf_ctrl_dev *ctrl_dev,
			struct rf_channel *chan);
};

struct	rf_ctrl_dev *allocate_rf_ctrl_dev(size_t priv_size,
			unsigned long flags);
int 	free_rf_ctrl_dev(struct rf_ctrl_dev *rf_dev);
int	register_rf_ctrl_dev(struct rf_ctrl_dev *rf_dev);
int	unregister_rf_ctrl_dev(struct rf_ctrl_dev *rf_dev);

struct  rf_phy_dev *allocate_rf_phy_dev(size_t priv_size,
		unsigned long flags);
int	free_rf_phy_dev(struct rf_phy_dev *phy);
int	register_rf_phy_dev(struct rf_phy_dev *phy_dev);
int	unregister_rf_phy_dev(struct rf_phy_dev *phy_dev);
int	rf_notify_dl_tti(struct rf_ctrl_dev *rf_dev);
int rf_send_event(struct rf_ctrl_dev *rf_dev,
	struct rf_event_handler *evt_handler, u16 evt);
#endif
