/*
 * RF device framework
 *
 * Author: pankaj chauhan <pankaj.chauhan@freescale.com>
 *
 * Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

#ifndef __RFDEV_IOCTL_H__
#define __RFDEV_IOCTL_H__

#include <linux/ioctl.h>
#include <linux/kernel.h>
#include <linux/types.h>

#define RIF_NAME_SIZE	20
#define RF_MAX_DEVS	6
#define MAX_NUM_ANTENNAS	4
#define PPS_TIMEOUT		4 /* seconds */
#define NUM_SYNTH_PARAMS	6
#define NUM_SYNTH_REGS		12
#define LTE_SUBFRM_PER_FRM	10
#define TIME_SYNC_NL_GROUP	1

enum rf_network_mode {
	LTE_TDD,
	LTE_FDD,
	NET_MODE_END
};

enum rf_txrxmode {
	TXRX_1T1R,
	TXRX_1T2R,
	TXRX_2T2R,
	TXRX_MODE_END
};

enum rf_band_width {
	BW_05_MHZ,
	BW_10_MHZ,
	BW_15_MHZ,
	BW_20_MHZ,
	BW_END
};

enum spi_cmd {
	SPI_WRITE = 1,
	SPI_READ,
	SPI_WAIT,
	SPI_WAIT_BBPLL_CAL,
	SPI_WAIT_RXCP_CAL,
	SPI_WAIT_TXCP_CAL,
	SPI_WAIT_RXFILTER_CAL,
	SPI_WAIT_TXFILTER_CAL,
	SPI_WAIT_BBDC_CAL,
	SPI_WAIT_RFDC_CAL,
	SPI_WAIT_TXQUAD_CAL,
	SPI_WAIT_CALDONE,
	SPI_END
};

enum rf_state {
	RF_CREATED,
	RF_PHY_ATTACHED,
	RF_PHY_INITIALIZED,
	RF_INITIALIZED,
	RF_TIMER_SYNC_AWAITED,
	RF_TIMER_SYNC_FAILED,
	RF_STOPPED,
	RF_READY,
	RF_DOWN,
	RF_SNIFFING,
	RF_STATE_END
};

enum rf_gain_ctrl_mode {
	MGC,
	FASTATTACK_AGC,
	SLOWATTACK_AGC,
	HYBRID_AGC
};

enum tmsyn_event {
	TIME_SYNC_EVT_NONE,           /* no event */
	TIME_SYNC_EVT_AD_DIS,  /* admin disable/stop */
	TIME_SYNC_EVT_AD_ENAB, /* admin enable/start */
	TIME_SYNC_EVT_TIMEINFO_RX,  /* timing info rxed */
	TIME_SYNC_EVT_TIMEINFO_MISS, /* timing info missed */
	TIME_SYNC_EVT_TIMEINFO_BO /* timing info blackout */
};

struct rf_synth_table {
	unsigned long long (*params)[NUM_SYNTH_PARAMS];
	__u8 (*reg_vals)[NUM_SYNTH_REGS];
	unsigned int count;
};

enum lte_tdd_subfrms {
	DL = 1,
	UL,
	SPL,
	END
};

struct rf_init_params {
	enum rf_network_mode mode;
	enum rf_txrxmode tx_rxmode;
	enum rf_band_width bw;
	unsigned int fq_band;
	int tdd_dl_ul_conf[LTE_SUBFRM_PER_FRM];
	unsigned int tdd_dwpts;
	unsigned int tdd_uppts;
	unsigned int tdd_gp;
	unsigned int long_cp;
	unsigned int ants;
	unsigned int capture_enabled;
	__u16 dl_delay;
	__u16 ul_delay;
};

enum rf_timer_src {
	RF_PPS_SRC_GPS = 1,
	RF_PPS_SRC_RAW_GPS,
	RF_PPS_SRC_PTP,
	RF_PPS_SRC_NLM,
	RF_PPS_SRC_END
};

enum rf_synthesizer {
	RF_RX,
	RF_TX,
	RF_BBPLL
};

struct rf_sniff_params {
	struct rf_init_params dev_params;
	unsigned long carrier_freq;
	enum rf_synthesizer rf_synth;
	enum rf_timer_src timing_src;
	unsigned int capture_duration;
};

struct rf_dev_info {
	char controller[RIF_NAME_SIZE];
	char phy[RIF_NAME_SIZE];
	enum rf_state state;
	enum rf_network_mode net_mode;
	enum rf_txrxmode tx_rxmode;
	enum rf_band_width bw;
	enum rf_timer_src timing_src;
	int sniff_enabled;
	enum rf_network_mode 	sniff_net_mode;
	enum rf_txrxmode 	sniff_tx_rxmode;
	enum rf_timer_src	sniff_timing_src;
	enum rf_band_width 	sniff_bw;
	unsigned long		sniff_carrier_freq;
	unsigned int ants;
	unsigned int symbol_len;
	unsigned int long_cp;
	unsigned int cp0_len;
	unsigned int cp1_len;
	unsigned int tx_atten[MAX_NUM_ANTENNAS];
};

#define RF_CHAN_MMAPED			0x00000001
#define RF_CHAN_DIR_TX			0x00000002
#define RF_CHAN_DIR_RX			0x00000004
#define RF_CHAN_XFER_MODE_LTE		0x00000008
#define RF_CHAN_XFER_MODE_SNIFF		0x00000010
#define RF_CHAN_PER_FRAME_EVENT		0x00000020
#define RF_CHAN_ASYNC			0x00000040

/**
 * struct rf_channel_params: RF data channel parameters
 * @ ant: Atnenna for which the channel will tx/rx IQ data
 * @ flags: Channel flags, see the flag #defs
 * @ tx_buf_size: Size of Tx buffer
 * @ rx_buf_size: Size of Rx buffer
 * @ tx_base_phys: Tx buffer base physical address. This is valid
 *   in case rf channel is mmaped (RF_CHAN_MMAPED)
 * @ rx_base_phys: Rx buffer base physical address. This is valid
 *   in case of rf channel is mmaped (RF_CHAN_MMAPED)
 */
struct rf_channel_params {

	unsigned int ant;
	__u32 flags;
	unsigned int tx_buf_size;
	unsigned int rx_buf_size;
	__u32 tx_base_phys;
	__u32 rx_base_phys;
};

struct rif_phy_cmd {
	enum spi_cmd	cmd;
	__u32	param1;
	__u32	param2;
	__u32	param3;
};

struct rif_phy_cmd_set {
	struct rif_phy_cmd *cmds;
	unsigned int count;
};

struct rif_reg_buf {
	__u32 addr;
	void *buf;
	unsigned int count;
};

struct rif_write_reg_buf {
	__u32 addr;
	__u32 data;
};

struct time_sync_1588_cnt {
	__u32 high;
	__u32 low;
};

struct time_info {
	struct time_sync_1588_cnt time_cnt;
	enum tmsyn_event event;
};

enum tmsyn_correction_mode {
	TMSYN_CORR_MODE_DAC,
	TMSYN_CORR_MODE_ADDEND,
	TMSYN_CORR_MODE_DAC_ADDEND,
	TMSYN_CORR_END
};

struct time_sync_data {
	enum tmsyn_correction_mode correction_mode;
	__u8 sync_source;
	__u32 lte_delay;
	__u32 ioctl_cmd;
	__u32 black_out_duration;
	struct time_info time_info;
};

struct rif_dac_params {
	__u32 correction_type;
	__u32 correction_value;
};

struct rif_dac_buf {
	__u32 correction_type;
	__u32 *buf;
};

struct rf_tx_buf {
	__u32 tx_if;
	__u32 tx_atten;
};

struct rf_tx_en_dis {
	__u32 tx_if;
	__u32 tx_cmd;
};

struct rf_rssi {
	__u32 ant;
	__u32 symbol;
	__u32 preamble;
	__u8 duration;
	int multiplier;
};

struct rf_rx_gain {
	int ant;
	int gain_db;
};

struct rf_gain_ctrl {
	int ant;
	enum rf_gain_ctrl_mode mode;
};

/* RF EVENTS: These Events are notifications
 * sent from kernel rf device f/w to user space
 */
struct rf_event_listener {
	__u16 evt_mask;
	unsigned int signal;
};

#define RF_EVT_ARG_SHIFT	0
#define RF_EVT_ARG_MASK		0xfff
#define RF_EVT_CHAN_SHIFT	12
#define RF_EVT_CHAN_MASK	0x3
#define RF_EVT_CMD_SHIFT	14
#define RF_EVT_CMD_MASK		0xffff

/* RF_EVENTS: Maximum 16 events as event bitmask is 16 bits
 * For defining new event, pick any unused bit
 */
#define RF_EVT_RX_FRAME		0x0001
#define RF_EVT_SNIFF_DONE	0x0002

#define RF_EVT_ALL	(RF_EVT_RX_FRAME | RF_EVT_SNIFF_DONE)
#define RF_MAKE_EVENT(evt, chan, arg)( \
			((evt & RF_EVT_CMD_MASK) << RF_EVT_CMD_SHIFT) | \
			((chan & RF_EVT_CHAN_MASK) << RF_EVT_CHAN_SHIFT) | \
			((arg & RF_EVT_ARG_MASK) << RF_EVT_ARG_SHIFT))

#define RF_EVENT_GET_CMD(evt) ((evt >> RF_EVT_CMD_SHIFT) & RF_EVT_CMD_MASK)
#define RF_EVENT_GET_CHAN(evt) ((evt >> RF_EVT_CHAN_SHIFT) & RF_EVT_CHAN_MASK)
#define RF_EVENT_GET_ARG(evt) ((evt >> RF_EVT_ARG_SHIFT) & RF_EVT_ARG_MASK)

enum dac_correction_type {
	FINE_CORRECTION = 1,
	COARSE_CORRECTION,
};

#define TIMING_SYNC_INIT	0x00000001
#define TIMING_SYNC_START	0x00000002
#define TIMING_SYNC_STOP	0x00000004
#define TIMING_SYNC_GET_TIME	0x00000010

#define FREQ_INVALID      	0xffffffff
#define TS_DELTA_INVALID      	0xffffffff
#define TIMESTAMP_INVALID	0xffffffffffffffffll

#define RF_MAGIC 'R'
#define RIF_DEV_INIT		_IOWR(RF_MAGIC, 1, struct rf_init_params*)
#define RIF_SET_TIMER_SOURCE	_IOW(RF_MAGIC, 2, unsigned int)
#define RIF_GET_STATE		_IOR(RF_MAGIC, 3, unsigned int)
#define RIF_SET_TIMER_CORRECTION _IOW(RF_MAGIC, 4, struct rif_dac_params *)
#define	RIF_RUN_PHY_CMDS	_IOW(RF_MAGIC, 5, struct rif_phy_cmd_set *)
#define RIF_READ_RSSI		_IOWR(RF_MAGIC, 6, struct rf_rssi *)
#define RIF_READ_PHY_REGS	_IOR(RF_MAGIC, 7, struct rif_reg_buf *)
#define RIF_READ_CTRL_REGS	_IOR(RF_MAGIC, 8, struct rif_reg_buf *)
#define RIF_START		_IO(RF_MAGIC, 9)
#define RIF_STOP		_IO(RF_MAGIC, 10)
#define RIF_GET_DEV_INFO	_IOWR(RF_MAGIC, 11, struct rf_dev_info *)
#define RIF_WRITE_PHY_REGS _IOR(RF_MAGIC, 12, struct rif_write_reg_buf *)
#define RIF_GET_DAC_VALUE	_IOR(RF_MAGIC, 13, struct rif_dac_buf *)
#define RIF_SET_TX_ATTEN	_IOW(RF_MAGIC, 14, struct rf_tx_buf *)
#define RIF_EN_DIS_TX		_IOW(RF_MAGIC, 15, struct rf_tx_en_dis *)
#define RIF_WRITE_CTRL_REGS 	_IOW(RF_MAGIC, 16, struct rif_write_reg_buf *)
#define RIF_READ_RX_GAIN 	_IOWR(RF_MAGIC, 17, struct rf_rx_gain *)
#define RIF_CONFIG_SNIFF 	_IOWR(RF_MAGIC, 18, struct rf_sniff_params *)
#define RIF_WRITE_RX_GAIN 	_IOW(RF_MAGIC, 19, struct rf_rx_gain *)
#define RIF_SET_GAIN_CTRL_MODE 	_IOW(RF_MAGIC, 20, struct rf_gain_ctrl *)
#define RIF_INIT_SYNTH_TABLE	_IOW(RF_MAGIC, 21, struct rf_synth_table *)
#define RIF_CHANNEL_OPEN	_IOW(RF_MAGIC, 22, struct rf_channel_params *)
#define RIF_CHANNEL_CLOSE	_IOW(RF_MAGIC, 23, unsigned int)
#define RIF_REGISTER_EVENT	_IOW(RF_MAGIC, 24, struct rf_event_listener*)
#define RIF_UNREGISTER_EVENT	_IO(RF_MAGIC, 25)

#endif
