/*
 * AF_PACKET shared function for PACKET_MMAP and PACKET_UM
 *
 *	Authors: Hemant Agrawal <hemant@freescale.com>
 *		Himanushu Seth <himanshu.seth@freescale.com>
 *	Modifier:
 *		Sachin Saxena <sachin.saxena@freescale.com>
 *		- Added SYSFS support for PACKET_UM socket options.
 *
 *	Copyright 2011-2012 Freescale Semiconductor, Inc.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */
#ifndef AF_PACKET_H
#define AF_PACKET_H

#ifdef CONFIG_PACKET_UM
#define PACKET_UM_MAX_RX_SOCK	8

#define PACKET_UM_RX_POLL_WT	12
#define PACKET_UM_TX_POLL_WT	12

#define TPKT_NAPI_WEIGHT	PACKET_UM_TX_POLL_WT
#endif

struct sock_data {
	struct sock *sk;
	int data_ready;
};

struct packet_ring_buffer {
	char		**pg_vec;
#ifdef CONFIG_PACKET_UM
	char		**pg_vec_tpkt;
#endif
	unsigned int	head;
	unsigned int	frames_per_block;
	unsigned int	frame_size;
	unsigned int	frame_max;

	unsigned int	pg_vec_order;
	unsigned int	pg_vec_pages;
	unsigned int	pg_vec_len;
#ifdef CONFIG_PACKET_UM
	unsigned int	pg_vec_order_tpkt;
	unsigned int	pg_vec_pages_tpkt;
	unsigned int	pg_vec_len_tpkt;
	unsigned int	tpacket_hdr_per_block;
	unsigned int	tpacket_hdr_max;
	unsigned int	*frame_addr;
	unsigned int	*tpacket_hdr_addr;
	struct pcb_s	*list_head;
	struct pcb_s	*list_tail;
	rwlock_t	frame_pool_lock;
#endif
	atomic_t	pending;
};

#ifdef CONFIG_PACKET_UM
/* Socket buffer recycling handler for Packet Buffers. This structure has own
 * spinlock to prevent simultaneous access. The member recycle_queue holds
 * top of recyclable socket buffer which are owned by this interface.
 * Maximum size of recyclable buffers are defined by recycle_max, and
 * current size of list is recycle_count.
 */
struct packet_skb_handler {
	/* Lock for buffer recycling queue */
	raw_spinlock_t	lock;
	short int	recycle_max;
	short int	recycle_count;
	struct sk_buff *recycle_queue;
};
#endif

#define PACKET_RING_TYPE_PM	0
#define PACKET_RING_TYPE_PUM	1

struct packet_sock {
	/* struct sock has to be the first member of packet_sock */
	struct sock		sk;
	struct tpacket_stats	stats;
#ifdef CONFIG_PACKET_MMAP
	struct packet_ring_buffer	rx_ring;
	struct packet_ring_buffer	tx_ring;
	int				copy_thresh;
#endif
	spinlock_t		bind_lock;
	struct mutex		pg_vec_lock;
	unsigned int		running:1,	/* prot_hook is attached*/
				auxdata:1,
				origdev:1;
	int			ifindex;	/* bound device	*/
	__be16			num;
	struct packet_mclist	*mclist;
#ifdef CONFIG_PACKET_MMAP
	atomic_t		mapped;
	enum tpacket_versions	tp_version;
	unsigned int		tp_hdrlen;
	unsigned int		tp_reserve;
	unsigned int		tp_loss:1;
	unsigned int		ring_type;
#endif

#ifdef CONFIG_PACKET_UM
	struct napi_struct	napi;
	struct msghdr		msg;
	struct sockaddr_ll	saddr;
#endif
	struct packet_type	prot_hook ____cacheline_aligned_in_smp;
};

struct packet_skb_cb {
	unsigned int origlen;
	union {
			struct sockaddr_pkt pkt;
			struct sockaddr_ll ll;
	} sa;
};

struct packet_filter_node {
	struct packet_filter_node *next;
	unsigned int	sock_cnxt;
	struct packet_filter_s filter;
};

#define PACKET_RB_LOCK		write_lock_bh
#define PACKET_RB_UNLOCK	write_unlock_bh
#define PACKET_RB_LOCK_INIT	rwlock_init

#ifdef CONFIG_PACKET_UM
extern struct sock_data tp_sock_data[PACKET_UM_MAX_RX_SOCK];
extern struct timer_list tpkt_timer;

extern struct net_device *priv_map_dev;

extern struct packet_filter_node *filter_head;
extern struct packet_filter_node *filter_tail;

extern struct packet_sock *po_tx;

extern unsigned int packet_sock_rx;
extern unsigned int packet_sock_tx;

extern struct packet_ring_buffer par_ring_rx;
extern struct packet_ring_buffer par_ring_tx;

extern struct packet_asf_cb g_packet_asf;
extern int tpum_tx_napi_wt;
extern int tpum_rx_poll_wt;
extern int debug_pmal;
extern unsigned long dbg_pum_sendmsg;
extern unsigned long dbg_pum_sendmsg_err;
extern unsigned long pkt_alloc_err;
extern unsigned long pum_rx_skb_err;
extern unsigned long pum_rx_skb;
extern unsigned long pum_tx_skb;
extern unsigned long pum_tx_skb_err;
extern unsigned long pmal_tx_kfree;
extern unsigned long pmal_rx_kfree;
extern unsigned long pmal_rx_free;
extern unsigned long pmal_tx_out;
#endif

extern int __packet_get_status(struct packet_sock *po,
	void *frame);

extern void __packet_set_status(struct packet_sock *po,
	void *frame, int status);

static inline struct packet_sock *pkt_sk(struct sock *sk)
{
	return (struct packet_sock *)sk;
}

#ifdef CONFIG_PACKET_UM
extern int packet_free_buffer_list(
	struct packet_ring_buffer *rb,
	unsigned int *ph,
	unsigned int nr_buffers);

extern unsigned int packet_alloc_buffer_list(
	struct packet_ring_buffer *rb,
	unsigned int ph[],
	unsigned int nr_buffers);

extern int packet_asft_add_connection(struct packet_con_s *conn);

extern int packet_asft_del_connection(struct packet_con_s *conn);

extern int packet_asft_flush_connection(void *ctxt);

extern int packet_add_buffer_to_list(struct packet_ring_buffer *rb,
			struct pcb_s *ph);

extern void packet_init_tpacket_hdr_addr(struct packet_sock *po,
			struct packet_ring_buffer *rb);

extern void packet_init_frame_desc(struct packet_sock *po,
			struct packet_ring_buffer *rb);

extern void packet_init_frame_addr(struct packet_sock *po,
			struct packet_ring_buffer *rb);

extern int tpum_rcv(struct sk_buff *skb, struct net_device *dev,
			struct packet_type *pt, struct net_device *orig_dev);

extern int tpum_snd(struct packet_sock *po, struct msghdr *msg);

extern int tpum_snd_napi(struct napi_struct *napi, int budget);

extern void *pum_lookup_frame(struct packet_sock *po,
		struct packet_ring_buffer *rb,
		unsigned int position,
		int status);

extern void packet_reset_recycle_queue(void);
extern void packet_free_recycle_queue(int lock_flag);
extern int gfar_packet_queue_deinit(void);
extern int gfar_packet_queue_init(void);
/* SYSFS interface */
extern int packet_um_sysfs_init(void);
#endif

#endif
